#!/bin/awk -f
## Replace MASS device parameters (prior to reprocessing)
## and check their final existence in output data.
## 
## Parameters: PA, PB, PC, PD - channels' A-D non-Poissonity factor
##             nA, nB, nC, nD - channels' A-D non-linearity time [ns]
##             k - magnification factor
## ATTENTION! Parameter names are case-SENSITIVE
## Parameters are taken from command line
##
## Description: replace occurrences of the mentioned parameters in the P-lines
## of mass-file with the specified values. If no parameters encountered in
## heading P- and #-lines, replace with given in parameters as 
## Channel\ChannelA(B,C,D)\NonPoisson and NonLinearity. 
##
## Use as: param.awk [<]<mass-file> [param1=val1 [param2=val2] ...] >result-file
BEGIN{
	# Default P, n-parameters are invalid and thus not substituted in output
	PA = PB = PC = PD = 0
	nA = nB = nC = nD = -1
	# Default scattering is left zero:
	SA = SB = SC = SD = 0
	k = -1
	# Correction indicators
	ncorr = 0
	# Parameter-seen indicator:
	seen_k = 0
	seen_n = seen_p = seen_s = ""
	IGNORECASE = 1
}

NR==1 {
	if (PA>0) s="PA="PA
	if (PB>0) s=s" PB="PB
	if (PC>0) s=s" PC="PC
	if (PD>0) s=s" PD="PD
	if (nA>=0) s=s" nA="nA
	if (nB>=0) s=s" nB="nB
	if (nC>=0) s=s" nC="nC
	if (nD>=0) s=s" nD="nD
	if (SA>0)  s=s" SA="SA
	if (SB>0)  s=s" SB="SB
	if (SC>0)  s=s" SC="SC
	if (SD>0)  s=s" SD="SD
	if (k>0)  s=s" k="k
	print "## ==== PARAM CORRECT: "s
}

$1=="P" && $0~/.*Magnification=.*/ && $0~/.*Optics.*/ {
	if (k>0) {
		gsub("=.*","="k)
		ncorr++
	}
	seen_k = k
}

# Here we catch Channel A, channelA etc varieties simultaneously:
$1=="P" && $0~/.*channel *[abcd].*/ && $0~/.*nonlinearity.*/ {
	nc = 0
	if ($0~/.*CHANNEL *A.*/) {if (nA>=0) nc+=gsub("=.*","="nA); seen_n=seen_n"A"}
	if ($0~/.*CHANNEL *B.*/) {if (nB>=0) nc+=gsub("=.*","="nB); seen_n=seen_n"B"}
	if ($0~/.*CHANNEL *C.*/) {if (nC>=0) nc+=gsub("=.*","="nC); seen_n=seen_n"C"}
	if ($0~/.*CHANNEL *D.*/) {if (nD>=0) nc+=gsub("=.*","="nD); seen_n=seen_n"D"}
	if (nc==1) {
		print "## PARAM n"substr(seen_n,length(seen_n),1)" REPLACED"
		ncorr++
	} else if (nc>1) {
		print $2"T"$3" PARAM MULTIPLE n-MATCHES FOUND!" > "/dev/stderr"
	}
}
$1=="P" && $0~/.*channel *[abcd].*/ && $0~/.*nonpoisson.*/ {
	nc = 0
	if ($0~/.*CHANNEL *A.*/) {if (PA>0) nc+=gsub("=.*","="PA); seen_p=seen_p"A"}
	if ($0~/.*CHANNEL *B.*/) {if (PB>0) nc+=gsub("=.*","="PB); seen_p=seen_p"B"}
	if ($0~/.*CHANNEL *C.*/) {if (PC>0) nc+=gsub("=.*","="PC); seen_p=seen_p"C"}
	if ($0~/.*CHANNEL *D.*/) {if (PD>0) nc+=gsub("=.*","="PD); seen_p=seen_p"D"}
	if (nc==1) {
		print "## PARAM P"substr(seen_p,length(seen_p),1)" REPLACED"
		ncorr++
	} else if (nc>1) {
		print $2"T"$3" PARAM MULTIPLE P-MATCHES FOUND!" > "/dev/stderr"
	}
}
$1=="P" && $0~/.*channel *[abcd].*/ && $0~/.*scatter.*/ {
	nc = 0
	if ($0~/.*CHANNEL *A.*/) {if (SA>0) nc+=gsub("=.*","="SA); seen_s=seen_s"A"}
	if ($0~/.*CHANNEL *B.*/) {if (SB>0) nc+=gsub("=.*","="SB); seen_s=seen_s"B"}
	if ($0~/.*CHANNEL *C.*/) {if (SC>0) nc+=gsub("=.*","="SC); seen_s=seen_s"C"}
	if ($0~/.*CHANNEL *D.*/) {if (SD>0) nc+=gsub("=.*","="SD); seen_s=seen_s"D"}
	if (nc==1) {
		print "## PARAM S"substr(seen_s,length(seen_s),1)" REPLACED"
		ncorr++
	} else if (nc>1) {
		print $2"T"$3" PARAM MULTIPLE S-MATCHES FOUND!" > "/dev/stderr"
	}
}

# No all nonpoissonity parameters encountered in preamble:
seen_p!="ABCD" && substr($1,1,1)!="#" && $1!="P" {
	print "## PARAM printing new P values"
	if (!(seen_p~/.*a.*/)) {
		if (PA<=0) {
			PA=1
			print $2"T"$3" PARAM PA defaulted to 1!" > "/dev/stderr"
		}
		print "P "$2" "$3" Channels\\ChannelA\\NonPoisson="PA
		ncorr++
	}
	if (!(seen_p~/.*b.*/)) {
		if (PB<=0) {
			PB=1
			print $2"T"$3" PARAM PB defaulted to 1!" > "/dev/stderr"
		}
		print "P "$2" "$3" Channels\\ChannelB\\NonPoisson="PB
		ncorr++
	}
	if (!(seen_p~/.*c.*/)) {
		if (PC<=0) {
			PC=1
			print $2"T"$3" PARAM PC defaulted to 1!" > "/dev/stderr"
		}
		print "P "$2" "$3" Channels\\ChannelC\\NonPoisson="PC
		ncorr++
	}
	if (!(seen_p~/.*d.*/)) {
		if (PD<=0) {
			PD=1
			print $2"T"$3" PARAM PD defaulted to 1!" > "/dev/stderr"
		}
		print "P "$2" "$3" Channels\\ChannelD\\NonPoisson="PD
		ncorr++
	}
	seen_p="ABCD"
}

# No all nonlinearity parameters encountered in preamble:
seen_n!="ABCD" && substr($1,1,1)!="#" && $1!="P" {
	print "## PARAM printing new n values"
	if (!(seen_n~/.*a.*/)) {
		if (nA<0) {
			nA=0
			print $2"T"$3" PARAM nA defaulted to 0!" > "/dev/stderr"
		}
		print "P "$2" "$3" Channels\\ChannelA\\NonLinearity="nA
		ncorr++
	}
	if (!(seen_n~/.*b.*/)) {
		if (nB<0) {
			nB=0
			print $2"T"$3" PARAM nB defaulted to 0!" > "/dev/stderr"
		}
		print "P "$2" "$3" Channels\\ChannelB\\NonLinearity="nB
		ncorr++
	}
	if (!(seen_n~/.*c.*/)) {
		if (nC<0) {
			nC=0
			print $2"T"$3" PARAM nC defaulted to 0!" > "/dev/stderr"
		}
		print "P "$2" "$3" Channels\\ChannelC\\NonLinearity="nC
		ncorr++
	}
	if (!(seen_n~/.*d.*/)) {
		if (nD<0) {
			nD=0
			print $2"T"$3" PARAM nD defaulted to 0!" > "/dev/stderr"
		}
		print "P "$2" "$3" Channels\\ChannelD\\NonLinearity="nD
		ncorr++
	}
	seen_n="ABCD"
}

# No all scattering parameters encountered in preamble:
seen_s!="ABCD" && substr($1,1,1)!="#" && $1!="P" && (SA*SA+SB*SB+SC*SC+SD*SD!=0) {
	print "## PARAM printing new S values"
	if (!(seen_s~/.*a.*/) && SA!=0) {
		print "P "$2" "$3" Channels\\ChannelA\\Scatter="SA
		ncorr++
	}
	if (!(seen_s~/.*b.*/) && SB!=0) {
		print "P "$2" "$3" Channels\\ChannelB\\Scatter="SB
		ncorr++
	}
	if (!(seen_s~/.*c.*/) && SC!=0) {
		print "P "$2" "$3" Channels\\ChannelC\\Scatter="SC
		ncorr++
	}
	if (!(seen_s~/.*d.*/) && SD!=0) {
		print "P "$2" "$3" Channels\\ChannelD\\Scatter="SD
		ncorr++
	}
	seen_s="ABCD"
}

# No magnification parameter encountered in preamble:
seen_k!=k && substr($1,1,1)!="#" && $1!="P" {
	if (k>0)
		print "P "$2" "$3" Optics\\Common\\Magnification="k
	else print $2"T"$3" PARAM k absent in mass-file!" > "/dev/stderr"
	seen_k = k
}

!($1=="##" && $0~/PARAM/) { print }

END{
    print "## PARAM CORRECTED="ncorr
}
