/*--------------------------------------------------------------------------*/
/** @file device.h
@brief  High-level interaction with the device bearing the manager functions 
which implement the algorithms of modes.

                  MASS project: TURBINA module file
                       header file for device.cpp

The algorithms represent the sequences of actions which implement the one 
accumulation-time cycle of measurements
by means provided by counter.h, motor.h, power.h, tracker.h and light.h.



 @author O. Voziakova, Sternberg Institute (ovoz@sai.msu.ru)
 
 @version 2.0
*/
#ifndef DEVICE_H
#define DEVICE_H

#include <qobject.h>
#include "exchange.h"
#include "module.h"
#include "channel.h"
#ifdef __MOTOR__
	#include "motor.h"
#elif defined __SCANMOTOR__
	#include "scanmotor.h"
#endif
#include "tracker.h"
#include "scan.hxx"
#include "trbtime.h"
#include "trbconst.h"
#include "auxiliary.h"
#include "svexchange.h"
#include "trbscenario.h"

#define TM_PREC			(1.5)	// time precision for count timeout interval

#define INIT_TMOUT				10	// max dlitel'most' initializatii
#define PARK_TMOUT				10	// max dlitel'most' parkovki
#define SV_DELAY				5	// max dopustimaya oshibka v ocenke vremeni vypolneniya

typedef double matrix43[N_CHAN][3];

class PointDouble {
private:
	double _x;
	double _y;
public:
	PointDouble(void) { _x = 0; _y = 0; }
	~PointDouble() {}
	double x(void) { return _x; }
	double y(void) { return _y; }
	void setX(double newX) { _x = newX; }
	void setY(double newY) { _y = newY; }
};

//void	timeout_action(void);
//void	*read_status(void *arg);
//void	*read_data(void);
//void	terminate_program(void);
//void*	count_cycle(void);
//void	initThreads(void);

//void	sendToScreen(const QString& text);
void	sendToOutFile(const QString& text,
			const char* prefix="> ");
void	sendErrorInfo(const QString& text,
				SVexchange::status erstatus=SVexchange::ERFAT);


class DevManager : public QObject {
	Q_OBJECT

protected:
	void timerEvent(QTimerEvent*);

private:
#ifdef __COUNTER_BUTTON__
	Counter *masterCounter;
	Counter *slaveCounter1;
	Counter *slaveCounter2;
	Counter *slaveCounter3;
	Counter *testCounter;
#else
	Bicounter *slaveBicounter;
	Bicounter *masterBicounter;
	Bicounter *testBicounter;
#endif

	scan_t scan[N_CHAN];

#ifdef __COUNTER_BUTTON__
	Counter counters[MAX_N_COUNTERS];
#else
	Bicounter bicounters[N_BICOUNTERS];
#endif
	Channel channel[N_CHAN];

#ifdef __MOTOR__
	Motor motor;
#elif defined __SCANMOTOR__
	ScanMotor scanmotor;
#endif
	Tracker tracker;
	Auxiliary auxiliary;

	bool	_withAutoGuide;

	int	_polling;

	int	_temperPolling;
	int	_overLightPolling;
#ifdef __COUNTER_BUTTON__
	int	_buttonsPolling;
#else
	int _mirrorPolling;
#endif
	int	_baseTimeTimerPolling;

	int	_baseTimeTimer;
	int	_temperTimer;
	int	_hvTimer;
#ifdef __COUNTER_BUTTON__
	int	_buttonsTimer;
#else
	int _mirrorTimer;
#endif

#ifdef __COUNTER_BUTTON__
	int	_processButtons_flag;
#else
	int	_processMirror_flag;
#endif
	int	_getTemper_flag;

// Flagi sostoyaniya:
	bool	isOverLight;
	bool	isScenario;
	bool	_isCounting;
	bool	_isPlayBack;
	bool	_isTracking;
	bool	_isUserStop;
	bool	isFirst;	// true if first loadConst & init
//	bool	isFirstInit;
	bool	isParking;	// true poka parkuyetsia
	bool	isQuiting;

	const char* _name_A;
	const char* _name_B;
	const char* _name_C;
	const char* _name_D;

	int _total_test_err1,_total_test_err2;
#ifdef __COUNTER_BUTTON__
	int _total_test_err3,_total_test_err4; 
#endif
	double _zShiftNormal,_zShiftGeneral,_zShiftGeneralSecond;

// Files & Dir-s
	int		_sciCountSave;		//	sci-mode: save to file or not
	int		_techCountSave;	//	tech-mode: save to file or not
	int		_logSave;		//	save to file or not
	QString		_logDirName;
	QString		_dataDirName;
	QString		_checkFile;	//for atmos 
	QString		_sumFileName;	// summary
	QString		_dumpFileName;	// summary
	QString		_massInputFileName; // Play Back

// Modes
	int	_dataBlockSize;
	int	_dataBlockSize0;
	int	_normalBlockNumber;
	int	_generalBlockNumber,_generalSecondBlockNumber;
	int	_focusBlockNumber;
	int	_centerBlockNumber;

	double	_gapRA; // for tracker
	double	_gapDec; // for tracker

	double	_normalExposition;
	double	_generalExposition;
	double	_focusExposition;
	double	_centerExposition;
	int	_focusScanStart,_focusScanFinish;
	int	_centerScanStart,_centerScanFinish;
	int	_workZone;
	int	_centralZone;
	int	_maxIterCentering;
	double	_speedRA;
	double	_speedDec;
	double	_normalBaseTime;
	double	_generalBaseTime,_generalSecondBaseTime;
	double	_focusBaseTime;
	double	_centerBaseTime;

// Cn2 restor.: fi[X]ed, f[L]oating layers
	bool	_normalFixedMethod;
	bool	_generalFixedMethod;

// Polling
	double	_temperPollingTime;
#ifdef __COUNTER_BUTTON__
	double	_buttonPollingTime;
#else
	double  _mirrorPollingTime;
#endif
	double	_highPollingTime;

#ifdef __COUNTER_BUTTON__
// Buttons mode
	int	 	_aperturesMode;
	int		_illumMode;
	int		_shiftMode;
	int		_doorMode;
#endif

// Initial & current:
#ifdef __MOTOR__
	int _inputDoor;
	int _initialStop;
#elif defined __SCANMOTOR__
	int _initScanMotorPosition;
#endif
	int _illumValue;
	int _illumination;
#ifdef __COUNTER_BUTTON__
	int _buttons;
	int _buttonMode;
#endif
	int _highVoltage;
	int _hvLevel;
	int _safety;
	QArray<double> _discrLevels;

	double	_temperature;


// Parking:
#ifdef __MOTOR__
	int _inputDoorPark;
	int _closeApert;
#elif defined __SCANMOTOR__
	int _parkScanMotorPosition;
#endif
#ifdef __COUNTER_BUTTON__
	int _buttonsPark;
#endif
	int _hvPark;
	int _illumPark;

// Tests:
	int _playInitialScenario; // On/Off

	double	_testVAmpl;

	int			_nStatisticTestLight;
	QArray<double>	_statisticTestLight;
	int			_nDetectorCountingLight;
	QArray<double>	_detectorCountingLight;
	int			_nDetectorCountingVoltage;
	QArray<int>	_detectorCountingVoltage;
	int			_nDetectorCountingDiscrimination;
	QArray<double>	_detectorCountingDiscrimination;
	int				_nDetectorStatisticLight;
	QArray<double>	_detectorStatisticLight;
	double			_detectorTestLight;	

	int		_showStarApert;
	int		_testApert;
	int		_normalApert,_generalApert,_generalSecondApert,_focusApert;
	int		_centerApert,_backgroundApert,_countApert,_statisticApert;

	int		_backgroundApShift;
	double		_backgroundTelShift;
	int		_backgroundMethod;

	double	_exchangeTestTime,_detectorTestTime;
	double	_detectorCountingTime,_detectorStatisticTime;
	double	_backgroundTime;
	double	_normalAccumTime;
	double	_generalAccumTime;
	double	_estimationAccumTime;
	double	_focusAccumTime;
	double	_centerAccumTime;
	double	_centerRefreshAccumTime;
	double	_fluxTolerance;
	double	_nonPoisTolerance;
	QArray<double>	_testCounts;

	QArray<double> _backgroundCorrectionCoefficients;
	QArray<double> _backgrounds;
	QArray<double> _fluxes;
	QArray<double> _nonPoisson;
	QArray<double> _nonLinearity;

// Optic & Site:
//	double	_scale;
	double	_sinPosAngle;
	double	_cosPosAngle;
	int	_mirror;

	double _shiftRA;
	double _shiftDec;

	QDateTime _startBaseTime;

// Scenario

//	QArray<char>		currentScenario;
	QArray<trbscenario::mode>		currentScenario;
	QArray<QString*>	scenarioName;
	QArray<QString*>	scenarioScript;
	int	_nScenario, _nUserScenario;
	int	_currentScenarioNumber;
	int	_stopOnError;

	int		_stopScenario;
	bool	_pauseScenario;
	int		_scenarioStartPosition;
	int		_playScenarioSVcomid;
	int		_stopScenarioSVcomid;
	int		_stopCountingSVcomid;

	matrix43 	*coef;
	PointDouble	*delta;

#ifdef	__COUNTER_BUTTON__
	int _nCorrected[MAX_N_COUNTERS];
#else
	int _nCorrected[N_CHAN];
#endif
// PlayBack
	long	massInFilePos;

private:
	void	configChannels(void);
	int		testMirror(int dialog);
	void	endCounting(void);
	void	disablePolling(void);
	void	enablePolling(void);
	void	startTemperPolling(void);
	void	stopTemperPolling(void);	
	void	startOverLightPolling(void);
	void	stopOverLightPolling(void);
#ifdef __COUNTER_BUTTON__
	void	startButtonsPolling(void);
	void	stopButtonsPolling(void);
#else
	void	startMirrorPolling(void);
	void	stopMirrorPolling(void);
#endif
//	void	startSVpolling(void);
//	void	stopSVpolling(void);

	int		setBlSize(int,int);
	int 		calcBlockNumber(double& baseTime, double exposition);
	int 		calcExpositionNumber(int _blockNumber);
	int		shiftAt(int,int);
	int		shiftTelescope(int coordinate, // RA or DEC
					double shift, // time-sec
					int dialog);

	void		openProgressor(const QString&);	//
	void		closeProgressor(void);		//
	
	void		openMode(const QString&,double);	//
	void		closeMode(void);		//

	void		sendScenarioName(const QString& name);	//
	void		openScenario(const QString&,double);	//
	void		closeScenario(void);		//

	int		runNormal(FILE* mass_in_file=NULL);
	int		runGeneralized(FILE* mass_in_file=NULL);
	int		runPlayBackMode(FILE* mass_in_file, trbscenario::mode which);
	int		runFluxEstimation(void);
	void		calcApertShift(//double& shiftRA,
					//double& shiftDec,
					double shiftX,
					double shiftY);
	bool		isStarInWorkZone(void);
	bool		isStarInCentralZone(void);
	void		starInDiaphragm(void);
	int		scanAperture(double accumTime,
				double baseTime,
				int blockNumber,
				double exposition,
				int aperture,
				int scanStart,
				const QString& modeName,
				bool isFocus);
	int		runFocusControl(void);
	int		runCenterControl(void);
	int		runCentering(void);
	int		runExchangeTest(void);
	int		runDetectorTest(void);
	int		runStatisticTest(void);
	int		runDetectorStatisticMeasurement(void);	//!
	int		runDetectorCountingMeasurement(void);		//!
	int		runBackgroundCorrection(void);			//!
	int		runBackgroundMeasurement(void);			//!
	int		runCenterCorrection(void);				//!
	int		runTrackCalibration(void);

	void	terminateCountMode(int);		//!
//	void	processSV(void);
	int		testOverLight(int);
	void	getTemper(int);
#ifdef __COUNTER_BUTTON__
	void	processButtons(int);
#endif

public:
	DevManager(void);		//
	~DevManager();
	void	closeDevice(bool withPark=true);
	void	quitProgram(void);
	bool	withAutoGuide(void) { return _withAutoGuide;}

	int		startCounting(int dialog, bool testViewerMirror=true);
	void	cleanStopScenario(void);
	void	setCoef(matrix43* p_matrix) {coef = p_matrix;}
	void	setDelta(PointDouble* p_point) {delta = p_point;}
	void	setMassInputFileName(const char* name) {_massInputFileName = QString(name);}
	void	setPlayScenarioSVcomid(int comid) { _playScenarioSVcomid = comid;}
	void	setStopScenarioSVcomid(int comid) { _stopScenarioSVcomid = comid;}
	void	setStopCountingSVcomid(int comid) { _stopCountingSVcomid = comid;}
	double	getRemainTimeMode(void);

	int		playInitialScenario(void);
	int		calcTimeoutInterval(double,int);
	void	startDevTimer(int);
	void	stopDevTimer(void);
	void	waitData(void);
	int	testUserStop(void);
	void	controlUserTrack(void);
	int	catchAppendix(void);
	void	outCountResult(void);

	void	resetDriverBlockCounter(void);

	bool	isCounterReady(void);	//i
	void	create(void);
	int		setInduceRegime(int dialog);	//i
//	Counter* getTestCounter(void) { return(testCounter); }
	void 	resetTestData(void);
	void	searchTestErrorsOneCountDevice(void);
	void	searchTestErrorsAllCountDevices(void);
	double	resumeTestOneCountDevice(void);
	double	resumeTestAllCountDevices(void);

	// Added for ver1.3: start and end-accum.Time 
	int	accumResume(QDateTime midAccumTime, 
				QDateTime startAccumTime, QDateTime endAccumTime,
				bool isgen,
				bool isFixedLayersMethod,
				bool isPlayBack=false);
	// Added for ver1.3: start and end-accum.Time 
	int	makeAccumResultScind(QDateTime midAccumTime,
				QDateTime startAccumTime, QDateTime endAccumTime,
				bool isgen);
	int	makeAccumResultAtmos(QDateTime midAccumTime,
				bool isgen, bool isFixedLayersMethod);
	int	showAccumResultScind(bool isgen, bool isStatTest=false);
	void	showAccumResultAtmos(int row, bool isgen);

	void	prepareFiles(void);
	void	searchCountersSyncStatus(void);
	int		prepareScenario(const char* section, const char* subsection,
				const char* key, int scenarioNumber, const char* name);
	void	loadConst(void);

	bool	cleanData(int& expositionNumber,
			long *addr = 0,
			bool isPlayBack = false);
	void	handleOverLight(void);

	void	setStartBaseTime(void);
	QDateTime	setMidTime(QDateTime startTime, QDateTime stopTime);

	int	isIllumOn(void) {return(auxiliary.isIllumOn()); }
	int	isLightOn(void) {return(auxiliary.isLightOn()); }
#ifdef __COUNTER_BUTTON__
	int	isButtonsOn(void) {return(auxiliary.isButtonsOn()); }
#endif
	int	isSafetyOn(void) {return(auxiliary.isSafetyOn()); }

	int	getStopCorrect(void);
	int	getStopCorrectNew(void);
	int	getNStops(void);
	int	currentAperture(void);
	int	getCurrentBrightness(void) { return(auxiliary.getCurrentBrightness()); }
	int	getScenarioNumber(void) {return(_currentScenarioNumber);}
	QString& getScenarioName(int number);
	int	getMaxScenarioNumber(void) {return(_nScenario);}
	int getMaxUserScenarioNumber(void) {return(_nUserScenario);}
	int	startShowStar(void);
	void	finishShowStar(int oldAperture);
//	void	playBack(QString& fileName);
	void	endPlayBack(void);
	bool	isPlayBack(void) {return _isPlayBack;}
	bool	isCounting(void) {return _isCounting;}
	bool	isTracking(void) {return _isTracking;}
	void	acceptBackgroundCorrection(void);
	void	acceptCenterCorrection(void);
	void	acceptTrackCalibration(void);
	void	endTrackCalibration(void);
	void	acceptComment(const QString& comment);
	void	getBackgroundCorrectionCoefficients(matrix43*);

	void	statistic_test_mode(void);
	void	detector_test_mode(void);
	void	exchange_test_mode(void);
	void	detector_counting_mode(void);
	void	detector_statistic_mode(void);
	void	background_corr_mode(void);
	void	background_measur_mode(void);
	void	centering_mode(void);
	void	centering_corr_mode(void);
	void	track_calibration_mode(void);
	void	focus_control_mode(void);
	void	center_control_mode(void);
	void	normal_mode(void);
	void	flux_estimation_mode(void);
	void	generalized_mode(void);
	double	getAccumTime(char mode);
	double	getCurrentMaxBaseTime(void);
	void	playback_mode(void);
	void	scenario_mode(void);
//VK
	const char* getBackground(void);
	const char*	getFluxes(void);
	const char* getScind(void);
	const char*	getIntegral(void);
	const char* getXProfile(void);
    const char* getLProfile(void);
    const char* getCentering(void);
    const char* getScenario( int numb );
    int         setScenario( int numb, const char *scen );
	
private:
	bool	isUserStop(void);
	long	getCountAddress(FILE* mass_in_file,
			bool isgen, QDateTime& startBaseTime);
	                                                
public slots:

	int		init(void);
//VK	
	int		gui_init(void);
	void	park(void);

	void	stopCounting(void);
	void	cleanUserStop(void);
	void	stopScenario(void);
	void	setPause(bool isPause);
	void	change(void);
	int	greatTrack(char coordinate, char direct,
				double trackTime); // trackTime in sec
	int	trackToCenter(void);
	int		fineShift(int);

//	void	playInitialScenario(void);
	trbscenario::mode	decodeMode(char mode);
	void	setCurrentScenario(int number);
	void	testVAmpl(void);

private:
	int		setBrightness(int brightness, int dialog);
	void	changeBrightness(int up);
#if defined __MOTOR__ || defined __SCANMOTOR__
	int		synchAt(int shift, int dialog);
#endif
	int		changeAperture(int shift, int dialog);
	int		setDiscriminations(QArray<double>& levels, int dialog);
	void	getDiscriminations(QArray<double>& levels);
	int		setBlockNumber(int blockNumber, int& expositionNumber,
							int dialog, bool isPlayBack=false);
	void	setProgress(double);
	int		setExpos(double expos, int dialog);
	void	setRemainTimeMode(double);
	void	setRemainTimeScenario(double);
	double	getRemainTimeScenario(void);

public:
#ifdef __COUNTER_BUTTON__
	int		testButtons(int dialog);
#endif
	int		testIllum(int dialog);
	int		testMotorStatus(int dialog);
	const QString& getStopName(int stopNumber);
	void	startAdjustAperture(void);
	void	acceptAdjustAperture(int dialog);
#ifdef __SCANMOTOR__
	int moveScanMotor(int position, bool windingOff, int dialog);
	int shiftScanMotor(int shift, int dialog);
#endif
	int		setVoltage(int hvLev, int dialog, bool wait=true);
	int		getVoltage(void) {return(auxiliary.getVoltage());}
	int		testPowerStatus(int dialog);
	int		isHVOn(int dialog);
	int		isDoorOpen(void);
	double	getFoclen(const QString& apName, int dialog=1);
	double	getFoclen(int aperture, int dialog=1);
#ifdef __COUNTER_BUTTON__
	int	resetTestCounter(void);
	int	switchTestCounter(void);
	int	runTestOneCounter(void);
	int runTestAllCounters(void);
#else
	int	resetTestBicounter(void);
	int	switchTestBicounter(void);
	int	runTestOneBicounter(void);
	int runTestAllBicounters(void);
#endif
	int	resetCountDevices(int dialog);
	int runAllCounters(void);

public slots:
	int		setLight(double value, int dialog,
				int waitInterval = LIGHT_WAIT_INTERVAL);
	double	getLight(void) {return(auxiliary.getLight());}
	int		turnIllum(const QString& state, int dialog);
	int		turnIllum(int state, int dialog);
	int		turnLight(int state, int dialog,
				int waitInterval=LIGHT_WAIT_INTERVAL);
//#ifdef __COUNTER_BUTTON__
	int		turnButtons(const QString& state, int dialog);
	int		turnButtons(int state, int dialog);
//#endif
	void	setBrightnessUp(void);
	void	setBrightnessDown(void);
	int		turnDoor(const QString& state, int dialog,
							bool restoreApert = true);
	int		turnDoor(int state, int dialog,
							bool restoreApert = true);
	int		shiftAperture(int shift, int dialog);
	int		setAperture(const QString& apName, int dialog,
							bool hard=false);
	int		setAperture(int stopNumber, int dialog,
							bool hard=false);
	void	nullSet(void);
	int		movePlus(int dialog);
	int		moveMinus(int dialog);
	int		stopTrack(void);
	int		forceTrack(void);
	int		track(char coordinate, char direct);
	void	turnHV(int state, int dialog);
	void	turnSafety(const QString& state, int dialog);
	void	turnSafety(int state, int dialog);
	int		stopCounters(int dialog);	//i
};

#endif
