/**     @file iocount.h
@brief	Input/Output of the individual photon counts in a binary "count-file".

	MASS project: header file for iocount.cpp
 
	This module implements the Input/Output operations with disk for saving and
	restoring the counts in count buffers of MASS detectors. The aim is to
	allow for saving the night-time count series in a binary file and for the
	\em  off-line playback the recorded series of input channel counts to
	repeat their processing and display the output parameters. Alternatively,
	the counts may be read and processed externally.

	The counts from all channels of the device are written in a binary format
	by the utility ioc::write() which returns the \em start-position \em
	address of the written series. This address is saved somewhere else (e.g.
	in main MASS output file). Afterwards, for reading the \em count-file, the
	address is fed to ioc::read() which restores the counts in the channel
	buffers. After this, the processing may go on as if the count data were
	just received from counters.

	The binary count file should be opened with an attribute "a+b" and closed
	upon completion of reading/writing externally. If problem occurs (bad input
	address for ioc::read() or low-level I/O problem), the error is set by
	these utilities to be detected by nr::erget() and nr::ermessage().

	The buffers are made of numbers of the type \em count_t which is defined in
	this module. All the bytes of count_t numbers are written to the disk.
	Thus, the two-byte unsigned or signed integers are seemed to be best suited
	for MASS, for which we do in this module:
	\verbatim
	#include <stdint.h> // or <sys/types.h>
		...
	typedef count_t	uint16_t ;	// or int16_t
	\endverbatim
	
	Normally, the buffers are written by ioc::write() or read by ioc::read()
	simultaneously if they contain the signal of the same time interval of
	acquisition (\em base-time). Thus, for 4-channel MASS device, \em four
	buffers are written/read in one time.  Each these count buffers is preceded
	by a \em descriptor consisting of two numbers of the same type \em
	count_t. 

	The first number is a number of count buffers which are written starting
	from the current position and which belong to the \b same base-time.
	This allows to check that the buffers are synchronized in acquisition time
	while reading the count file. 

	The second number is the length of the following buffer in counts (one
	should multiply this number by the sizeof(count_t) to get length in bytes).
	Currently, MASS channels are forced to fill their count buffers always with
	equal amount of counts.
	
	The example of the record of one base-time of MASS with 4 channels each
	having 1000 counts (binary count_t-type numbers are denoted as <..>) :
	\verbatim
<4> <1000> <channel_1 count_1> <channel_1 count_2> ... <channel_1 count_1000> 
<3> <1000> <channel_2 count_1> <channel_2 count_2> ... <channel_2 count_1000> 
<2> <1000> <channel_3 count_1> <channel_3 count_2> ... <channel_3 count_1000> 
<1> <1000> <channel_4 count_1> <channel_4 count_2> ... <channel_4 count_1000>
	\endverbatim
	
	Thus, reading the first two bytes of the file or of the base-time record
	allows one to get know the number of buffers written further for this
	base-time and the size of (each) buffer.

	@author	N.Shatsky (kolja@sai.msu.ru)
	@version	1.1	Descriptor added for each record
*/
#ifndef _IOCOUNT_H
#define _IOCOUNT_H

/** Include the definitions of the fixed-length integers: may alternatively be
<sys/types.h>
*/
#include <stdint.h>

#include <stdio.h>

/** The count type will be unsigned 16-bit integer type */
typedef uint16_t count_t ;
/** Maximal count_t (implementation works for unsigned integral types) */
const count_t max_count_t = (count_t)(-1) ; 

/**	@brief
	Declarations of the channel count type and counts I/O operations */
namespace ioc {

/**
	@brief	Write the channel count series to disk
	@param	fcnt	structure of the opened binary file of counts
	@param	buffer	array of \em nbuf pointers to channel count buffers
	@param	buflen	length of the channel buffer in units of counts (not bytes!)
	@param 	nbuf	Number of buffers (e.g. number of active channels)
	@return	Starting position in file (in bytes from beginning) where
		the counts were written or (-1) on error
	
	The function uses the \em fseek() to set the file position to the end of
	file (where it is, normally, set already), gets this current position to be
	returned on success, and writes sequentially all \em nbuf channel buffers to
	disk. Before writing of each \em i-th channel buffer (\em i=0..nbuf-1), the
	descriptor is written consisting of two numbers: {\em nbuf-i, \em buflen }.

	After writing, the file pointer is retained at the next position after the
	last written data number.
	
*/
long write(FILE * fcnt, count_t ** buffer, int buflen, int nbuf) ;

/**
	@brief	Read the channel count series from disk
	@param	fcnt	structure of the opened binary file of counts
	@param	buffer	array of \em nbuf pointers to channel count buffers
	@param	buflen	On enter:
		Maximal length of the channel buffer in units of counts (not bytes!)
					On exit:
		Actual length of channel buffers
	@param 	nbuf	On enter:
		Expected number of buffers (e.g. number of active channels)
					On exit:
		Encountered number of buffers (error if differs from \em nbuf on enter)
	@param	addr	Starting position in file (in bytes from beginning) where
		the counts should be read from. 
		
	The address is checked to be only the multiple of the size of the \em
	count_t type.
	
	The function uses the \em fseek() to set the file position to the given \em
	addr, and reads sequentially \em nbuf records \em each consisting of a
	two-number descriptor plus a count buffer read into	the respective \em
	buffer[i] (\em i=0..nbuf-1).

	Each of descriptors is checked to be consistent with the parameters \em nbuf
	and \em buflen:
	- the first number is checked to be equal to \em nbuf-i
	- the second number must not exceed the parameter \em buflen value
	- the second number must be equal for all \em nbuf records
	
	If one of these conditions are not justified, the error with the code
	nr::EROFL is returned and the further reading stops. The error nr::ERFIO is
	returned when not all the counts are successfully read or \em addr is not
	fseek()-able.

	The content of a first descriptor is returned in place-holders \em nbuf and
	\em buflen in any case, either on error or on success.
	
	After reading, the file pointer is retained at the next position after the
	last data number read.
	
*/
void read(FILE * fcnt, count_t ** buffer, int * buflen, int * nbuf, 
	long addr) ;

}	// end of namespace ioc

#endif
