/*--------------------------------------------------------------------------*/
/** @file light.h
@brief  Implementation of interaction (command calls and replies from modules, 
their current data and status keeping, etc.) with buttons and control/illumination
LEDs module.

                  MASS project: TURBINA module file
                       header file for light.cpp



 @author O. Voziakova, Sternberg Institute (ovoz@sai.msu.ru)
 
 @version 2.0
*/
#ifndef LIGHT_H
#define LIGHT_H

#include "module.h"

#define MIN_ILLUM_LEVEL	0
#define MAX_ILLUM_LEVEL	15

#define MIN_LIGHT_LEVEL	(0.0)
#define MAX_LIGHT_LEVEL (1.0)
#define MIN_VAMPL_LEVEL	(0.0)
#define MAX_VAMPL_LEVEL (1.0)

#define MIN_BMODE	0
#define MAX_BMODE	7
#define N_BMODES	4

// Status of Light:

#define BUTT_LOCK	0x01
#define BMODE		0x02
#define BPLUS		0x04
#define BMINUS		0x08
#define LIGHT_ON	0x10
#define ILLUM_ON	0x20
#define VARY_ON		0x40
#define AUTO		0x80

#define LIGHT_WAIT_INTERVAL			30000	// 30 sec
#define SHORT_LIGHT_WAIT_INTERVAL	5000	// 5 sec

#ifdef __COUNTER_BUTTON__
class Light : public HardModule {
#else
class Light : virtual public HardModule {
#endif

private:
	uchar _setIllum;
	uchar _getIllum;
	uchar _setLight;
	uchar _getLight;
	uchar _setVAmpl;
	uchar _getVAmpl;
#ifdef __COUNTER_BUTTON__
	uchar _setBMode;
	uchar _getBMode;
	uchar _setParam;
	uchar _getParam;
	uchar _buttLock;
	uchar _buttUnLock;
#endif
	uchar _illumOn;
	uchar _illumOff;
	uchar _lightOn;
	uchar _lightOff;
	uchar _varyOn;
	uchar _varyOff;

	int _illumValue;		// 0..255 for device
	int	_illumLevel;		// 0..15 (logarithmic scale) for User
	int _lightValue;		// 0..255 for device
	int _vAmplValue;		// 0..255 for device
#ifdef __COUNTER_BUTTON__
	int _bMode;				// 0..7
#endif

	int _illumination;	// on/off
	int _light;			// on/off
	int _vary;			// on/off
#ifdef __COUNTER_BUTTON__
	int _buttons;		// lock/unlock
#endif

public:
	Light(void);
	virtual ~Light();

	virtual void resetVariables(void);
	virtual int loadOwnConst(void);
	virtual int loadConst(void);

#ifdef __COUNTER_BUTTON__
	int testButtons(int dialog=1);
#endif
	int testIllum(int dialog=1);
	int turnIllum(int state,			// ON or OFF
				int dialog=1);
	int turnIllum(const char* state,	// "ON" or "OFF"
				int dialog=1);
	int turnLight(int state,			// ON or OFF
				int dialog=1, int waitTime=LIGHT_WAIT_INTERVAL);
	int turnLight(const char* state,	// "ON" or "OFF"
				int dialog=1, int waitInterval=LIGHT_WAIT_INTERVAL);
#ifdef __COUNTER_BUTTON__
	int turnButtons(int state,			// ON or OFF
				int dialog=1);
	int turnButtons(const char* state,// "ON" or "OFF"
				int dialog=1);
#endif
	int turnVary(int state,				// ON or OFF
				int dialog=1);

	int setBrightness(int illumLevel,
				int dialog=1);			//[0..15]
	int setBrightnessUp(int dialog=1);
	int setBrightnessDown(int dialog=1);
	int setLight(double lightLevel,		//[0.0..1.0]
				int dialog=1, int waitTime=LIGHT_WAIT_INTERVAL);
	int setVAmpl(double vAmpl,			//[0.0..1.0]
				int dialog=1);
#ifdef __COUNTER_BUTTON__
	int getBMode(uchar& bMode,			//[0..3]
				int dialog=1);
	int setBMode(uchar bMode,
				int dialog=1);
	int getBParam(uchar& bParam,		//[0..255]
				int dialog=1);
	int setBParam(uchar bParam,
				int dialog=1);
#endif

	int isIllumOn(void) {return _illumination;}
	int isLightOn(void) {return _light;}
#ifdef __COUNTER_BUTTON__
	int isButtonsOn(void) {return _buttons;}
#endif
	int getCurrentBrightness(void) {return _illumLevel;}
	double getLight(void) {return double(_lightValue)/256.0;}
};

#endif
