/*--------------------------------------------------------------------------*/
/** @file motor.h
@brief  Implementation of interaction (command calls and replies from modules, 
their current data and status keeping, etc.) with stepper motor of the aperture 
wheel.

                  MASS project: TURBINA module file
                       header file for motor.cpp



 @author O. Voziakova, Sternberg Institute (ovoz@sai.msu.ru)
 
 @version 2.0
*/
#ifndef MOTOR_H
#define MOTOR_H

#include <qstring.h>
#include "module.h"

#define	MIN_STEP	1
#define	MAX_STEP	8
#define MAX_N_STOPS	8
#define MAX_MOTOR_TEST_INTERVALS	500	//  2.5 sec if dev_wait_int = 5ms

#define MAX_POSITION		3199
#define SYNCH_STEP			1
#define MAX_SYNCH_SHIFT		127
#define MIN_SCAN_POSITION	(-64)
#define MAX_SCAN_POSITION	64

// Status of motor:

#define	DOOR_CLOSE	0x04
#define	ZERO_POINT	0x08
#define	ACCEL		0x10
#define	FORWARD		0x20
#define	SYNCHRON	0x40
#define	MOTION		0x80

// Status of sensors:

#define	ZERO	0x02
#define	DOOR	0x04

class Motor : virtual public HardModule {

private:

	uchar _getSensors;
	uchar _getPosition;
	uchar _moveTo;
	uchar _shiftAt;
	uchar _synchAt;
	uchar _loadStops;
	uchar _getStops;
	uchar _setAperture;
	uchar _getAperture;
	uchar _setStep;
	uchar _getStep;
	uchar _nullSet;
	uchar _openDoor;
	uchar _closeDoor;
	uchar _powerDown;

	int _currentDoor;
	int	_currentAperture;
	uchar	_stop[MAX_N_STOPS];
	char	_stopCorrect[MAX_N_STOPS];
	char	_stopCorrect_new[MAX_N_STOPS];
	QString	_stopName[MAX_N_STOPS];
	double	_stopFoclen[MAX_N_STOPS];
	int _nStops;

public:

	Motor(void);
	virtual ~Motor();

	virtual void resetVariables(void);
	virtual int loadOwnConst(void);
	virtual int loadConst(void);

	int getSensors(uchar& status, int dialog=1);
	int getPosition(int& position, int dialog=1);
	int setStep(uchar step, int dialog=1);
	int loadStops(int dialog=1);
	int getApertureNumber(const QString& apName, int dialog=1);
	double getFoclen(const QString& apName, int dialog=1);
	double getFoclen(int aperture, int dialog=1);
	int waitMotor(int dialog=1);
	int setStop_(uchar absStopNumber, int dialog=1);
	int setAperture(const QString& apName, int dialog=1, bool hard=false);
	int setAperture(int aperture, int dialog=1, bool hard=false);
	int setStop(uchar absStopNumber, int dialog=1);
	int moveTo(int position, int dialog=1);
	int synchAt(char shift, int dialog=1);
	int shiftAt(char shift, int dialog=1);
	int turnDoor(const char* state, int dialog=1);
	int turnDoor(int state, int dialog=1);
	int testDoor(int dialog=1);
	int testInitStatus(int dialog=1);
	int nullSet(int dialog=1);
	int movePlus(int dialog=1);
	int moveMinus(int dialog=1);
	int shiftAperture(int shift, int dialog=1);
	int isDoorOpen(void) {return(_currentDoor);}
	int currentAperture(void) {return(_currentAperture);}
	int getNStops(void) {return(_nStops);}
	int getMaxNStops(void) {return(MAX_N_STOPS);}
	const QString& getStopName(int stopNumber);
	int getStopCorrect(void);
	int getStopCorrectNew(void);
	void startCorrection(void);
	void acceptCorrection(int dialog=1);
};

#endif
