#define QT_THREAD_SUPPORT
#include <qthread.h>
#include <qapplication.h>
#include <qfile.h>
#include <qfiledialog.h>
#include <qtextstream.h>
#include <qtooltip.h>
#include <qstatusbar.h>
#include <qmenubar.h>
#include <qmessagebox.h>
#include <math.h>

#include "mywidget.h"
#include "tree.h"
#include "trbexport.h"
#include "trbtime.h"
#include "trbthread.h"
#include "scenariothr.h"
#include "messages.h"
#include "tools.h"
#include "cfg.h"
#include "device.h"
#include "star.h"
#include "names.h"
#if defined __MOTOR__ || defined __SCANMOTOR__
	#include "diaphragm.xpm"
#endif
#include "sv.xpm"
#include "buttons.xpm"

extern DevManager	devmanager;
extern Star		star;
extern CFGTree*		tree;
extern QMutex		infoExportMutex;
extern bool		isMassRoot;
extern bool		withDevice;

static const char* Undefin	= "?";
#ifdef __MOTOR__
static const char* Open		= "Open";
static const char* Close	= "Close";
#endif
static const char* On		= "On";
static const char* Off		= "Off";

// ToolTips:
#if defined __MOTOR__ || defined __SCANMOTOR__
static const char* diaphragmText	= "Star in diaphragm";
#endif
static const char* svButtonText		= "Super Visor";
static const char* raPlusText		= "RA+";
static const char* raMinusText		= "RA-";
static const char* decPlusText		= "Dec+";
static const char* decMinusText		= "Dec-";
static const char* stopText 		= "Stop Scenario";
static const char* pauseText		= "Pause";
static const char* playText			= "Press to run scenario";
static const char* navigateText		= "Press to show star info";
static const char* selectText		= "Press to select new star";  
static const char* illumText		= "Illumination";
#ifdef __MOTOR__
static const char* doorText			= "Input Door";
static const char* apertText		= "Current Aperture";
#endif
#ifndef __COUNTER_BUTTON__
static const char* mirrorText		= "Viewer mirror";
#endif
static const char* hvText		= "High Voltage";
static const char* tmprText		= "Temperature, C";
static const char* starText		= "Observed Star";
static const char* zenitText		= "Air Mass";
static const char* scenarioText		= "Current Scenario";
static const char* modeText		= "Current Mode";
static const char* modeTimeText		= "Mode Remained Time";
static const char* scenarioTimeText	= "Scenario Remained Time";
static const char* clockText		= "Local Time";

void MyWidget::setLabelText(QLabel *label, const QString& text) {
	if(label->text() != text) {
		int width = fontMetrics().width(text);
		label->setText(text);
		if(label->width() < width) {
			label->setFixedSize(label->sizeHint());
		}
	}
}

int MyWidget::addItem(QMenuData* menuData,
				const char* helpText,
				bool withDeviceEnable,
				bool forRootEnable,
				bool countDisable,
				bool withSuperVisorDisable) {
	int ret = item_id;
	if(menu_help.size() <= (uint)item_id) menu_help.resize(menu_help.size()+10); 
	menu_help[item_id] = new QString(helpText);
	if(withDeviceEnable) {
		if(device_enable.size() <= (uint)n_dev) 
				device_enable.resize(device_enable.size()+10);
		device_enable[n_dev] = item_id;
		n_dev++;
	}
	bool deviceEnable = !withDeviceEnable || withDevice;
	bool rootEnable = !forRootEnable || isMassRoot;
	bool enable = deviceEnable && rootEnable;
	menuData->setItemEnabled(item_id,enable);
	if(countDisable && enable) {
		if(disabl_count.size() <= (uint)n_dis) 
				disabl_count.resize(disabl_count.size()+10);
		disabl_count[n_dis] = item_id;
		n_dis++;
	}
	if(withSuperVisorDisable && enable) {
		if(disabl_sv.size() <= (uint)n_dis_sv) 
				disabl_sv.resize(disabl_sv.size()+10);
		disabl_sv[n_dis_sv] = item_id;
		n_dis_sv++;
	}
	item_id++;
	return ret;
}


MyWidget::MyWidget(QWidget *parent, const char *name)
					: QMainWindow(parent, name)	{
	isTimerEventLocked = false;
	isMyQuitLocked = false;
	isScenario = false;
	oldStarPosition = -1;
	existTalkDialog = false;
	existAdjustDialog = false;
	existIllumDialog = false;
	existLightDialog = false;
	existHVDialog = false;
	existSelectStarDialog = false;
	existShowStarDialog = false;
	existCommentDialog = false;
	existCenteringDialog = false;
	existBackCorrectionDialog = false;
	existCenterCorrectionDialog = false;
	existTrackCalibrationDialog = false;
	existHelpView = false;
	_updateHA = true;
//	_disableKeyEvent = false;
	toolBarDevice = NULL;
	toolBarObject = NULL;
	toolBarProgram = NULL;
	toolBarButtons = NULL;
	mainView = NULL;
	progressor = NULL;
	playButtonGroup = NULL;
	file = NULL;
	measure = NULL;
	tests = NULL;
	help = NULL;
	config = NULL;
	tools = NULL;
	functions = NULL;
#ifdef __MOTOR__
	aperture_set = NULL;
#endif
	scenario_set = NULL;
	cnfg = NULL;
	

	mainView = new QJKMainView(this,"MainView");
	CHECK_PTR(mainView);

	loadConst();

	if(mainView) mainView->setFont(QFont("Helvetica",_fontSizeSmall),true);
	if(mainView) mainView->fitToFont();
	if(mainView) setCentralWidget(mainView);

/*
	progressor = new QProgressBar(100,statusBar());
	CHECK_PTR(progressor);
	statusBar()->addWidget(progressor,0,true);
	progressor->hide();
*/
	toolBarDevice = new QToolBar(this);
	CHECK_PTR(toolBarDevice);
	addToolBar(toolBarDevice,"Device status",Bottom,true);
	toolBarDevice->setHorizontalStretchable(true);

	toolBarObject = new QToolBar(this);
	CHECK_PTR(toolBarObject);
	addToolBar(toolBarObject,"Object status",Top,true);
	toolBarObject->setHorizontalStretchable(true);

	toolBarProgram = new QToolBar(this);
	CHECK_PTR(toolBarProgram);
	addToolBar(toolBarProgram,"Program status",Top,true);
	toolBarProgram->setHorizontalStretchable(true);

	toolBarButtons = new QToolBar(this);
	CHECK_PTR(toolBarButtons);
	addToolBar(toolBarButtons,"",Left,true);
	toolBarButtons->setVerticalStretchable(true);

#ifdef __MOTOR__
	if(toolBarDevice) labelApert = new QLabel("",toolBarDevice);
	CHECK_PTR(labelApert);
	labelApert->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelApert->setAlignment(AlignCenter);
	labelApert->setFixedSize(labelApert->sizeHint());
	QToolTip::add(labelApert,apertText);

	if(toolBarDevice) labelDoor = new QLabel("",toolBarDevice);
	CHECK_PTR(labelDoor);
	labelDoor->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelDoor->setAlignment(AlignCenter);
	labelDoor->setFixedSize(labelDoor->sizeHint());
	QToolTip::add(labelDoor,doorText);
#endif
#ifndef __COUNTER_BUTTON__
	if(toolBarDevice) labelMirror = new QLabel("",toolBarDevice);
	CHECK_PTR(labelMirror);
	labelMirror->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelMirror->setAlignment(AlignCenter);
	labelMirror->setFixedSize(labelMirror->sizeHint());
	QToolTip::add(labelMirror,mirrorText);
#endif

	if(toolBarDevice) labelIllum = new QLabel("",toolBarDevice);
	CHECK_PTR(labelIllum);
	labelIllum->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelIllum->setAlignment(AlignCenter);
	labelIllum->setFixedSize(labelIllum->sizeHint());
	QToolTip::add(labelIllum,illumText);

	if(toolBarDevice) labelHV = new QLabel("",toolBarDevice);
	CHECK_PTR(labelHV);
	labelHV->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelHV->setAlignment(AlignCenter);
	labelHV->setFixedSize(labelHV->sizeHint());
	QToolTip::add(labelHV,hvText);

	if(toolBarDevice) labelTmpr = new QLabel("-99.9",toolBarDevice);
	CHECK_PTR(labelTmpr);
	labelTmpr->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelTmpr->setAlignment(AlignCenter);
	labelTmpr->setFixedSize(labelTmpr->sizeHint());
	QToolTip::add(labelTmpr,tmprText);
	if(toolBarDevice) toolBarDevice->setStretchableWidget(labelTmpr);

	if(toolBarDevice) toolBarDevice->addSeparator();

#if defined __MOTOR__ || defined __SCANMOTOR__
	if(toolBarObject) labelDiaphragm = new QLabel(toolBarObject);
	CHECK_PTR(labelDiaphragm);
	labelDiaphragm->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelDiaphragm->setAlignment(AlignCenter);
	QToolTip::add(labelDiaphragm, diaphragmText);
	displayStarPosition(-1);
#endif

	QIconSet svIconSet;
	svIconSet.setPixmap(svOn,QIconSet::Automatic,QIconSet::Normal);
	svIconSet.setPixmap(svOff,QIconSet::Automatic,QIconSet::Disabled);

	if(toolBarObject) svBigRedButton = new QToolButton(svIconSet, svButtonText,"",
		this, SLOT(toggleSVbigRedButton()), toolBarObject, "SVButton");
	CHECK_PTR(svBigRedButton);

	svBigRedButton->setUsesBigPixmap(true);
	svBigRedButton->setAutoRaise(false);
	svBigRedButton->setToggleButton(true);
	svBigRedButton->setOn(true); 
	svBigRedButton->setDisabled(true);

	if(toolBarObject) labelStar = new QLabel("",toolBarObject);
	CHECK_PTR(labelStar);
	labelStar->setText(star.getName());
	labelStar->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelStar->setAlignment(AlignCenter);
	labelStar->setFixedSize(labelStar->sizeHint());
	QToolTip::add(labelStar,starText);
	if(toolBarObject) toolBarObject->setStretchableWidget(labelStar);

	if(toolBarObject) labelZenit = new QLabel("99.99",toolBarObject);
	CHECK_PTR(labelZenit);
	labelZenit->setText(star.getAirMass());
	labelZenit->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelZenit->setAlignment(AlignCenter);
	labelZenit->setFixedSize(labelZenit->sizeHint());
	QToolTip::add(labelZenit,zenitText);

	if(toolBarObject) digitalclock = new DigitalClock(toolBarObject);
	CHECK_PTR(digitalclock);
	digitalclock->setFrameStyle(QFrame::Panel | QFrame::Sunken);
//	digitalclock->setFixedSize(digitalclock->sizeHint());
// Konstructor stavit size
	QToolTip::add(digitalclock,clockText);
	if(toolBarObject) toolBarObject->addSeparator();

	if(toolBarProgram) labelMode = new QLabel("",toolBarProgram);
	CHECK_PTR(labelMode);
	labelMode->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelMode->setAlignment(AlignCenter);
	labelMode->setFixedSize(labelMode->sizeHint());
	QToolTip::add(labelMode,modeText);

	if(toolBarProgram) labelModeTime = new QLabel("",toolBarProgram);
	CHECK_PTR(labelModeTime);
	labelModeTime->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelModeTime->setAlignment(AlignCenter);
	labelModeTime->setFixedSize(labelModeTime->sizeHint());
	QToolTip::add(labelModeTime,modeTimeText);

	if(toolBarProgram) labelScenario = new QLabel(devmanager.getScenarioName(
				devmanager.getScenarioNumber()),
				toolBarProgram);
	CHECK_PTR(labelScenario);
	labelScenario->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelScenario->setAlignment(AlignCenter);
	labelScenario->setFixedSize(labelScenario->sizeHint());
	QToolTip::add(labelScenario,scenarioText);	

	if(toolBarProgram) labelScenarioTime = new QLabel("",toolBarProgram);
	CHECK_PTR(labelScenarioTime);
	labelScenarioTime->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	labelScenarioTime->setAlignment(AlignCenter);
	labelScenarioTime->setFixedSize(labelScenarioTime->sizeHint());
	QToolTip::add(labelScenarioTime,scenarioTimeText);

	if(toolBarProgram) toolBarProgram->setStretchableWidget(labelScenario);

	QIconSet navigateIconSet,selectStarIconSet,playIconSet;
	QIconSet stopIconSet,pauseIconSet;
	QIconSet raPlusIconSet,raMinusIconSet,decPlusIconSet,decMinusIconSet;
	navigateIconSet.setPixmap(navigate,QIconSet::Automatic,QIconSet::Normal);
	navigateIconSet.setPixmap(navigateDisabled,QIconSet::Automatic,QIconSet::Disabled);

	selectStarIconSet.setPixmap(selectStar,QIconSet::Automatic,QIconSet::Normal);
	selectStarIconSet.setPixmap(selectStarDisabled,QIconSet::Automatic,QIconSet::Disabled);

	playIconSet.setPixmap(play,QIconSet::Automatic,QIconSet::Normal);
	playIconSet.setPixmap(playDisabled,QIconSet::Automatic,QIconSet::Disabled);

	stopIconSet.setPixmap(setStop,QIconSet::Automatic,QIconSet::Normal);
	stopIconSet.setPixmap(setStopDisabled,QIconSet::Automatic,QIconSet::Disabled);

	pauseIconSet.setPixmap(setPause,QIconSet::Automatic,QIconSet::Normal);
	pauseIconSet.setPixmap(setPauseDisabled,QIconSet::Automatic,QIconSet::Disabled);

	raPlusIconSet.setPixmap(raPlusIcon,QIconSet::Automatic,QIconSet::Normal);
	raPlusIconSet.setPixmap(raPlusDisabled,QIconSet::Automatic,QIconSet::Disabled);

	raMinusIconSet.setPixmap(raMinusIcon,QIconSet::Automatic,QIconSet::Normal);
	raMinusIconSet.setPixmap(raMinusDisabled,QIconSet::Automatic,QIconSet::Disabled);

	decPlusIconSet.setPixmap(decPlusIcon,QIconSet::Automatic,QIconSet::Normal);
	decPlusIconSet.setPixmap(decPlusDisabled,QIconSet::Automatic,QIconSet::Disabled);

	decMinusIconSet.setPixmap(decMinusIcon,QIconSet::Automatic,QIconSet::Normal);
	decMinusIconSet.setPixmap(decMinusDisabled,QIconSet::Automatic,QIconSet::Disabled);

	if(toolBarButtons) navigateButton = new QToolButton(navigateIconSet,navigateText,"",
		this,SLOT(slotShowStarDialog()),toolBarButtons,"NavigateButton");
	CHECK_PTR(navigateButton);
	navigateButton->setUsesBigPixmap(true);
	navigateButton->setAutoRaise(false);

	if(toolBarButtons) selectButton = new QToolButton(selectStarIconSet,selectText,"",
		this,SLOT(slotSelectStarDialog()),toolBarButtons,"SelectButton");
	CHECK_PTR(selectButton);
	QToolTip::add(selectButton,selectText);
	selectButton->setUsesBigPixmap(true);
	selectButton->setAutoRaise(false);

	playButtonGroup = new QButtonGroup(this);
	CHECK_PTR(playButtonGroup);
	playButtonGroup->setExclusive(true);

	if(toolBarButtons) playButton = new QToolButton(playIconSet, playText,"",
		this, SLOT(playScenario()),toolBarButtons,"PlayButton");
	CHECK_PTR(playButton);
	if(playButtonGroup) playButtonGroup->insert(playButton,1);
	playButton->setToggleButton(true);
	playButton->setUsesBigPixmap(true);
	playButton->setAutoRaise(false);
	playButton->setDisabled(!withDevice);

	if(toolBarButtons) stopButton = new QToolButton(stopIconSet, stopText, "",
		this, SLOT(stopScenario()), toolBarButtons,"StopButton");
	CHECK_PTR(stopButton);
	if(playButtonGroup) playButtonGroup->insert(stopButton,2);
	stopButton->setToggleButton(true);
	stopButton->setUsesBigPixmap(true);
	stopButton->setAutoRaise(false);
	stopButton->setDisabled(true);

	if(toolBarButtons) pauseButton = new QToolButton(pauseIconSet, pauseText,"",
		this, SLOT(togglePause()), toolBarButtons ,"PauseButton");
	CHECK_PTR(pauseButton);
	pauseButton->setUsesBigPixmap(true);
	pauseButton->setAutoRaise(false);
	pauseButton->setToggleButton(true);
	pauseButton->setDisabled(true);

	if(toolBarButtons) labelTrack = new QLabel("",toolBarButtons);
	CHECK_PTR(labelTrack);
	if(toolBarButtons) toolBarButtons->setStretchableWidget(labelTrack);

	if(toolBarButtons) raPlus = new QToolButton(toolBarButtons,"RAplusButton");
	CHECK_PTR(raPlus);
	raPlus->setIconSet(raPlusIconSet);
	raPlus->setTextLabel(raPlusText);
	raPlus->setUsesBigPixmap(true);
	raPlus->setAutoRaise(false);
	if(!devmanager.withAutoGuide()) raPlus->setDisabled(true);
	connect(raPlus, SIGNAL(pressed()), this, SLOT(moveRAplus()));
	connect(raPlus, SIGNAL(released()),
			this, SLOT(stopTrack()));

	if(toolBarButtons) raMinus = new QToolButton(toolBarButtons,"RAplusButton");
	CHECK_PTR(raMinus);
	raMinus->setIconSet(raMinusIconSet);
	raMinus->setTextLabel(raMinusText);
	raMinus->setUsesBigPixmap(true);
	raMinus->setAutoRaise(false);
	if(!devmanager.withAutoGuide()) raMinus->setDisabled(true);
	connect(raMinus, SIGNAL(pressed()), this, SLOT(moveRAminus()));
	connect(raMinus, SIGNAL(released()),
			this, SLOT(stopTrack()));

	if(toolBarButtons) decPlus = new QToolButton(toolBarButtons,"DecPlusButton");
	CHECK_PTR(decPlus);
	decPlus->setIconSet(decPlusIconSet);
	decPlus->setTextLabel(decPlusText);
	decPlus->setUsesBigPixmap(true);
	decPlus->setAutoRaise(false);
	if(!devmanager.withAutoGuide()) decPlus->setDisabled(true);
	connect(decPlus, SIGNAL(pressed()), this, SLOT(moveDecPlus()));
	connect(decPlus, SIGNAL(released()),
			this, SLOT(stopTrack()));

	if(toolBarButtons) decMinus = new QToolButton(toolBarButtons,"DecMinusButton");
	CHECK_PTR(decMinus);
	decMinus->setIconSet(decMinusIconSet);
	decMinus->setTextLabel(decMinusText);
	decMinus->setUsesBigPixmap(true);
	decMinus->setAutoRaise(false);
	if(!devmanager.withAutoGuide()) decMinus->setDisabled(true);
	connect(decMinus, SIGNAL(pressed()), this, SLOT(moveDecMinus()));
	connect(decMinus, SIGNAL(released()),
			this, SLOT(stopTrack()));

	if(toolBarProgram) toolBarProgram->addSeparator();

	item_id = 0;
	menu_help.resize(100);
	n_dis = 0;
	disabl_count.resize(100);
	n_dis_sv = 0;
	disabl_sv.resize(100);
	n_dev = 0;
	device_enable.resize(100);

	file = new QPopupMenu;
	CHECK_PTR(file);

	if(file) file->insertItem("Comment...",
		this,SLOT(slotCommentDialog()),ALT+Key_C,item_id);
	if(file) addItem(file,"Comment",false,false,false,false);

	if(file) file->insertItem("Select star...",
		this,SLOT(slotSelectStarDialog()),0,item_id);
	if(file) addItem(file,"Select star",false,false,false,false);

	if(file) file->insertItem("Show star...",
		this,SLOT(slotShowStarDialog()),0,item_id);
	if(file) addItem(file,"Show star",false,false,false,false);

#ifdef __WITH_PLAYBACK__
	if(file) file->insertItem("Play Back", this, SLOT(slotPlayBack()),0,item_id);
	if(file) pb_id = addItem(file,"Play Back",false,false,false);
#endif

	if(file) file->insertSeparator();

	if(file) file->insertItem("Exit", this, SLOT(myQuit()),CTRL+Key_Q,item_id);
	if(file) addItem(file,"Exit",false,false,true);

	help = new QPopupMenu;
	CHECK_PTR(help);
	help->insertItem("Help", this, SLOT(slotHelp()),CTRL+Key_H,item_id);
	addItem(help,"Help",false,false,false,false);

	measure = new QPopupMenu;
	CHECK_PTR(measure);

	if(measure) {
#ifdef __MOTOR__
		measure->insertItem("Normal", this, SLOT(normal()),
						0, item_id);
		addItem(measure,"Normal");
		measure->insertItem("Generalized", this, SLOT(generalized()),
						0, item_id);
		addItem(measure,"Generalized");
#endif
#if defined __MOTOR__ || defined __SCANMOTOR__
		measure->insertItem("Center control", this,
				SLOT(centerControl()), 0, item_id);
		addItem(measure,"Star centering control");
#endif
#ifdef __MOTOR__
		measure->insertItem("Focus control",
				this,SLOT(focusControl()),
						0, item_id);
		addItem(measure,"Focus control");
#else
		measure->insertItem("Run measurement", this, SLOT(normal()),
						0, item_id);
		addItem(measure,"Run measurement");
#endif
		measure->insertItem("Background",
			this,SLOT(backgroundMeasurement()),
							0, item_id);
		addItem(measure,"Background measurement");
		measure->insertItem("Flux Estimation",this,SLOT(fluxEstimation()),
						CTRL+Key_F,item_id);
		addItem(measure,"Flux Estimation");
		measure->insertSeparator();

		scenario_set = new QPopupMenu;
		CHECK_PTR(scenario_set);
		measure->insertItem("Select Scenario",scenario_set,
				item_id);
		addItem(measure,"Select Scenario");
	}
	int nscen = devmanager.getMaxUserScenarioNumber();
	scen_id.resize(nscen);
	for(int i=0; i<nscen; i++) {
		scenario_set->insertItem(devmanager.getScenarioName(i),
				item_id);
		scen_id[i] = addItem(scenario_set,
				devmanager.getScenarioName(i).data(),
				true,false,false);
	}
	connect(scenario_set, SIGNAL(activated(int)),
		this, SLOT(slotSetScenario(int)));

	if(measure) measure->insertSeparator();

	if(measure) measure->insertItem("Run scenario",
				this,SLOT(playScenario()),0,item_id);
	if(measure) play_item = addItem(measure,playText,true,false,false);

	if(measure) measure->insertItem("Stop scenario",
				this,SLOT(stopScenario()),0,item_id);
	if(measure) stop_item = addItem(measure,stopText,true,false,false);
	if(measure) measure->setItemEnabled(stop_item, false);

	if(measure) measure->insertItem("Pause scenario",
				this,SLOT(togglePauseMenu()),0,item_id);
	if(measure) pause_item = addItem(measure,pauseText,true,false,false);
	if(measure) measure->setItemEnabled(pause_item, false);

 	tests = new QPopupMenu;
	CHECK_PTR(tests);

	tests->insertItem("Exchange Test",this,SLOT(exchangeTest()),
						0,item_id);
	addItem(tests,"Run full exchange test");

	tests->insertItem("Detectors Test",this,SLOT(detectorTest()),
						0,item_id);
	addItem(tests,"Run detectors test");

	tests->insertItem("Statistic Test",this,SLOT(statisticTest()),
						0,item_id);
	addItem(tests,"Run statistic test");

	tools = new QPopupMenu;
	CHECK_PTR(tools);
//VK
	tools->insertItem("Initialization", &devmanager,
				SLOT(gui_init()),0,item_id);
	int id = addItem(tools,"Initialization");

	tools->insertItem("Detectors Counting measurement",this,SLOT(detectorCountingMeasurement()),
						0,item_id);
	addItem(tools,"Detectors Counting measurement...");

	tools->insertItem("Detectors Statistic measurement",this,SLOT(detectorStatisticMeasurement()),
						0,item_id);
	addItem(tools,"Detectors Statistic measurement...");

#ifdef __MOTOR__
	tools->insertItem("Centering...",
		this,SLOT(slotCenteringDialog()),0,item_id);
	addItem(tools,"Star centering");

	tools->insertItem("Track calibration...",
		this, SLOT(slotTrackCalibrationDialog()),0,item_id);
	id = addItem(tools,"Track calibration");
	if(!devmanager.withAutoGuide()) tools->setItemEnabled(id,false);

	tools->insertItem("Background correction...",
		this,SLOT(slotBackCorrectionDialog()),
						0, item_id);
	addItem(tools,"Background correction");

	tools->insertItem("Aperture Center...",
			this,SLOT(slotCenterCorrectionDialog()),
							0,item_id);
	addItem(tools,"Fixation of aperture center");

	tools->insertItem("Apertures Adjust...",
		this,SLOT(slotAdjustDialog()),0,item_id);
	addItem(tools,"Apertures Adjust");
#endif

	tools->insertSeparator();

	tools->insertItem("Talk",this,SLOT(slotTalkDialog()),0,item_id);
	id = addItem(tools,"Talk (debug mode)",true,true,false);

	tools->insertItem("Break Accumulation",&devmanager,SLOT(stopCounting()),
						CTRL+Key_S,item_id);
	break_item = addItem(tools,"Break Accumulation",true,false,false);
	tools->setItemEnabled(break_item, false);

	tools->insertItem("Parking", &devmanager,
			SLOT(park()),0,item_id);
	addItem(tools,"Parking");

	functions = new QPopupMenu;
	CHECK_PTR(functions);

#ifdef __MOTOR__
	functions->insertItem("",
		this, SLOT(switchDoor()),0,item_id);
	door_id = addItem(functions,"Door Open/Close");
	slotTurnDoor(-1);
#endif

	functions->insertItem("",
			this,SLOT(switchHV()),0,item_id);
	hv_id = addItem(functions,"HV On/Off");
	slotTurnHV(-1);

	functions->insertItem("",
		this, SLOT(turnSafety()),0,item_id);
	safety_id = addItem(functions,"OverLight safety On/Off",true,true);
	slotTurnSafety(-1);

#ifdef __COUNTER_BUTTON__
	functions->insertItem("",
		this,SLOT(switchButtons()),0,item_id);
	buttons_id = addItem(functions,"Buttons On/Off");
	slotTurnButtons(-1);

	functions->insertItem("",
		this, SLOT(turnIllum()),0,item_id);
	illum_id = addItem(functions,"Illumintaion On/Off");
	slotTurnIllum(-1);

#else
	slotTurnMirror(-1);
	slotTurnIllum(-1);
#endif

	functions->insertItem("",
		this, SLOT(turnLight()),0,item_id);
	light_id = addItem(functions,"Control Light On/Off",true,true);
	slotTurnLight(-1);

	functions->insertSeparator();

	functions->insertItem("Set Illumination level...",
		this,SLOT(slotIllumDialog()),0,item_id);
	id = addItem(functions,"Set Illumination level");

	functions->insertItem("Set Light level...",
		this,SLOT(slotLightDialog()),0,item_id);
	id = addItem(functions,"Set Light level",true,true);

	functions->insertItem("Set High Voltage...",
		this,SLOT(slotHVDialog()),0,item_id);
	id = addItem(functions,"Set High Voltage",true,true);

#ifdef __MOTOR__
	aperture_set = new QPopupMenu;
	CHECK_PTR(aperture_set);
	functions->insertItem("Set &Aperture",aperture_set,
				item_id);
	id = addItem(functions,"Set Aperture");

	int ns = devmanager.getNStops();
	apert_id.resize(ns);
	for(int i=0; i<ns; i++) {
		aperture_set->insertItem(devmanager.getStopName(i),
			item_id);
		apert_id[i] = addItem(aperture_set,devmanager.getStopName(i).data());
	}
	slotAperture(devmanager.currentAperture());
	connect(aperture_set, SIGNAL(activated(int)),
		this, SLOT(setAperture(int)));

	functions->insertItem("Set Null-Point",
		&devmanager,SLOT(nullSet()),0,item_id);
	id = addItem(functions,"Set Null-Point");
#endif

	config = new QPopupMenu;
	CHECK_PTR(config);
	cnfg = new CFGDialog(config,&item_id,menu_help);

	if(file) menuBar()->insertItem("&File",file);
	if(measure) menuBar()->insertItem("&Measurement",measure);
	menuBar()->insertItem("T&est",tests);
	menuBar()->insertItem("&Tools",tools);
	menuBar()->insertItem("&Control",functions);
	menuBar()->insertItem("Confi&g",config);
	menuBar()->insertSeparator();
	menuBar()->insertItem("&Help",help);

	connect(menuBar(),SIGNAL(highlighted(int)),
				this,SLOT(slotHighlighted(int)));
	connect(cnfg,SIGNAL(signalShowStatus(const char*,int)),
				this,SLOT(slotShowMessage(const char*,int)));
	connect(tree, SIGNAL(changed()), this,SLOT(change()));


	normalTimer = startTimer(1000);	// 1 sec

	setGeometry(_winX, _winY, _winWidth, _winHeight);
	statusBar()->message("Ready",2000);
	repaint();
	StateExport::setApplicationReady();
}

void MyWidget::myQuit(void) {
	if(isMyQuitLocked) return;
	isMyQuitLocked = true;

	jdebug("mywidget::myQuit...");
//VK	if(cnfg) cnfg->testCfgChang();
	StateExport::cleanApplicationReady();
	trbthread::stop();	// with waiting countThread
	devmanager.closeDevice(true);	// with Park
	trbthread::destruct();
	qApp->quit();
}

void MyWidget::playScenario(void) {
	scenariothr::playScenario();
//	showPlayScenario();
}

void MyWidget::enableSelectScenario(bool enable) {
	for(uint i=0; i<scen_id.size(); i++) {
		menuBar()->setItemEnabled(scen_id[i],enable);
	}
}

void MyWidget::showPlayScenario(void) {
	playButton->setOn(true);
	stopButton->setOn(false);

//	infoExportMutex.lock();		uzhe locked!
	bool withSV = InfoExport::withSV;
//	infoExportMutex.unlock();
	if(withSV) return;
	
	playButton->setDisabled(true);
	menuBar()->setItemEnabled(play_item, false);
	enableSelectScenario(false);

	stopButton->setDisabled(false);

	menuBar()->setItemEnabled(stop_item, true);
	pauseButton->setDisabled(false);
	menuBar()->setItemEnabled(pause_item, true);
}

void MyWidget::stopScenario(void) {
	devmanager.stopScenario();
//	showStopScenario();
}

void MyWidget::showStopScenario(void) {
	playButton->setOn(false);
	stopButton->setOn(true);

//	infoExportMutex.lock();  // uzhe locked!!!!
	bool withSV = InfoExport::withSV;
//	infoExportMutex.unlock();
	if(withSV) return;

	playButton->setDisabled(false);
	menuBar()->setItemEnabled(play_item, true);
	enableSelectScenario(true);

	stopButton->setDisabled(true);

	menuBar()->setItemEnabled(stop_item, false);
	pauseButton->setOn(false);
	pauseButton->setDisabled(true);
	menuBar()->setItemEnabled(pause_item, false);
}

void MyWidget::togglePauseMenu(void) {
	pauseButton->toggle();
	togglePause();
}

void MyWidget::togglePause(void) {
	jdebug("&&&&&&&&&&&&&&&&&&&&&&&&&&&&& toggle Pause!");
	if(pauseButton->isOn()) {
		jdebug("&&&&&&&&&& pauseOn!");
		stopButton->setDisabled(true);
		menuBar()->setItemEnabled(stop_item, false);
		scenariothr::setPause(true);
	} else {
		jdebug("&&&&&&&&&& pauseOff!");
		stopButton->setDisabled(false);
		menuBar()->setItemEnabled(stop_item, true);
		scenariothr::setPause(false);
	}
	jdebug("&&&&&&&&&& end toggle");
}

void MyWidget::toggleSVbigRedButton(void) {
	jdebug("&&&&& toggle svBigRedButton!");
	if(svBigRedButton->isOn()) {
		jdebug("&&&&&&&&&& svOn!");
//		printf("&&&&&&&&&& svOn!\n");
		infoExportMutex.lock();
		InfoExport::withSV = false;
		infoExportMutex.unlock();
		slotEnableSV();
	} else {
		jdebug("&&&&&&&&&& svOff!");
//		printf("&&&&&&&&&& svOff!\n");
		infoExportMutex.lock();
		InfoExport::withSV = true;
		infoExportMutex.unlock();
		slotDisableSV();
	}
//	if(InfoExport::isScenario || devmanager.isCounting()) slotDisableItems();
//	else slotEnableItems();
}

void MyWidget::slotTurnIllum(int _on) {
	jdebug("Illum On/off: %d",_on);
	switch (_on) {
		case  ON : functions->changeItem(illum_id, 
			"Turn Illumintaion\tOff");
			setLabelText(labelIllum, On);
			break;
		case  OFF : functions->changeItem(illum_id,
			"Turn Illumintaion\tOn");
			setLabelText(labelIllum,Off);
			break;
		case -1 : functions->changeItem(illum_id,
			"Turn Illumintaion\tOn/Off");
			setLabelText(labelIllum,Undefin);
			break;
	}
}

void MyWidget::slotTurnSafety(int _on) {
	jdebug("Safety On/off: %d",_on);
	switch (_on) {
		case  ON : functions->changeItem(safety_id, 
			"Turn Safety\tOff");
			break;
		case  OFF : functions->changeItem(safety_id,
			"Turn Safety\tOn");
			break;
		case -1 : functions->changeItem(safety_id,
			"Turn Safety\tOn/Off");
			break;
	}
}

void MyWidget::slotTurnLight(int _on) {
	jdebug("Light On/off: %d",_on);
	switch (_on) {
		case  ON : functions->changeItem(light_id, 
			"Turn Light\tOff");
			break;
		case  OFF : functions->changeItem(light_id,
			"Turn Light\tOn");
			break;
		case -1 : functions->changeItem(light_id,
			"Turn Light\tOn/Off");
			break;
	}
}

void MyWidget::slotTemperature(double _temperature) {
	QString str;
	str.setNum(_temperature,'f',1);
	setLabelText(labelTmpr, str);
}

int MyWidget::moveRAplus(void) {
	int success = SUCCESS;
	if(devmanager.isCounting() || isScenario)
		StateExport::startUserTrack(RA,1);
	else
		if((success=devmanager.track(RA,1)) != SUCCESS)
			raPlus->setDown(false);
	return success;
}

int MyWidget::moveRAminus(void) {
	int success = SUCCESS;
	if(devmanager.isCounting() || isScenario)
		StateExport::startUserTrack(RA,-1);
	else
		if((success=devmanager.track(RA,-1)) != SUCCESS)
			raMinus->setDown(false);
	return success;
}

int MyWidget::moveDecPlus(void) {
	int success = SUCCESS;
	if(devmanager.isCounting() || isScenario)
		StateExport::startUserTrack(DECLINATION,1);
	else
		if((success=devmanager.track(DECLINATION,1)) != SUCCESS)
			decPlus->setDown(false);
	return success;
}

int MyWidget::moveDecMinus(void) {
	int success = SUCCESS;
	if(devmanager.isCounting() || isScenario)
		StateExport::startUserTrack(DECLINATION,-1);
	else
		if((success=devmanager.track(DECLINATION,-1)) != SUCCESS)
			decMinus->setDown(false);
	return success;
}

void MyWidget::stopTrack(void) {
	StateExport::stopUserTrack();
	if(!devmanager.isCounting() && !isScenario)
		devmanager.stopTrack();
}

void MyWidget::turnIllum(void) {
	if(devmanager.isIllumOn()) {
		devmanager.turnIllum(OFF,1);
	} else {
		devmanager.turnIllum(ON,1);
	}
}

void MyWidget::turnSafety(void) {
	if(devmanager.isSafetyOn()) {
		devmanager.turnSafety(OFF,1);
	} else {
		devmanager.turnSafety(ON,1);
	}
}

void MyWidget::turnLight(void) {
	if(devmanager.isLightOn()) {
		devmanager.turnLight(OFF,1);
	} else {
		devmanager.turnLight(ON,1,0);
	}
}

void MyWidget::slotTurnHV(int _on) {
	jdebug("HV On/off: %d",_on);
	switch(_on) {
		case  ON : functions->changeItem(hv_id,
			"Turn HV\tOff");
			setLabelText(labelHV, On);
			break;
		case  OFF : functions->changeItem(hv_id,
			"Turn HV\tOn");
			setLabelText(labelHV, Off);
			break;
		case -1 : functions->changeItem(hv_id,
			"Turn HV\tOn/Off");
			setLabelText(labelHV, Undefin);
			break;
	}
}

void MyWidget::switchHV(void) {
	int state = devmanager.isHVOn(1);
	if(state == ON) {
		devmanager.turnHV(OFF,1);
	} else {
		devmanager.turnHV(ON,1);
	}
}

void MyWidget::slotTurnButtons(int state) {
#ifdef __COUNTER_BUTTON__
	jdebug("Buttons On/off: %d", state);
	switch(state) {
		case  ON : functions->changeItem(buttons_id,
			"Buttons\tLock"); break;
		case  OFF : functions->changeItem(buttons_id,
			"Buttons\tUnLock"); break;
		case UNDEFINED : functions->changeItem(buttons_id,
			"Buttons\tLock/UnLock"); break;
	}
#endif
}

void MyWidget::switchButtons(void) {
#ifdef __COUNTER_BUTTON__
	if(devmanager.isButtonsOn()) {
		devmanager.turnButtons(OFF,1);
	} else {
		devmanager.turnButtons(ON,1);
	}
#endif
}

void MyWidget::slotTurnMirror(int state) {
#ifndef __COUNTER_BUTTON__
	jdebug("Set Mirror On/off: %d", state);
	switch(state) {
		case  ON :
			setLabelText(labelMirror, On);
			break;
		case  OFF :
			setLabelText(labelMirror, Off);
			break;
		case UNDEFINED :
			setLabelText(labelMirror, Undefin);
			break;
	}
#endif
}

void MyWidget::slotTurnDoor(int state) {
#ifdef __MOTOR__
	jdebug("Set Door On/off: %d", state);
	switch(state) {
		case  OPEN : functions->changeItem(door_id,
			"Door\tClose"); 
			setLabelText(labelDoor, Open);
			break;
		case  CLOSE : functions->changeItem(door_id,
			"Door\tOpen");
			setLabelText(labelDoor, Close);
			break;
		case UNDEFINED : functions->changeItem(door_id,
			"Door\tOpen/Close");
			setLabelText(labelDoor, Undefin);
			break;
	}
#endif
}

void MyWidget::switchDoor(void) {
#ifdef __MOTOR__
	if(devmanager.isDoorOpen() == OPEN) {
		devmanager.turnDoor(CLOSE,1);
	} else {
		devmanager.turnDoor(OPEN,1);
	}
#endif
}

void MyWidget::slotSetScenario(int id) {
	for(uint i=0; i<scen_id.size(); i++) {
		if(id == scen_id[i]) {
			jdebug("MyWidget: set Scenario %d",i);
			devmanager.setCurrentScenario(i);
			return;
		}
	}
}

void MyWidget::slotAperture(int stopNumber) {
#ifdef __MOTOR__
	setLabelText(labelApert, devmanager.getStopName(stopNumber));

	for(uint i=0; i<apert_id.size(); i++) {
		menuBar()->setItemEnabled(apert_id[i], true);
	}
	if((stopNumber < 0) || 
		(stopNumber >= (int)apert_id.size())) return;
	menuBar()->setItemEnabled(apert_id[stopNumber], false);
#endif
}

void MyWidget::setAperture(int id) {
#ifdef __MOTOR__
	for(uint i=0; i<apert_id.size(); i++) {
		if(id == apert_id[i]) {
			jdebug("MyWidget: set Aperture %d",i);
			devmanager.setAperture(i,1);
			jdebug("MyWidget::setAperture return");
			return;
		}
	}
#endif
}

void MyWidget::setProgress(const QString& text, double progress) {
	if(!progressor) return;
	if(progress == -1) openProgressor(text);
	else if(progress == -2) closeProgressor();
	else progressor->setProgress((int)progress);
}

void MyWidget::openProgressor(const QString& str) {
	if(!progressor) return;
	statusBar()->message(str);
	setLabelText(labelMode, str);
	progressor->reset();
	if(!progressor->isVisible()) progressor->show();
}

void MyWidget::closeProgressor(void) {
	if(!progressor) return;
	statusBar()->clear();
	if(progressor->isVisible()) progressor->hide();
}

void MyWidget::setMode(const QString& text, double sec) {
	setLabelText(labelMode, text);
	QTime t(0,0,0);
	int isec = int(rint(sec));
	setLabelText(labelModeTime, t.addSecs(isec).toString());
	if(isScenario) {
		t.setHMS(0,0,0);
		setLabelText(labelScenarioTime, t.addSecs(int(rint(scenarioTime))+isec).toString());
	} else setLabelText(labelScenarioTime, "");
}

void MyWidget::setScenario(const QString& text, double sec) {
	setLabelText(labelScenario, text);

	scenarioTime = sec;
//	Time t = sec;
//	setLabelText(labelScenarioTime, t.toString());
}

/*
void MyWidget::resizeEvent(QResizeEvent* e) {

	int _chng = 0;
	if(_winWidth != e->size().width()) {
		_winWidth = e->size().width();
		_chng = 1;
	}
	if(_winHeight != e->size().height()) {
		_winHeight = e->size().height();
		_chng = 1;
	}
}
*/

/*
void MyWidget::moveEvent(QMoveEvent* e) {
	_winX = e->pos().x();
	_winY = e->pos().y();
	jdebug("MyWidget::MoveEvent: %d %d",_winX,_winY);

}
*/

void MyWidget::change(void) {
		jdebug("MyWidget::change");
	bool first = false;
	loadConst(first);
}

void MyWidget::fitToFont(void) {
#ifdef __MOTOR__
	labelApert->setFixedSize(labelApert->sizeHint());
	labelDoor->setFixedSize(labelDoor->sizeHint());
#endif
#ifndef __COUNTER_BUTTON__
	labelMirror->setFixedSize(labelMirror->sizeHint());
#endif
	labelIllum->setFixedSize(labelIllum->sizeHint());
	labelHV->setFixedSize(labelHV->sizeHint());
	labelTmpr->setFixedSize(labelTmpr->sizeHint());
	labelZenit->setFixedSize(labelZenit->sizeHint());
	labelStar->setFixedSize(labelStar->sizeHint());
	digitalclock->fitToFont();

	labelMode->setFixedSize(labelMode->sizeHint());
	labelScenario->setFixedSize(labelScenario->sizeHint());
	labelModeTime->setFixedSize(labelModeTime->sizeHint());
	labelScenarioTime->setFixedSize(labelScenarioTime->sizeHint());
}

void MyWidget::loadConst(bool first) {
		jdebug("MyWidget::loadConst");
	if(isLeafChanged(cfgFile,"Preferences",
			"Text window settings",
			"WindowGeometry")) {
		const char* _geometry;
		leafDataToStr(&_geometry, first, cfgFile,
				"Preferences",
				"Text window settings",
				"WindowGeometry");
		if(!strcasecmp(_geometry,"AUTO")) {
			_winX = 0;
			_winY = 0;
			_winWidth = qApp->desktop()->width()/2;
			_winHeight = qApp->desktop()->height()-50;
		} else {
			int key = sscanf(_geometry,"%d%d%d%d", &_winX, &_winY,
				&_winWidth, &_winHeight);
			if(key != 4) {
				QString warn = 
					"TextWindow: Invalid geometry! \nWill be modified.";
				showStrongMessage(warn);
				_winX = 0;
				_winY = 0;
				_winWidth = qApp->desktop()->width()/2;
				_winHeight = qApp->desktop()->height()-50;
			}
		}
		if(!first) {
			setGeometry(_winX+10, _winY+68,
					_winWidth, _winHeight);
//   Pochemu +10 i +68 NE ZNAYU !!!!!
		}
	}

	if(isLeafChanged(cfgFile, "Preferences",
			"Text window settings",
			"FontSizeLarge")) {
		leafDataToInt(_fontSizeLarge, first, cfgFile, "Preferences",
				"Text window settings",
				"FontSizeLarge");
		qApp->setFont(QFont("Helvetica",_fontSizeLarge),true);
		if(existCommentDialog) commentDialog->fitToFont();
		if(existCenteringDialog) centeringDialog->fitToFont();
		if(existCenterCorrectionDialog) trackCalibrationDialog->fitToFont();
		if(existTrackCalibrationDialog) trackCalibrationDialog->fitToFont();
		if(!first) fitToFont();
	}
	if(isLeafChanged(cfgFile, "Preferences",
			"Text window settings",
			"FontSizeSmall")) {
		leafDataToInt(_fontSizeSmall, first, cfgFile, "Preferences",
				"Text window settings",
				"FontSizeSmall");
		statusBar()->setFont(QFont("Helvetica",_fontSizeSmall),true);
		if(mainView) mainView->setFont(QFont("Helvetica",_fontSizeSmall),true);
		if(mainView) mainView->fitToFont();
	}

	QString str;
	leafDataToStr(str, first, cfgFile, "Operations",
		"Star Pointing", "CoordinateSystem");
	if(str.upper() == "RA") _showHA = 0;
	else _showHA = 1;
}

void MyWidget::slotDisableItems(void) {

	infoExportMutex.lock();
	bool withSV = InfoExport::withSV;
	infoExportMutex.unlock();
	if(withSV) return;

    for(int i=0; i<n_dis; i++) {
		menuBar()->setItemEnabled(disabl_count[i], false);
	}
#ifdef __WITH_PLAYBACK__
	if(!devmanager.isPlayBack()) menuBar()->setItemEnabled(pb_id,false);
#endif
	cnfg->setDisabled();
//	selectButton->setDisabled(true);
//	navigateButton->setDisabled(true);
	playButton->setDisabled(true);
	menuBar()->setItemEnabled(play_item,false);
	enableSelectScenario(false);
	menuBar()->setItemEnabled(break_item, true);
	if(existBackCorrectionDialog) backCorrectionDialog->setButtonsEnabled(false);
	if(existCenterCorrectionDialog) centerCorrectionDialog->setButtonsEnabled(false);
	if(existTrackCalibrationDialog) trackCalibrationDialog->setButtonsEnabled(false);
	_updateHA = false;
}

void MyWidget::slotEnableItems(void) {
	jdebug("slotEnableItems");
	QApplication::beep();

	infoExportMutex.lock();
	bool withSV = InfoExport::withSV;
	infoExportMutex.unlock();
	if(withSV) return;

	for(int i=0; i<n_dis; i++) {
		menuBar()->setItemEnabled(disabl_count[i], true);
	}
#ifdef __WITH_PLAYBACK__
	if(!devmanager.isPlayBack()) menuBar()->setItemEnabled(pb_id,true);
#endif
	cnfg->setEnabled();
//	selectButton->setDisabled(false);
//	navigateButton->setDisabled(false);
	if(!isScenario) {
		playButton->setDisabled(false);
		menuBar()->setItemEnabled(play_item,true);
		enableSelectScenario(true);
	}
	menuBar()->setItemEnabled(break_item, false);
	if(existBackCorrectionDialog) backCorrectionDialog->setButtonsEnabled(true);
	if(existCenterCorrectionDialog) centerCorrectionDialog->setButtonsEnabled(true);
	if(existTrackCalibrationDialog) trackCalibrationDialog->setButtonsEnabled(true);
	_updateHA = true;
}

void MyWidget::slotDisableSV(void) {
//		printf("slotDisableSV\n");
    for(int i=0; i<n_dis_sv; i++) {
		menuBar()->setItemEnabled(disabl_sv[i], false);
	}
//	cnfg->setDisabled();
//	selectButton->setDisabled(true);
//	navigateButton->setDisabled(true);
	playButton->setDisabled(true);
	stopButton->setDisabled(true);
	pauseButton->setDisabled(true);
	menuBar()->setItemEnabled(play_item,false);
	enableSelectScenario(false);
	menuBar()->setItemEnabled(break_item,false);
//	_updateHA = false;
}

void MyWidget::slotEnableSV(void) {
	jdebug("slot enable SV");
//		printf("slotEnableSV\n");
	QApplication::beep();
	for(int i=0; i<n_dis_sv; i++) {
		menuBar()->setItemEnabled(disabl_sv[i], true);
	}
//	cnfg->setEnabled();
//	selectButton->setDisabled(false);
//	navigateButton->setDisabled(false);
//	playButton->setDisabled(false);
	menuBar()->setItemEnabled(play_item,true);
	enableSelectScenario(true);
	menuBar()->setItemEnabled(break_item, false);
//	_updateHA = true;
	if(InfoExport::isScenario || devmanager.isCounting()) slotDisableItems();
	else slotEnableItems();
	if(InfoExport::isScenario) showPlayScenario();
	else showStopScenario();
}

void MyWidget::disableDevice(bool disable) {
    	for(int i=0; i<n_dev; i++) {
		menuBar()->setItemEnabled(device_enable[i], !disable);
	}
	playButton->setDisabled(disable);
	disableTracker(disable);
}

void MyWidget::disableTracker(bool disable) {
	raPlus->setDisabled(disable);
	raMinus->setDisabled(disable);
	decPlus->setDisabled(disable);
	decMinus->setDisabled(disable);
}

MyWidget::~MyWidget() {
	jdebug("................. MyWidget: destructor...");

	killTimer(normalTimer);

	jdebug("....delete toolBarDevice");
	if(toolBarDevice) delete toolBarDevice;
	jdebug("....deletetoolBarObject ");
	if(toolBarObject) delete toolBarObject;
	jdebug("....delete toolBarProgram");
	if(toolBarProgram) delete toolBarProgram;
	jdebug("....delete toolBarButtons");
	if(toolBarButtons) delete toolBarButtons;

	jdebug("....delete mainView");
	if(mainView) delete mainView;
//	jdebug("....delete progressor");
//	if(progressor) delete progressor;

	jdebug("....delete playButtonGroup");
	if(playButtonGroup) delete playButtonGroup;

	jdebug("....delete file");
	if(file) delete file;
	jdebug("....delete measure");
	if(measure) delete measure;
	jdebug("....delete tests");
	if(tests) delete tests;
	jdebug("....delete help");
	if(help) delete help;
	jdebug("....delete config");
	if(config) delete config;
	jdebug("....delete tools");
	if(tools) delete tools;
	jdebug("....delete functions");
	if(functions) delete functions;
#ifdef __MOTOR__
	jdebug("....delete aperture_set");
	if(aperture_set) delete aperture_set;
#endif
	jdebug("....delete scenario_set");
	if(scenario_set) delete scenario_set;
	jdebug("....delete cnfg");
	if(cnfg) delete cnfg;

	jdebug("....delete menu_help");
	for(int i=0; i<item_id; i++) {
		if(menu_help[i]) delete menu_help[i];
	}
	jdebug("...deleted");
	if(existTalkDialog) {
		jdebug("....delete talkDialog");
		delete talkDialog;
		existTalkDialog = false;
	}
	if(existAdjustDialog) {
		jdebug("....delete adjustDialog");
		delete adjustDialog;
		existAdjustDialog = false;
	}
	if(existSelectStarDialog) {
		jdebug("....delete selectStarDialog");
		delete selectStarDialog;
		existSelectStarDialog = false;
	}
	if(existShowStarDialog) {
		jdebug("....delete showStarDialog");
		delete showStarDialog;
		existShowStarDialog = false;
	}
	if(existCommentDialog) {
		jdebug("....delete CommentDialog");
		delete commentDialog;
		existCommentDialog = false;
	}

	if(existCenteringDialog) {
		jdebug("....delete CenteringDialog");
		delete centeringDialog;
		existCenteringDialog = false;
	}

	if(existTrackCalibrationDialog) {
		jdebug("....delete trackCalibrationDialog");
		delete trackCalibrationDialog;
		existTrackCalibrationDialog = false;
	}

	if(existIllumDialog) {
		jdebug("....delete illumDialog");
		delete illumDialog;
		existIllumDialog = false;
	}
	if(existLightDialog) {
		jdebug("....delete lightDialog");
		delete lightDialog;
		existLightDialog = false;
	}
	if(existHVDialog) {
		jdebug("....delete hvDialog");
		delete hvDialog;
		existHVDialog = false;
	}
	if(existBackCorrectionDialog) {
		jdebug("....delete backCorrectionDialog");
		delete backCorrectionDialog;
		existBackCorrectionDialog = false;
	}
	if(existCenterCorrectionDialog) {
		jdebug("....delete centerCorrectionDialog");
		delete centerCorrectionDialog;
		existCenterCorrectionDialog = false;
	}
	if(existHelpView) {
		jdebug("...delete helpView");
		delete helpView;
		existHelpView = false;
	}

	jdebug("end of ~MyWidget");
}

void MyWidget::slotHighlighted(int id) {
	if(id >= 0) {
		statusBar()->message(menu_help[id]->data(),4000);
	}
}

void MyWidget::slotShowMessage(const char* _text, int ms) {
	if(ms > 0) { statusBar()->message(_text,ms); return; }
	if(ms < 0) { statusBar()->message(_text); return; }
	if(ms == 0) { statusBar()->clear(); return; }
}

void MyWidget::slotShowMessage(const char* _text ) {
	statusBar()->message(_text);
}

void MyWidget::slotHelp() {
	if(!existHelpView) {
		QString fileName = helpFile;
		if(!testFileName(fileName,1,false)) return;
		helpView = new HelpView(fileName,0,"Help");
		CHECK_PTR(helpView);
		existHelpView = true;
	}
	helpView->exec();
}

void MyWidget::slotPlayBack() {
	if(!devmanager.isPlayBack() &&
			!devmanager.isCounting()) { // start PlayBack
		QString pbDirName;
		leafDataToStr(pbDirName, true, cfgFile, "General",
			"Inputs", "PlayBackDir");
		QString fileExt = "*.";
		fileExt += pbInFileExt;
		QString fn = QFileDialog::getOpenFileName(dataPath+pbDirName,
							fileExt,this);
		if(testFileName(fn,0)) {
			if(file) file->changeItem(pb_id, "End Play Back");
			scenariothr::playBack(fn.data());
		}
	} else if(!devmanager.isPlayBack() &&
			devmanager.isCounting()) { // stop PlayBack
		devmanager.endPlayBack();
		if(file) file->changeItem(pb_id, "Play Back");
	} else if(devmanager.isPlayBack() &&
			devmanager.isCounting()) {
		devmanager.stopCounting();
		if(file) file->changeItem(pb_id, "Play Back");
	}
}

void MyWidget::slotTalkDialog(void) {
	if(!existTalkDialog) {
		talkDialog = new TalkDialog();
		CHECK_PTR(talkDialog);
		existTalkDialog = true;
	}

	talkDialog->show();
}

void MyWidget::slotAdjustDialog(void) {
	QString str;
	if(!existAdjustDialog) {
		adjustDialog = new AdjustDialog();
		CHECK_PTR(adjustDialog);
		existAdjustDialog = true;
	}

	int oldApert = devmanager.currentAperture();
	if(oldApert == UNDEFINED) {
		str = "Invalid current aperture";
		switch(showWarning(str, 2, "Ignore", "Cancel")) {
			case  QMessageBox::Abort  : return;
			case  QMessageBox::Ignore : break;
			default	: break;
		}
	}
	int oldIllum = devmanager.isIllumOn();
	devmanager.turnIllum(ON,1);
	devmanager.startAdjustAperture();
	adjustDialog->update();
	int ret = adjustDialog->exec();
	if(ret == QDialog::Accepted) {
		devmanager.acceptAdjustAperture(1);
	}
	devmanager.setAperture(oldApert,1,1);
// !!! Nasil'stvenno stavim dazhe tu zhe aperturu, t.k. mogla byt'
// !!!sdvinuta na shagi!!!	
	devmanager.turnIllum(oldIllum,1);
}

void MyWidget::slotSelectStarDialog(void) {
	QString str;
	if(!existSelectStarDialog) {
		selectStarDialog = new SelectStarDialog();
		CHECK_PTR(selectStarDialog);
		existSelectStarDialog = true;
	} else selectStarDialog->update();

	selectButton->setDisabled(true);
	int ret = selectStarDialog->exec();
	if(ret == QDialog::Accepted) {
		setLabelText(labelStar, star.getName());
		setLabelText(labelZenit, star.getAirMass());
	}
	selectButton->setDisabled(false);
}

void MyWidget::slotCommentDialog(void) {
	if(!existCommentDialog) {
		commentDialog = new CommentDialog();
		CHECK_PTR(commentDialog);
		existCommentDialog = true;
	}
	QString comment;
	commentDialog->update(&comment);

	int ret = commentDialog->exec();
	if(ret == QDialog::Accepted) {
		devmanager.acceptComment(comment);
	}

}

void MyWidget::slotCenteringDialog(void) {
#ifdef __MOTOR__
	if(!existCenteringDialog) {
		centeringDialog = new CenteringDialog();
		CHECK_PTR(centeringDialog);
		existCenteringDialog = true;
	}
	centeringDialog->update();

	centeringDialog->exec();
#endif
}

void MyWidget::slotShowStarDialog(void) {
	int oldAperture = -1;
	if(withDevice) oldAperture = devmanager.startShowStar();
	navigateButton->setDisabled(true);
	if(!existShowStarDialog) {
		showStarDialog = new ShowStarDialog(_showHA);
		CHECK_PTR(showStarDialog);
		existShowStarDialog = true;
	} else showStarDialog->update(_showHA);

	showStarDialog->exec();
	if(withDevice) devmanager.finishShowStar(oldAperture);
	navigateButton->setDisabled(false);
}

void MyWidget::timerEvent(QTimerEvent *e) {
	if(isTimerEventLocked) return;
	isTimerEventLocked = true;

	infoExportMutex.lock();
	if(ExportText::isChanged) {
		ExportText::isChanged = false;
		if(mainView) mainView->showTextLine();
	}
	if(ExportTable::isChanged) {
		ExportTable::isChanged = false;
		if(mainView) mainView->fillTable();
	}
	if(InfoExport::isProgressChanged) {
		InfoExport::isProgressChanged = false;
		setProgress(InfoExport::progress.message,
				InfoExport::progress.progress);
	}

	if(InfoExport::isMessage) {
		InfoExport::isMessage = false;
		slotShowMessage(InfoExport::message.data(),
				InfoExport::ms);
	}

//	if(isScenario != InfoExport::isScenario) { // scenario state is changed
	if(InfoExport::isScenarioStateChanged) { // scenario state is changed
		InfoExport::isScenarioStateChanged = false;
		isScenario = InfoExport::isScenario;
		if(!isScenario) showStopScenario();
		else showPlayScenario();
	}

	if(InfoExport::isScenarioChanged) {
		InfoExport::isScenarioChanged = false;
		setScenario(InfoExport::scenario.message,
				InfoExport::scenario.progress);
	}

	if(InfoExport::isModeChanged) {
		InfoExport::isModeChanged = false;
		setMode(InfoExport::mode.message,
				InfoExport::mode.progress);
	}

	if(InfoExport::isStarPositionChanged) {
		jdebug("star position is changed");
		jdebug("oldPos = %d",oldStarPosition);
		jdebug("newPos = %d",InfoExport::starPosition);
		InfoExport::isStarPositionChanged = false;
		if(oldStarPosition != InfoExport::starPosition) {
			jdebug("oldPos != newPos");
			oldStarPosition = InfoExport::starPosition;
			displayStarPosition(oldStarPosition);
		}
	}

	infoExportMutex.unlock();

	if(StateExport::isChanged()) {
		StateExport::cleanChanged();

		if(StateExport::isStartCounting()) {
			StateExport::cleanStartCounting();
			slotDisableItems();
		}
		if(StateExport::isEndCounting()) {
			StateExport::cleanEndCounting();
			slotEnableItems();
#ifdef __WITH_PLAYBACK__
			if(file) file->changeItem(pb_id, "Play Back");
#endif
			slotShowMessage("End of count");
		}

		if(StateExport::isQuitProgram()) {
			myQuit();
		}

		if(StateExport::isSVconnected()) {
//				printf("Connected!\n");
			StateExport::cleanSVconnected();
			slotDisableSV();
			svBigRedButton->setOn(false);
			svBigRedButton->setDisabled(false);
		}

		if(StateExport::isSVdisconnected()) {
//				printf("Disconnected!\n");
			StateExport::cleanSVdisconnected();
			slotEnableSV();
			svBigRedButton->setOn(true);
			svBigRedButton->setDisabled(true);
//			if(devmanager.isCounting()) slotDisableItems();
//			else slotEnableItems();
		}

		if(StateExport::isWithoutDevice()) {
			StateExport::cleanWithoutDevice();
			disableDevice(true);
		}
		if(StateExport::isWithDevice()) {
			StateExport::cleanWithDevice();
			disableDevice(false);
		}

		if(oldState.hv != StateExport::hv()) {
			oldState.hv = StateExport::hv();
			slotTurnHV(oldState.hv);
		}
		if(oldState.safety != StateExport::safety()) {
			oldState.safety = StateExport::safety();
			slotTurnSafety(oldState.safety);
		}
		if(oldState.illum != StateExport::illum()) {
			oldState.illum = StateExport::illum();
			slotTurnIllum(oldState.illum);
		}
		if(oldState.light != StateExport::light()) {
			oldState.light = StateExport::light();
			slotTurnLight(oldState.light);
		}
#ifdef __MOTOR__
		if(oldState.aperture != StateExport::aperture()) {
			oldState.aperture = StateExport::aperture();
			slotAperture(oldState.aperture);
		}
		if(oldState.door != StateExport::door()) {
			oldState.door = StateExport::door();
			slotTurnDoor(oldState.door);
		}
#endif
#ifdef __COUNTER_BUTTON__
		if(oldState.buttons != StateExport::buttons()) {
			oldState.buttons = StateExport::buttons();
			slotTurnButtons(oldState.buttons);
		}
#else
 		if(oldState.mirror != StateExport::mirror()) {
			oldState.mirror = StateExport::mirror();
			slotTurnMirror(oldState.mirror);
		}
#endif
		if(oldState.temperature != StateExport::temperature()) {
			oldState.temperature = StateExport::temperature();
			slotTemperature(oldState.temperature);
		}
	}
	digitalclock->showTime();
	if(_updateHA) {
		star.updateCurrent();
		if(existShowStarDialog && _showHA) showStarDialog->showHA();
	}
	setLabelText(labelZenit, star.getAirMass());
	setLabelText(labelStar, star.getName());
	isTimerEventLocked = false;
}

void MyWidget::slotIllumDialog(void) {
	if(!existIllumDialog) {
		illumDialog = new IllumDialog();
		CHECK_PTR(illumDialog);
		existIllumDialog = true;
	}
	illumDialog->update();
	illumDialog->show();
}

void MyWidget::slotLightDialog(void) {
	if(!existLightDialog) {
		lightDialog = new LightDialog();
		CHECK_PTR(lightDialog);
		existLightDialog = true;
	}
	lightDialog->update();
	lightDialog->show();
}

void MyWidget::slotHVDialog(void) {
	if(!existHVDialog) {
		hvDialog = new HVDialog();
		CHECK_PTR(hvDialog);
		existHVDialog = true;
	}
	hvDialog->update();
	hvDialog->show();
}

void MyWidget::slotBackCorrectionDialog(void) {
	jdebug("Slot BackCorrection Dialog!!!");
	int ret = QMessageBox::information(0,"background",
			"Point to the starless area of sky","Continue","Cancel");
	if(ret != 0) {
		slotShowMessage("Canceled...",4000);
		return;
	}

	if(!existBackCorrectionDialog) {
		backCorrectionDialog = new BackCorrectionDialog();
		CHECK_PTR(backCorrectionDialog);
		existBackCorrectionDialog = true;
	}
	backCorrectionDialog->update();

	int oldApert = devmanager.currentAperture();
	if(oldApert == UNDEFINED) {
		QString str = "Invalid current aperture";
		switch(showWarning(str, 2, "Ignore", "Cancel")) {
			case  QMessageBox::Abort  : return;
			case  QMessageBox::Ignore : break;
			default	: break;
		}
	}
	ret = backCorrectionDialog->exec();
	if(ret == QDialog::Accepted) {
		slotShowMessage("Accepted...",4000);
	} else {
		slotShowMessage("Canceled...",4000);
	}
	backCorrectionDialog->finish();
	devmanager.setAperture(oldApert, 1);
}

void MyWidget::slotCenterCorrectionDialog(void) {
#ifdef __MOTOR__
	int ret = QMessageBox::information(0,"centering",
			"Point to the star","Continue","Cancel");
	if(ret != 0) {
		slotShowMessage("Canceled...",4000);
		return;
	}

	if(!existCenterCorrectionDialog) {
		centerCorrectionDialog = new
			CorrectionDialog("Aperture Center","ShiftRA",
				"ShiftDec",CENTER);
		CHECK_PTR(centerCorrectionDialog);
		existCenterCorrectionDialog = true;
	}
	centerCorrectionDialog->update();

	int oldApert = devmanager.currentAperture();
	if(oldApert == UNDEFINED) {
		QString str = "Invalid current aperture";
		switch(showWarning(str, 2, "Ignore", "Cancel")) {
			case  QMessageBox::Abort  : return;
			case  QMessageBox::Ignore : break;
			default	: break;
		}
	}
	ret = centerCorrectionDialog->exec();
	if(ret == QDialog::Accepted) {
		slotShowMessage("Accepted...",4000);
	} else {
		slotShowMessage("Canceled...",4000);
	} 
	centerCorrectionDialog->finish();
	devmanager.setAperture(oldApert, 1);
#endif
}

void MyWidget::slotTrackCalibrationDialog(void) {
#ifdef __MOTOR__
	int ret = QMessageBox::information(0,"centering",
			"Point to the star","Continue","Cancel");
	if(ret != 0) {
		slotShowMessage("Canceled...",4000);
		return;
	}

	if(!existTrackCalibrationDialog) {
		trackCalibrationDialog = new
			CorrectionDialog("Track Calibration","SpeedRA",
				"SpeedDec",TRACK);
		CHECK_PTR(trackCalibrationDialog);
		existTrackCalibrationDialog = true;
	}
	trackCalibrationDialog->update();

	int oldApert = devmanager.currentAperture();
	if(oldApert == UNDEFINED) {
		QString str = "Invalid current aperture";
		switch(showWarning(str, 2, "Ignore", "Cancel")) {
			case  QMessageBox::Abort  : return;
			case  QMessageBox::Ignore : break;
			default	: break;
		}
	}
	ret = trackCalibrationDialog->exec();
	if(ret == QDialog::Accepted) {
		slotShowMessage("Accepted...",4000);
	} else {
		slotShowMessage("Canceled...",4000);
	} 
	trackCalibrationDialog->finish();
//	devmanager.endTrackCalibration();
	devmanager.setAperture(oldApert, 1);
#endif
}

void MyWidget::displayStarPosition(int code) {
#if defined __MOTOR__ || defined __SCANMOTOR__
	jdebug("display Star position. Code=%d", code);
	switch(code) {
		case 11 : labelDiaphragm->setPixmap(ld); break;
		case  1 : labelDiaphragm->setPixmap(dd); break;
		case 21 : labelDiaphragm->setPixmap(rd); break;
		case 10 : labelDiaphragm->setPixmap(ll); break;
		case  0 : labelDiaphragm->setPixmap(cc); break;
		case 20 : labelDiaphragm->setPixmap(rr); break;
		case 12 : labelDiaphragm->setPixmap(lu); break;
		case  2 : labelDiaphragm->setPixmap(uu); break;
		case 22 : labelDiaphragm->setPixmap(ru); break;
		case -1 : labelDiaphragm->setPixmap(empt); break;
	}
#endif
}

void MyWidget::keyPressEvent(QKeyEvent *k) {
	if(devmanager.isTracking()
		&& (k->key() == Key_Control)
		&& !k->isAutoRepeat()) { //ForceMove
		devmanager.forceTrack();
	}
	if(k->state() != ShiftButton) return;
//	if(_disableKeyEvent) return;
	if(!devmanager.withAutoGuide()) return;
	if(devmanager.isTracking()) return;
	if(k->isAutoRepeat()) return;
	switch(k->key()) {

		case Key_Right	: //RA+
				if(moveRAplus() == SUCCESS)
					raPlus->setDown(true);
				break;

		case Key_Left	: //RA-
				if(moveRAminus() == SUCCESS)
					raMinus->setDown(true);
				break;

		case Key_Down	: //Dec-
				if(moveDecMinus() == SUCCESS)
					decMinus->setDown(true);
				break;

		case Key_Up	: //Dec+
				if(moveDecPlus() == SUCCESS)
					decPlus->setDown(true);
				break;
        }
}

void MyWidget::keyReleaseEvent(QKeyEvent *k) {
//	debug("ReleaseEvent!");
	if(!devmanager.isTracking() 
		&& !StateExport::isUserTrack()) return;
	if(k->isAutoRepeat()) return;
//	debug("ReleaseEvent after isTracking!");
	switch(k->key()) {
		case Key_Right	: //RA+
		case Key_Left	: //RA-
		case Key_Down	: //Dec-
		case Key_Up	: //Dec+
				stopTrack();
				raPlus->setDown(false);
				raMinus->setDown(false);
				decPlus->setDown(false);
				decMinus->setDown(false);
				break;
	}
}

void MyWidget::closeEvent(QCloseEvent* e) {
	jdebug("CloseEvent!!!");
	myQuit();	
	e->accept();
}

void MyWidget::normal(void) {
	scenariothr::normal();
}

void MyWidget::generalized(void) {
	scenariothr::generalized();
}

void MyWidget::centerControl(void) {
	scenariothr::centerControl();
}

void MyWidget::focusControl(void) {
	scenariothr::focusControl();
}

void MyWidget::backgroundMeasurement(void) {
	scenariothr::backgroundMeasurement();
}

void MyWidget::fluxEstimation(void) {
	scenariothr::fluxEstimation();
}

void MyWidget::exchangeTest(void) {
	scenariothr::exchangeTest();
}

void MyWidget::detectorTest(void) {
	scenariothr::detectorTest();
}

void MyWidget::statisticTest(void) {
	scenariothr::statisticTest();
}

void MyWidget::detectorCountingMeasurement(void) {
	scenariothr::detectorCountingMeasurement();
}

void MyWidget::detectorStatisticMeasurement(void) {
	scenariothr::detectorStatisticMeasurement();
}
