/*--------------------------------------------------------------------------*/
/** @file outcontrol.h
@brief  Manipulation of the output in the main results and graphic results 
windows.

                  MASS project: TURBINA module file
                       header file for outcontrol.cpp

The module "tells" the other modules which data should be displayed and which
should be hidden depending on the (current) state of CFG parameters which are 
listed in the \c Display section of \c turbina.cfg.



 @author O. Voziakova, Sternberg Institute (ovoz@sai.msu.ru)
 
 @version 1.0
*/
#ifndef OUTCONTROL_H
#define OUTCONTROL_H

#include <qobject.h>
#include <qstring.h>
#include <qcolor.h>
#include "summary.h"

#define LUT_LENGTH	256
#define MAX_DSI_NUMBER	4

class GraphSet {
private:
	QString		_pname;
	double*		_xData;
	double*		_yData;
	double		_refer;
	double		_xMin;
	double		_xMax;
	double		_yMin;
	double		_yMax;
	int		_nPoint;
	int		_nPointGraph;
	int		_size;
	double		_quantum; // 0.0 .. 1.0
	const char*	_color;
	bool		_changed;


public:
	GraphSet( void ) { _changed = true; _xData = NULL;
		_yData = NULL; _nPoint = 0; 
		_nPointGraph = 0; _refer = sum::NODATA; }
	void load( QString& pname_,
			const char* color,
			int size_,
//			sum::reftype rtype,
			double refer,
			double* x = NULL,
			double* y = NULL,
			int nP = 0,
			double scale = 1 );

	void		setData( double* x,
					double* y,
					int nP );
	void		setNPoint( int nP );
	void		setNPointGraph( int nP );
#if defined __MOTOR__ || defined __SCANMOTOR__
	void		copyScan( double* y, int startX, int nP );
#endif
	void		setQuantum( double quant ) {_quantum = quant;}
	void		setChanged( bool chng ) { _changed = chng; }
	bool		isChanged( void ) { return _changed; }
	double*		xData( void ) { return _xData; }
	double*		yData( void ) { return _yData; }
	double		refer( void ) { return _refer; }
	void		setRefer( double refer ) { _refer = refer; }
	int		nPoint( void ) {return _nPoint; }
	int		nPointGraph( void ) {return _nPointGraph; }
	bool		searchMinMax( void );
	double		xMin( void ) { return _xMin; }
	double		xMax( void ) { return _xMax; }
	double		yMin( void ) { return _yMin; }
	double		yMax( void ) { return _yMax; }
	const char*	color( void ) {return _color;}
	double		quantum( void ) {return _quantum;}
	int		size( void ) {return _size;}
	const char*	pname( void ) {return _pname.data(); }
	bool		isScan( void );
	bool		isProfile( void );
};

class GraphData {
private:
	QArray<GraphSet*> graphSet;
	int _nData;
public:
	GraphData( void ) { _nData = 0; }
	~GraphData();
	void clean( void );
	void add( QString& pname,
			const char* color,
			int size,
//			sum::reftype rtype,
			double refer,
			double* xData = NULL,
			double* yData = NULL,
			int nPoint = 0,
			double scale = 1 );
	void setSummaryData( void );
	GraphSet* getGraphSet( int num ) { 
		if( num < _nData ) return graphSet[num];
		else return NULL; }
	GraphSet* getGraphSet( const char* what,
			const char* entity );
	int nGraph( void ) {return _nData;}
	void setNPoint( int nP );
};

class ShowSet {
private:
	bool _graph;
	bool _text;
	bool _err;
public:
	ShowSet( void ) {_graph = false;
			_text = false;
			_err = false; }
	void set( const char* show );
	void set( ShowSet& showSet );
	bool graph(void) {return _graph;}
	bool text(void) {return _text;}
	bool err(void) {return _err;}
};

class DisplayData {

private:
	const char*	_entity;
	double		_refer;
	ShowSet		_showSet;
	const char*	_color;
	int		_size;

public:
	DisplayData( void ) { _entity = "A";
				_showSet.set("N");
				_color = "white";
				_size = 1; }
	void		setEntity( const char* _entity_ ) {_entity = _entity_;}
	void		setRefer( double refer ) {_refer = refer;}
	ShowSet*	showSet( void ) {return &_showSet;}
	void		setColor( const char* _color_ ) {_color = _color_;}
	void		setSize( int _size_ ) {_size = _size_;}

	const char*	entity( void ) {return _entity;}
	bool		isgen( void );
	bool		iscross( void );
	bool		showGraph( void ) {return _showSet.graph();}
	bool		showText( void ) {return _showSet.text();}
	bool		showErr( void ) {return _showSet.err();}
	const char*	color( void ) {return _color;}
	int		size( void ) {return _size;}
	double		refer( void ) {return _refer;}
};

class OutControl : public QObject {
	Q_OBJECT
private:
	GraphData	graphData;

// Profile:
	QString		_lutDirName;
	QArray<QString*> lutList;
	int		_nLut;
	QColor		lutTable[LUT_LENGTH];
	bool		_percentType;
	double*		_zSmooth;
	double*		_cn2Smooth;
	int		_nZsmooth;
	double		_cn2ScaleFactor;

	sum::reftype _fluxRefType;
	sum::reftype _indexRefType;
	sum::reftype _desiRefType;
	sum::reftype _integralRefType;
	sum::reftype _profileRefType;

	const char* dsiName[MAX_DSI_NUMBER];
	int	_nDSI;
	bool	_showFlux;
	bool	_showFluxErr;
	bool	_showIndex;
	bool	_showIndexErr;
	bool	_showIndexG;
	bool	_showIndexGerr;
	bool	_showDSI;
	bool	_showDSIerr;
	bool	_showDSIg;
	bool	_showDSIgErr;
	bool	_showDESI;
	bool	_showDESIErr;

	bool	_showSee;
	bool	_showHeff;
	bool	_showIsopl;
	bool	_showM2;
	bool	_showM0;
	bool	_showTau;
	bool	_showProfile;

	bool _drawProfile;

	bool isFirst;
	bool _graphReady;
	bool _graphReDraw;

private:
	void loadLutTable( const QString& fileName );
	void prepare( QArray<DisplayData>& scan,
			QArray<DisplayData>& flux,
			QArray<DisplayData>& desi,
			QArray<DisplayData>& index,
			QArray<DisplayData>& integral,
			DisplayData& profile );

public:
	OutControl( void );
	~OutControl();
	void loadConst( void );

	GraphSet* graphSet( int num ) { return graphData.getGraphSet( num ); }
	GraphSet* graphSet( const char* what,
			const char* entity ) { 
			return graphData.getGraphSet( what,entity ); }
	const int nGraph( void ) { return graphData.nGraph(); }
	const int getNlut( void ) { return _nLut; }
	double* zSmooth( void ) { return _zSmooth; }
	double* cn2Smooth( void ) { return _cn2Smooth; }
	const int nZsmooth( void ) { return _nZsmooth; }
	const QString& getLutName( int nL );
	const QColor& getLutColor( double val );
	bool isPercent( void ) { return _percentType; }

	bool showFlux( void ) { return _showFlux; }
	bool showFluxErr( void ) { return _showFluxErr; }
	bool showFluxG( void ) { return _showFlux; }
	bool showFluxGerr( void ) { return _showFluxErr; }
	bool showSI( void ) { return _showIndex; }
	bool showSIerr( void ) { return _showIndexErr; }
	bool showSIg( void ) { return _showIndexG; }
	bool showSIgErr( void ) { return _showIndexGerr; }
	bool showDESI( void ) { return _showDESI; }
	bool showDESIerr( void ) { return _showDESIErr; }
	bool showDSI( void ) { return _showDSI; }
	bool showDSIerr( void ) { return _showDSIerr; }
	bool showDSIg( void ) { return _showDSIg; }
	bool showDSIgErr( void ) { return _showDSIgErr; }
	int getNdsiName( void ) { return _nDSI;}
	const char** getDSIname( void ) { return dsiName; }
	bool showSee( void ) { return _showSee; }
	bool showHeff( void ) { return _showHeff; }
	bool showIsopl( void ) { return _showIsopl; }
	bool showM2( void ) { return _showM2; }
	bool showM0( void ) { return _showM0; }
	bool showTau( void ) { return _showTau; }
	bool showProfile( void ) { return _showProfile; }

	bool drawProfile( void ) { return _drawProfile; }

	bool isGraphReady( void );
	bool reDraw( void );
	void setGraphReady( bool ready, bool reDraw = false );
	void resetGraphReady( void );
	void setNPoint( int nP ) { graphData.setNPoint( nP ); }
	void addSummaryData( double ut,
				double mag,
				double airmass,
				bool ready,
				bool isgen );
	void resetSummaryData( void ); // if sum::done/sum::init() bylo

private slots:
	void change( void );
};

#endif
