/*--------------------------------------------------------------------------*/
/** @file procounter.h
@brief  Implementation of interaction (command calls and replies from modules, 
their current data and status keeping, etc.) with photometric modules.

                  MASS project: TURBINA module file
                       header file for counter.cpp



 @author O. Voziakova, Sternberg Institute (ovoz@sai.msu.ru)
 
 @version 1.0
*/
#ifndef PROCOUNTER_H
#define PROCOUNTER_H

#include <qarray.h>
#include "module.h"
#include "iocount.h"

#define BYTE_PER_COUNT	2	// in principle, may be "short" regime, then - one BYTE
#define NS2MS			(1.0E-6)	// nseconds to millisends conversion

#define MAX_LEVEL_VALUE			255
#define MIN_LEVEL_VALUE			0
#define MIN_EXPOSITION_NUMBER	1
#define MAX_EXPOSITION_NUMBER	32767
#define MIN_EXPOSE				0.2	// ms
#define MAX_EXPOSE				32.0	// ms
#define MIN_BLOCK_SIZE			1
#define MAX_BLOCK_SIZE			16

// Status of counter:

#define ACTIVE		0x01	// active_on
#define INDUCE		0x02	// induce_on
#define SHORT		0x04	// 1-byte data
#define SLAVE		0x08	// slave - 1, master - 0
#define SHUTTER		0x10	// shutter_on
#define TEST		0x20	// test
#define COUNT		0x80	// Integration in process

class BlockCounter {
private:
	int _number;
public:
	BlockCounter(void);
	int incr(void);
	int decr(void);
	int get(void) {return _number;}
	int set(int);
	void reset(void) {_number=0;}
};

class ProCounter : virtual public HardModule {
protected:
	uchar _stop;
	uchar _masterOn;
	uchar _masterOff;
	uchar _activeOn;
	uchar _activeOff;
	uchar _induceOn;
	uchar _induceOff;
	uchar _setInduc;
	uchar _getInduc;
	uchar _runMst;
	uchar _runTest;
	uchar _getNumber;
	uchar _setNumber;
	uchar _getBlSize;
	uchar _setBlSize;
	uchar _setExpos;
	uchar _getExpos;

	static const count_t INVALID_COUNT = max_count_t;
	int			_maxNData;
	QArray<int>	blockLabel;
	int			_nBlockLabel;


	int			_exposValue;
	int			_expositionNumber;
	int			_blockSize;
	int			_lastLabelNumber;
	int			_lastLabel;

public:
	ProCounter(void);
	virtual ~ProCounter();

	virtual void resetVariables(void);	// !!! virtual, t.k. vyzyvaetsia iz roditelia (HardModule)
	virtual void create(int sizeOfData);
	virtual int loadOwnConst(void);

	virtual void resizeData(int newSize) = 0;
	virtual void resetData(void) = 0;
	void addBlockLabel(int newLabel);
	/**
	@return	The fact that data were expanded due to the loss */
	virtual bool expandData(void) = 0;
	virtual int insertBlock(int label, count_t value) = 0;
	virtual int removeBlock(int label) = 0;
	virtual int searchTestErrors(count_t test_number) = 0;

	char getModule(void) {return _module;}
	void setName(const char* name);
	const char* getName(void) {return _identName;}

	int stop(int dialog=1);
	int setMasterOn(int dialog=1);
	int setMasterOff(int dialog=1);
	int setActiveOn(int dialog=1);
	int setActiveOff(int dialog=1);
	int setInduceOn(int dialog=1);
	int setInduceOff(int dialog=1);
	int setInduc(uchar masterAddress, int dialog=1);
	int setBlSize(uchar blockSize, int dialog=1);
	int setNumber(int expositionNumber, int dialog=1);


	int setExpos(double expos, int dialog=1);
	int runMst(int dialog=1);
	int runTest(int dialog=1);

	BlockCounter blockCounter;
};

#endif
