/**
	@file	readmass.h
	@brief	Reading the content of the mass-file for Playback function
	 and CFG parameters last-written values.

	This module implements reading and parsing of lines of a \e mass-file.
	This is done in course of following tasks:
	- comparison of the configuration file (CFG) parameters written previously in a
	current mass-file with those read from the CFG (repetitive start of TURBINA
	in the same night). Deals with the so-called \e preamble-type lines
	('P'-lines).
	- the "playback" of an old mass-file (normally named as \c YYMMDD.mass.in)
	for re-reducing of the previous nights data. Here, apart from the 'P'-lines,
	we take care mostly of the \e base-time originated lines (as that with
	prefixes 'i','j') and new modes declarations ('M'-lines). Also, the object
	info lines ('O') are parsed.
	
	\b Usage

	Reading the single line:
	
	The function rdm::readline() reads and (possibly) parses the (next) line
	from the opened stream  of the mass-file. It returns the line prefix which
	denotes the type of line and parsing results. Some line types are parsed and
	the rest are simply provided "as is" by rdm::getvalue() (see below).

	The parsed data are available with following functions:
	- rdm::gettype() returns the line prefix just as rdm::readline() does, for
	\b all types, naturally;
	- rdm::isheader() tells for the comment-type lines (prefix \b '#' ) whether
	the line is a header-line (second character is a prefix also).
	- rdm::getut() returns the UT moment reference, for \b all types except for
	header-lines (starting with \b '#*' where '*' is a prefix character);
	- rdm::getname() returns the name of parameter (\b 'P'-type) or started mode
	(\b 'M'-type) or catalogue (HR) number of a star (\b 'O'-type) in \b
	upper-case (unless parameter 1 is given for original case);
	- rdm::getvalue() [no arguments] returns the parameter value for \b
	'P'-type line, or the content of the line to the right of the star number
	for \b 'O'-type (name, coordinates etc.) or the airmass (if there is one)
	for \b 'M'-type or the rest of the line to the right of the UT reference
	for \e non-parsed lines;
	- rdm::getaddr() returns the \em count-file address, for \b 'i','j'-types;

	The parsed parameters are stored in global (private) variables, and are
	modified solely by overwriting. This means that, for example, the count-file
	address	returned by rdm::getaddr() will stay unchanged while the lines
	without one are read.
	
	Reading the mass-file from SCIND ver2.0 and later (turbina ver1.3 and later)
	and accompanying count stat.moments file:
	
	The function rdm::readline(m,sw,s) is provided being the interface to all 
	other rdm functions which reads one of two files maintaining the correct 
	order of output data.

	Reading the whole mass-file for the parameter values set:

	The function rdm::grabparams() should be called on the opened mass-file
	stream to collect the preamble-type information. The pairs of parameter
	names and their associated latest assigned values are stored in a \e map.
	Then, these latest written parameter values are available with
	rdm::getvalue() with an argument \b exactly \b equal to the parameter name
	as a \e key of this map (in upper-case, as returned by rdm::getname(false)).

	\note
	No error is set by this simple module to be checked as nr::erget(). The
	reading failures are to be traced by the return values (see e.g.
	rdm::getut())

	@author N.Shatsky
	@version 1.0 Creation
	
		1.1 Bugs (minor) correction, overloaded getut()
		
		1.5 readline(two files) for Turbina ver1.3 implemented
		
		1.6 new time format: date added as YYYY-MM-DD in mass-file. Both old
		and new formats are still supported (date is read if exists only)
		
		1.7 minor corrections to docs; readline(m,sw,s) now accepts char* sw to 
		allow several switchers (for example, "IJ").
		
		1.8 getname(true) returns name in original case
*/
//#define __DOS

#include <stdio.h>
#ifndef __DOS
#include <string>


/**
	@brief	Reading and parsing the mass-file lines
*/
namespace rdm {
#else
#include <string.h>
#include <ctype.h>
#define string char*
#define bool char
const char false = 0;
const char true = 1;
#endif

/** Invalid count-file address, returned when not found or failed to convert */
const long NOADDR = -1;

/**
	@brief Read the next line from the mass-file stream \a mf
	@param	mf	the stream pointer of an opened mass-file
	@return	line prefix or \0 for failed reading/parsing or if EOF

	Reading the line and parsing it if possible: for line types \b
	'P','M','i','j','O' . No action is performed if null-file \a mf is
	supplied (segm.fault-protected).
	
	Before parsing the line, the carriage-return character is removed from the
	line.

	The first character of line determines the line \e type and thus the
	returned value (see also gettype()).

	For any line but that of header-type the UT is extracted from the second
	field and can be obtained with getut(). UT is either UT time only or both 
	UT date and	time.

	The \e name (returned by getname()) is extracted for 'P'-, 'M'-lines and
	'O'-lines starting \em after the \em last space character after UT record:
	- 'P' : name is ended \em before the (first) '=' character (\em after this
	'=' the \em value of the parameter starts)
	- 'M' : name of mode is ended at the last non-space character or - if the 
	last word in the line is a number (airmass) - at the last character of the
	previous word (normally, the first). 
	- 'O' : name (HR-number) lasts till the end of line. Note that the 
	length of the HR-record depends on the numeric value of HR (e.g. 3 chars for
	HR=999).

	The \e value (returned by getvalue()) is extracted for 'P'-lines (after the
	first '='-character) and for non-parsed and comment-lines.

	The \e address (see getaddr()) is converted from "@..." substrings of
	the	mass-file line or set to rdm::NOADDR if not found.
*/
char readline(FILE * mf);

/**
	@brief	Return the last line prefix
	@return prefix of the last successively read/parsed line or \0 if no lines
	were read
*/
char gettype();

/**
	@brief	Disentangle the comment- and header-type lines
	@return	Header-line flag: \b true for lines started as '#*' and \b false for
	any other line type (comment or non-comment)
*/
bool isheader();

/**
	@brief	Return starting position of last-read record
	@return file position before reading last line
	
	File position is saved before each readline(); thus, with help of this 
	function, one line may be read twice:
	\verbatim
	rdm::readline(f) ; // some line read
	fseek(f, rdm::getstartpos(), SEEK_SET);
	rdm::readline(f) ; // the same line again
	\endverbatim ; 
	
*/
long getstartpos();


/**
	@brief	Return the parsed UT moment as hours, minutes and seconds
	@param	UTh	Resulted UT hour (may be set zero if no UTh needed)
	@param	UTm	Resulted UT minute (may be left zero)
	@param	UTs	Resulted UT second (may be left zero)
	@return	Fractional UT hour or -1 on error of UT parsing
*/
double getut(int * UTh, int * UTm = 0, int * UTs = 0);

/**
	@brief	Return the parsed UT moment as date components, hours, minutes and seconds
	@param	UTY	Resulted UT year (may be set zero if not needed)
	@param	UTM	Resulted UT month number (may be set zero if not needed)
	@param	UTD	Resulted UT day (may be set zero if not needed)
	@param	UTh	Resulted UT hour (may be set zero if no UTh needed)
	@param	UTm	Resulted UT minute (may be left zero)
	@param	UTs	Resulted UT second (may be left zero)
	@return	Fractional UT hour or -1 on error of UT parsing
*/
double getut(int * UTY, int * UTM, int * UTD, int * UTh, int * UTm = 0, int * UTs = 0);

/**
	@brief	Return the string-type UT moment (unparsed)
	@return	UT-string
*/
std::string getut();

/**
	@brief	Return the parameter name or started mode name or star catalogue ID
	@param orig flag to return the name in original case instead of upper-case
	@return	Name of parameter as Section[/SubSection[/SubSubSection]] or star's
	catalogue ID (first field after UT in 'O'-lines, normally HR) or mode name
*/
std::string getname(bool orig = false);

/**
	@brief	Return the value of parameter or star parameters or line content
	@param	pname	parameter name in \b upper-case for accessing the map of
		parameter name/values created by grabparams(). Empty by default for
		getting the "value" last read with readline().
	@return The rest of line to the right of the first '=' for 'P'-lines or
		acquired parameter name \a pname; or the catalogue ID for 'O'-lines
		or the full line for header-type lines; or the mode parameters for 
		'M'-lines or the line content to the right of UT for other 
		non-parsed lines

	This function is of the dual use. First, it returns the \em value for a
	mass-file line which was just read with readline() (see \b Returns
	description). Second, it returns the value associated with the provided
	parameter name \a pname taken from the parameters map created by
	grabparams().
*/
std::string getvalue(const std::string pname = "");

/**
	@brief	Return the count-file address for the last read base-time
	results line ('i','j'-type)
*/
long getaddr();

/**
	@brief	Add the parameter and its value to the map filled with grabparams()
	@param	pname	parameter name in \b upper-case for adding in the map of
		parameter name/values created by grabparams(). Empty by default for
		simply getting the size of the map.
	@param	pvalue	parameter \a pname value. May be empty if, say, no value is
		associated with \a pname.
	@return	The size of the parameters map after adding the parameter \a pname.
	
	This function simply assigns the value \a pvalue (if not
	empty/defaulted) to the map \em key \a pname (i.e., creates a new entry
	\a pname if not existent before and associates it with the (new) value
	\a pvalue). Returned is the size() of the parameters map.
*/
int add2map(const std::string pname = "", const std::string pvalue = "");

/**
	@brief	Read the whole mass-file to collect the parameter values from the
	preamble-type lines
	@param	mf	the stream pointer of an opened mass-file
	@return	Number of preamble-type lines encountered (\b not of parameters
	stored in map - it is less if the same parameter was encountered, say, twice).

	The utility first sets the file position at its origin and then collects
	all the preamble-type line contents in an associative array (map). The
	values of parameters, if encountered repetitively are thus overwritten and
	upon the end of reading represent the last-written values.

	After the function finishes its job, the file pointer is left at the end
	of file. The parameter values may be acquired with rdm::getvalue() with an
	exact parameter name in upper-case as an argument.
	
	If null \a mf is supplied, the accumulated map of parameter-value pairs is
	only emptied.
	
	\note	grabparams() uses the function readline() for reading the mass-file.
	So, the content of the variables returned by gettype(), getvalue() etc. is
	modified compared to pre-grabparams() state.
*/
int grabparams(FILE * mf);

/**
	@brief  Read synchronously mass- and stm-files
	@param  mass    Primary file: mass-file
	@param  sw  prefix(es) on which to switch to \a stm -file reading
	@param  stm     Secondary file: stat.moments file
 
	This overloaded procedure mimics the behavior of the one-parameter version
	of rdm::readline() inserting the content of one of two files on the output
	from rdm. Switching from the primary file (read first by default) to the
	secondary is done when either of chars of the prefix string \a sw is
	encountered as the line prefix \em and the last field in this line contains
	the valid UT moment reference - "UTend". The UT reference of the line
	itself (second column) is called "UTbeg".  The content of the switching
	line is released before switching.  Then, upon the next calls of the
	function, the reading of the secondary file is performed while its UT
	column contains the reference between the UTbeg and UTend, including
	boundaries. After passing all records between UTbeg and UTend or of the EOF
	of the secondary file, the reading will go again from the primary file.

	Thus, the order of reading of the mass-file information and the
	stm-file (raw) data resembles the data flow from the old-format
	mass-files where the basetime-index lines were mixed with the rest
	of information.
 
	The utility uses the usual tools provided by the module readmass.h
	which, thus, "has no idea" which file is under reading now - mass-
	or stm-file.
 
	\note   If two latter parameters are defaults (omitted), the reading
	of the first file is simply performed.
*/
char readline(FILE * mass, const char* sw, FILE * stm);

#ifndef __DOS
} // of namespace rdm
#endif
