/****************************************************************************/
/*                 MASS project: TURBINA module file                        */
/*                            atmos.cpp                                     */
/*                    Description is given in atmos.hxx                     */
/* Editor tabular size: 4                                                   */
/****************************************************************************/
/* Uncomment ATMOSDEBUG to get the helpful info on STDERR about parameters
in action */
//#define ATMOSDEBUG 
/* Insert the macro ATMOSTEST in compilation for production of independent
executable (e.g. for tests) or uncomment below: */
//#define ATMOSTEST

/* Following block is converted into the Reference Manual with DOXYGEN program:
1. Generate Doxygen file with $ doxygen -g
2. Replace following lines in Doxygen with any text editor:
 PROJECT_NAME           = "Standalone MASS-data processor"
 PROJECT_NUMBER         = "Version 2.4"
 OUTPUT_DIRECTORY       = doc
 INPUT                  = atmos.cpp
3. Generate documentation: $ doxygen
4. Generate Manual in doc/latex/refman.ps: $ (cd doc/latex; make; make ps)
*/

/** \mainpage Standalone ATMOS
	\section intro Introduction
	This program makes off-line processing of the MASS files. In the
	version 2.1 and later it contains the full pipeline starting from
	calculation of the instant scintillation indices. This means that all the
	parameters which affect the data processing in MASS observations play role
	here and may be altered.
	
	The program uses the procedures and function which are identical to
	those used in Turbina program (v2.04 and later) which are called in
	the same order as in Turbina. The parameters which affect the data
	processing are read solely from the preamble ('P'-started) lines of
	the input mass-file, so, if one wants to modify, e.g., the
	magnification, the respective line in the mass-file should be
	edited.

	The program outputs the atmospheric parameters in the same format as
	Turbina does. The standard output is used which should be redirected to
	the new mass-file. The original mass-file and statistic moments (stm)
	file constitute the input of the program. Since each line contains the
	UT moment reference and the unique line prefix, two files may be
	concatenated and sorted in a single input file (as is expected in a
	future Turbina generation).

	\section params Parameters used in data processing
	
	Since the currently utilized versions of Turbina do not output all
	the needed parameters in the MASS-file, the extraction of the
	following device parameters in the header of the mass-file is
	obligatory (with help of a script cfg2mass.awk for example):

	- Bicounters\\Bicounter 1(2)\\NonLinearityA(B)
	- Bicounters\\Bicounter 1(2)\\NonPoissonA(B)
	
	These parameters are assigned to the characteristics of the specific device
	channel (A,B,C,D) with help of four cross-linking parameters which are
	normally written in the mass-file: 
	
	- Bicounters\\Channels\\ChannelA(B,C,D)=Counter 1(2)A(B)
	
	Alternatively, the ``device-independent'' form is also accepted for
	reprocessing (and overrides former definitions if both are present):
	
	- Channels\\ChannelA(B,C,D)\\NonPoisson
	- Channels\\ChannelA(B,C,D)\\NonLinearity
	- Channels\\ChannelA(B,C,D)\\Scatter
	
	This form is generated by the script param.awk. The last parameter is a new capability 
	of the program which allows to take into account the scattered light fraction in a given
	channel. See SCIND module description about changes in the scintillation indices formulii.

	Timing of measurements (number of base-times in an accumulation time
	and number of microexposures in a base-time) is taken from:
	
	- Operations\\Normal mode\\BaseTime
	- Operations\\Normal mode\\AccumTime
	- Operations\\Normal mode\\Exposition

	Together with the background fluxes taken from the line

	- Operations\\Background measurement\\Background
	
	all the mentioned parameters participate in calculation of
	scintillation indices.

	Apertures geometry is taken from following 9 parameters:
	
	- Segmentator\\Channel A(B,C,D)\\Inner
	- Segmentator\\Channel A(B,C,D)\\Outer
	- Optics\\Common\\Magnification
	
	Given the provided paths to spectral energy distribution files (option
	-s) and spectral response file (option -r), this information is used
	for checking and calculation of the scintillation weight functions.
	
	The altitude grid for the turbulence altitude moments decomposition
	if read from 
	
	- WeightFunction\\MinAltitude
	- WeightFunction\\MaxAltitude
	- WeightFunction\\Step
	- WeightFunction\\MinStep
	
	OR, preferentially, from 4-numbers parameter
	
	- Data processing\\Integrals\\AltitudeGrid

	Restoration of the turbulence profiles by 2 methods is
	performed on the grid from:
	
	- Data processing\\Floating Layers\\AltitudeGrid
	- Data processing\\Fixed Layers\\AltitudeGrid

	These AltitudeGrid-parameters define the grid of altitudes (in [km])
	on which the restoration and SVD of moments are performed. They are
	either "Z0, Zmax, dZ, dZmin" (see wf_t::setzgrid()) or simply a
	ascending-sorted list of altitudes like "1, 2, 4, 8, 16".  In case
	when these parameters (Data processing or WeightFunction) are
	absent, atmos uses the default grids specified in its header file.
	Maximal altitude is also restricted to 40km or more.


	The correction of the results by the airmass is performed. For this,
	the airmass is computed using the UT, star coordinates from the last
	O-line, and from the site geographic coordinates from

	- General\\Site\\Longitude
	- General\\Site\\Latitude
	

	\section data Input data line types and process organization
	
	The input lines which are considered by the atmos-processor are following:
	
	- P - parameter lines;
	- M - mode lines (only Normal mode is processed)
	- O - object lines
	- m - stm-file data for Normal mode
	- I - index lines (used only for swapping the read from mass- and stm-files)
	
	All the rest data is ignored. O-, P- and M-lines are printed in
	the output making the repetitive reprocessing possible. Processing
	of Generalized mode (mode Generalized and s-lines stm-file data
	for shifted part of data) is also supported but still not tested.
	
	Old format of the mass-file where the instant indices are given as i-lines
	is still supported, but needs the UT-stamps modification in the
	mass-file. This possibility is retained but currently was not tested.

	The work of the processor consists of the line-by-line reading of the input.
	The statistics file lines are read whenever the I-line with the UT moments span
	(beginning and end of the accumulation time) is encountered. 
	
	Each line is parsed and then processing depends on the line prefix. 
	First of all, the following condition is tested: prefix is be either
	'M' or 'O' or end-of-file is true, and some scintillation data are
	already collected (see below) and time passed since beginning of mode 
	is equal or more than accumulation time. If condition is met, then
	interpretation of accumulation-time data set is made. It consists of
	calculation of the middle-UT time reference (from first- and
	last-measurement UT references, see below), averaging the
	scintillation indices and turbulence integral moments, restoration
	of the turbulence profile by Fixed and Floating layers methods and
	writing the resulting indices (I-line), integral characteristics
	(A-line), profiles restored (two T-lines) and relative O-C of
	indices (two R-lines). Header line for each of these line types is
	output only once per program start (unless -H option is specified). 
	The interpreted data are finally reset.
	
	After this, the processing is made depending on the prefix only:
	
	- In case of P-line, the parameters are read into the internal
	text variables and echoed in the output as well. No calculations are
	made.
	
	- O-lines data are converted into the star coordinates and
	spectrum type, used for selection of the proper weight function file 
	and spectrum energy distribution file for its checking or
	calculation.
	
	- M-lines are processed only for Normal (or Generalized) modes;
	all the rest modes are simply printed in the output for reference. 
	If any parameter line was read after last M-line, then all the
	parameters used in calculations (see above section) are re-read from
	their saved text representation. Then the airmass value is
	recomputed and the first-measurement UT reference is saved. In case
	star changed, the weight file is checked and, if not found or found
	differing in parameters, it is recomputed (running percentage is
	output on the standard error output).  Finally, an Update of the
	altitude grids (according to the new airmass value) and some other
	quantities for turbulence restoration is made and the mode line is
	printed.
	
	- m-lines (and 's' for shifted altitude data) give the count
	statistics which are read from the line and the instant
	scintillation indices are computed. The turbulence moments according
	to their decompositions performed in Update are generated. The
	last-measurement UT reference is updated.

	- i-lines (and 'j' for shifted altitude mode) originating for old mass-files are read into
	the internal array of instant scintillation indices; no
	recomputation is possible even if the detector characteristics
	changed. The turbulence moments according to their decompositions
	performed in Update are calculated. The last-measurement UT
	reference is updated.

	- #-lines are echoed in the output if the second character in the line is also '#'. Thus, the
	messages from the filtering scripts are visible in the processed
	data.
	
	The use of the processor is following:

	1. Compilation in Turbina directory: make atmos

	2. Usage: see ./atmos -h

	\note
	This version of a Standalone ATMOS acts like Turbina respecting the
	checks and computation of weights (unless -n option is given). To
	specify where to take the spectra and response curve, use the -s and
	-r options; the place to take weights from is given by -w option.
	Finally, if all three kinds of information is stored in standard
	directories/files data/mass.crv, spectra/..sp and weight/..wf in the
	single parent catalogue, this catalogue (/opt/turbina/data by default)
	may be given by	-d option. If no parameters are given at all, the 
	weights are checked and computed in /opt/turbina/data weight using 
	spectra from /opt/turbina/data/spectra and response function from
	/opt/turbina/data/data/mass.crv).

*/
#include <math.h>
#include "atmos.hxx"
#include "scind.hxx"
#include "weif.hxx"
#include "nr.h"
#include <string.h>
#include <string>
#include <iostream>
// redefine CERR as cout for proper order of messages if needed
#define CERR cerr
using namespace std;

#ifdef ATMOSTEST
#include "readmass.h"
#include <map>
#include <getopt.h>
#include <time.h>

// define GCC33 for modern >=3.3 gcc versions:
#define GCC33
 
#ifdef GCC33
#include <sstream>
#else
#include <strstream>
#endif
#include <iostream>

/*---------------------------- TEST PART ---------------------------------*/

string get_midtime(const string b, const string e) {
	tm beg, end;
	strptime( b.c_str(), "%Y-%m-%d %T", &beg);
	strptime( e.c_str(), "%Y-%m-%d %T", &end);
	beg.tm_isdst = end.tm_isdst = -1;
	time_t tb = timegm(&beg);
	time_t te = timegm(&end);
	double diff = difftime(te,tb);
	tb += (int) (diff / 2);
	gmtime_r(&tb,&end);
	char res[20]; strftime(res,20,"%Y-%m-%d %H:%M:%S",&end);
//	CERR << b << " -- " << e << ": " << res << '\n';
	return string(res);
}

int julian( int year, int month, int day ) {
	int result = (3057*month)/100 + (365*year) + 1721028L;
	result = result + (year/4 - year/100 + year/400 + day);
	if( month > 2 ) return( result);
	int j = 2;
	if( year%4 == 0 )	j -= 1;
	if( year%100 == 0 ) j -= 1;
	if( year%400 == 0 ) j += 1;
	result += j;
	return( result );
}

/* Calculation of sidereal time in the Greewhich midnigth (in days) */

double greenwich_st0( int _julian ) {
	double c3 = 1.0775925925925926E-06;
	double c2 = 2.7379093507870371E-03;
	double c1 = 0.2790572732638889;

	double d = (double)_julian - (double)2451545.5;
	double t2 = d/(double)36525.0;
	double result = c1 + c2*d + c3*t2*t2;
	return ( result - floor( result ) );
}

/* Calculation of the airmass(radian args) */
double airmass(double ha, double dec, double lat) {
	double sinH = sin(lat) * sin(dec) +
		cos(lat) * cos(dec) * cos(ha);
	double secz1 = 1.0 / sinH;
	double secz2 = secz1 - 1.0;
	double _airMass = secz1 - secz2 * (0.0018167 +
			secz2 * (0.002875 + 0.0008083 * secz2));
	return _airMass;
}

int main( int argc, char **argv ) {
	// FILES:
	const int FNL = 255;
	char tdir[FNL+1] = "/opt/turbina/data";
	char wdir[FNL+1] = "";
	char rfile[FNL+1] = "";
	char sdir[FNL+1] = "";
	const char SEDEXT[] = ".sp";	// SED file extension
	const char WFEXT[] = ".wf";	// weight file extension

	bool nocheck = false, oneheader = true;
	
	// parse parameters:
	while (1) {
		int c = getopt(argc,argv,"r:s:d:w:nHhv");
		if (argc<2) c = 'h'; // no mass(+stm) file(s) - output help
		if (c==-1) break;
		switch (c) {
			case 'r': strncpy(rfile, optarg, FNL); break;
			case 's': strncpy(sdir , optarg, FNL); break;
			case 'd': strncpy(tdir , optarg, FNL); break;
			case 'w': strncpy(wdir , optarg, FNL); break;
			case 'n': nocheck = true; break;
			case 'H': oneheader = false; break;
			case 'h': default:
				CERR << "ATMOS: atmospheric turbulence and parameters computer\n"
				"Usage: \n./atmos [-d tdir] [-w wfdir] [-r rfile] [-s spdir] [-nH] <mass> [<stm>]\n"
				"  Options:\n"
				"\t-d tdir - directory with /weight/*.wf, /data/mass.crv, /spectra/*.sp\n"
				"\t-w weight-files directory (overrides -d) or file (if no O-lines in mass)\n"
				"\t-r response-file (overrides -d)\n"
				"\t-s spectra-files directory (overrides -d)\n"
				"\t-n - force no weights checking/recalculations (hence ignore -r and -s)\n"
				"\t-H - force header-line for each result line\n"
				"  mass - mass-file, may also be '-' for standard input.\n"
				"  stm  - stm-file (content may, properly sorted, be mixed into mass-file.\n";
				exit (-1);
		}
	}
	if (!strlen(rfile)) {
		strcpy(rfile, tdir);
		strcat(rfile, "/data/mass.crv");
	}
	if (!strlen(wdir)) {
		strcpy(wdir, tdir);
		strcat(wdir, "/weight/");
	}
	if (!strlen(sdir)) {
		strcpy(sdir, tdir);
		strcat(sdir, "/spectra/");
	}

	// dynamically created file names:
	string wfile = wdir;
	string sfile = "NO STAR SELECTED";

	string dfile = string(wdir)+"/atmos.tmp";	// Cn2 profile moments decomposition file

	FILE * fout = stdout;
	FILE * fmass = strcmp(argv[optind],"-") ? fopen(argv[optind],"r") : stdin ;
	if (!fmass) {
		CERR << "Problems with file "<<argv[optind] <<". Exit\n";
		exit(-1);
	}

	FILE * fstm = NULL;
	if (argc>optind+1) { fstm = fopen(argv[optind+1],"r"); }

#ifdef ATMOSDEBUG
	CERR << "Tdir="<<tdir<<"\nWdir="<<wdir<<"\nSdir="<<sdir<<"\nRfile="<<
	rfile<<"\nNocheck="<<nocheck<<"\nmass-file="<<argv[optind]<<"\nstm-file="<<
	(fstm ? argv[optind+1]:"no")<<'\n';
#endif

	// switching-to-stm-file mass-file line prefixes: see rdm::readmass(f,sw,s)
	// This will have no effect if no separate stm-file is provided.
	char sw[] = "IJ";	

	// number of MASS channels: a constant
	const int NCHAN = 4;

	// starting defaults: overridden by P-lines in mass-file:
	const double min_zmax = 40;
	double z0=0, zmax=40, zstep=0.25, zminstep=0.04; // z-grid for Cn2 moments
	double longitude=0; // fractional eastern longitude of the site (0<l<1)
	double latitude=0; // radian latitude of the site
	double zshift1=0, zshift2=0 ;	// zshifts of normal and 1st half of generalized modes
	double bkgr[NCHAN]={0,0,0,0};
	double zgrid[atm::NFIXED];	// user z-grid (see atm::setzcn2())
	double innerDiam[NCHAN], outerDiam[NCHAN];
	double magnification = 0;
	double deadtime[NCHAN] = {1e-8,1e-8,1e-8,1e-8};//[s]
	double nonpois[NCHAN] = {1,1,1,1};
	double scatter[NCHAN]={0,0,0,0};
	double microexp = 1; // [ms]
	
	// overridden by M-lines:
	double secz   = 1;	

	const double NS_TO_MS = 1.0E-6;

	// derived quantities:
	// number of indices:
	const int NIDX = sc::napcomb(NCHAN)+NCHAN;	// SI,DSI,DESI
	// old mass-file processing only:
	double idx[NIDX*2] ; 	// normal and shifted indices
	// maximal expected quantity of numbers in a mass-file line:

	// Generalized-mode features (left for the safe case, future experiments)
	std::map <string,double> aperdata;
	std::map <string,string> withaper;
	char mode = '?';
	bool headout = false;	// flag of headers written (#R-header must know naper)
	bool pchanged  = true;	// flag of parameters changed: re-convertion needed
	bool ochanged  = true;	// flag of object changed: weights testing needed
	bool isgen   = false;	// flag of Generalized mode started
	int measnorm = 0, measgen = 0;	// last normal and shifted meas. numbers
	double ra=0, dec=0; // fractional (0<ra<1) RA and radian DEC
	double accumtime = 0; // accumulation time in seconds
	const double DEG2RAD = 0.0174532925199433;
	const double PI2 = 6.283185307179586;
	
	string UTbeg, UTend;
	double uth, uthbeg = 0;
	char t = 0;

	do {
		if (!feof(fmass)) t = rdm::readline(fmass,sw,fstm);
		string ut = rdm::getut();
		/* Make fractional monotonically increasing through night UT-hour: */
		int dummy = 0;
		uth = rdm::getut(&dummy);
		if (longitude<0) { 
			if (uth > 12.-longitude) uth -= 24;
		} else { 
			if (uth < 12.-longitude) uth += 24;
		}	
		// check: is it time to interpret the accumulated results?
		double dtime = (uth - uthbeg)*3600;
		if ((t=='O' || (t=='M' && dtime>=accumtime) || feof(fmass)) && measnorm>0) {
			// one of prefixes listed in stdchr() matches the current one:
			// taken from makeAccumResultScind() (see device.cpp of Turbina):
			sc::average(isgen, sc::MAXLAG);
			string UTmid = get_midtime(UTbeg, UTend);

			if (!headout) {
				/* write heading: */
				sc::writeavg(fout,sc::DSI,sc::MODENORM,	"#I        UTbeg      "," UTend   ",true);
			}
			string stamp = string("I ") + UTbeg;
			sc::writeavg(fout,sc::DSI,sc::MODENORM,stamp.c_str(),UTend.c_str());

			if (!headout) {
				/* write heading: */
				sc::writeavg(fout,sc::AFLUX,sc::MODENORM, "#F        UT         ", 0,true);
			}
			stamp = string("F ") + UTmid;
			sc::writeavg(fout,sc::AFLUX,sc::MODENORM,stamp.c_str());
			if (isgen) {
				stamp[0] = 'G';
				sc::writeavg(fout,sc::AFLUX,sc::MODEGEN,stamp.c_str());
				stamp[0] = 'J';
				sc::writeavg(fout,sc::DSI,sc::MODEGEN,stamp.c_str(),UTend.c_str());
			}

			// taken from makeAccumResultScind() (see device.cpp of Turbina):
			double *scind = sc::getavgidxptr(sc::DSI,sc::MODENORM);
			double *scindShft = isgen ? sc::getavgidxptr(sc::DSI,sc::MODEGEN) : 0 ;
			/* Average Cn2-integrals and compute atm. parameters */
			const int desichan = 0 ;        /* DESI - in A-channel */
			double desi = sc::getavgidx(sc::DESI,isgen?sc::MODEGEN:sc::MODENORM, desichan);
			double edesi = sc::geterridx(sc::DESI,isgen?sc::MODEGEN:sc::MODENORM,desichan);
			atm::avgint(desi, edesi, secz) ;

			if (!headout) {
				/* write heading: */
				atm::write(fout,"#A        UT         ", atm::ATMPAR, atm::HEAD) ;
			}

			stamp[0] = 'A';
			atm::write(fout, stamp.c_str(), atm::ATMPAR);
			
			double * e2scind = sc::geter2idxptr(sc::DSI,sc::MODENORM);
			double * e2scindShft = isgen ? sc::geter2idxptr(sc::DSI,sc::MODEGEN) : 0;

			// Restore the profile with both methods:
			for (int w=0; w<=1; w++) {
				atm::calcn2(scind, scindShft, e2scind, e2scindShft,
					secz, w ? atm::FIXEDLAY : atm::FLOATLAY) ;
				stamp[0] = 'T';
				if (!headout && !w) {
					/* write heading: */
					atm::write(fout,"#T        UT         ", atm::CN2PROF, atm::HEAD) ;
				}
				atm::write(fout, stamp.c_str(), atm::CN2PROF);
				// write the normalized O-C for indices restored by Cn2 profile:
				if (!headout && !w) {
					/* write heading: */
					atm::write(fout,"#R        UT         ", atm::O_CSCIND, atm::HEAD) ;
					if (oneheader) headout = true;
				}
				stamp[0] = 'R';
				atm::write(fout, stamp.c_str(), atm::O_CSCIND);
			}
			// reset local and SCIND's counters of saved instant indices:
			measnorm = measgen = 0;
			sc::again();
		}
		
/* ****** This may be uncommented for comparison with online proc.results: ****/			
//		if (strchr("IARTF",t)) printf("%c %s %s\n",rdm::gettype(),
//			rdm::getut().c_str(),rdm::getvalue().c_str());
		switch (t) {
			case 'P':
				// get the preamble parameters:
				rdm::add2map(rdm::getname(),rdm::getvalue());
				printf("P %s %s=%s\n", rdm::getut().c_str(),
					rdm::getname(1).c_str(),rdm::getvalue().c_str());
				pchanged = true;
				break;
			case 'O':
				{
					string val = rdm::getvalue();
#ifdef GCC33
					istringstream ist(val.c_str()) ; // string input
#else
					istrstream ist(val.c_str()) ; // string input
#endif
					string name, name2, spcl;
					ist >> name;
					if (name.length()<4) {
						// another word is also star name (which we don't need):
						ist >> name2;
						name += ' ';
						name += name2;
					}
					int rah=0,ram=0,ras=0,decd=0,decm=0,decs=0;
					string sdecd;
					ist >> rah >> ram >> ras >> sdecd >> decm >> decs;
					decd = abs(atoi(sdecd.c_str())) ;
					ra = (rah + ram/60. + ras/3600.)/24.;
					dec = (decd + decm/60. + decs/60)*DEG2RAD;
					char sdec = '+';
					if (strchr(sdecd.c_str(),'-')!=NULL) {sdec = '-', dec = -dec;}
					
					// ok, we're by the spectrum:
					ist >> spcl;

					printf("O %s %s %s %i %i %i %c%i %i %i %s\n", rdm::getut().c_str(),
						rdm::getname(1).c_str(),
						name.c_str(), 
						rah, ram, ras, sdec, decd, decm, decs, 
						spcl.c_str());

					// convert the SED name into lower-case and produce the weight
					// file name:
					for (int i=0; i<(int)spcl.length(); i++) spcl[i] = tolower(spcl[i]);
					if (wdir[strlen(wdir)-1]=='/') 
						wfile = string(wdir) + spcl + WFEXT;
					else
						wfile = string(wdir) + '/' + spcl + WFEXT;
#ifdef ATMOSDEBUG
					CERR << "WEIGHT="<< wfile<<'\n';
#endif
					sfile = string(sdir) + '/' + spcl + SEDEXT;
					ochanged = true;
				}
				break;
			case 'M':
				mode = rdm::getname()[0];
				/* If results are not processed - we skip this mode */
				if (measnorm) break;
				/* here we check that star IS selected (hense airmass is a 
				number and thus moved to getvalue()-string; see readmass.h):*/
				if ((mode=='N' || mode=='G') && rdm::getvalue().length()>0) {
					// compute airmass from time:
					int yy,mm,dd,uth,jd;
					double utd,s0,ha;
					// fractional UT-day
					utd = rdm::getut(&yy,&mm,&dd,&uth)/24.;
					// julian day of beginning of the UT-day
					jd = julian(yy,mm,dd);
					// Greenwich stellar time of the beginning of UT-day
					s0 = greenwich_st0(jd);
					const double mu = 1.0027379093; // stellar to mean solar time rate ratio

					if (pchanged) {
						// (re)conversion of P-parameters into internal vars:
						// old (current?) format to set the Cn2 moments rest.grid:
						z0 = strtod(rdm::getvalue("WEIGHTFUNCTION\\MINALTITUDE").c_str(),0);
						zmax = strtod(rdm::getvalue("WEIGHTFUNCTION\\MAXALTITUDE").c_str(),0);
						if (zmax<min_zmax) zmax = min_zmax;
						zstep = strtod(rdm::getvalue("WEIGHTFUNCTION\\STEP").c_str(),0);
						zminstep = strtod(rdm::getvalue("WEIGHTFUNCTION\\MINSTEP").c_str(),0);
						microexp = strtod(rdm::getvalue("OPERATIONS\\NORMAL MODE\\EXPOSITION").c_str(),0);
						magnification = strtod(rdm::getvalue("OPTICS\\COMMON\\MAGNIFICATION").c_str(),0);
						
						// determine fractional longutude and radian latitude:
						string scrd = rdm::getvalue("GENERAL\\SITE\\LONGITUDE");
						istringstream is1(scrd);
						int dh,mm,ss;
						is1 >> dh >> mm >> ss;
						longitude = (abs(dh) + mm/60. + ss/3600.)/24.;
						if (strchr(scrd.c_str(),'-')!=NULL) longitude *= -1.0;
						
						scrd = rdm::getvalue("GENERAL\\SITE\\LATITUDE");
						istringstream is2(scrd);
						is2 >> dh >> mm >> ss;
						latitude = (abs(dh) + mm/60. + ss/3600.)*DEG2RAD;
						if (strchr(scrd.c_str(),'-')!=NULL) latitude *= -1.0;

//						CERR << "Longitude="<<longitude<<"fr Latitude="<<latitude<<"rad\n";


						// new format to set the gridS:
						string s = rdm::getvalue("DATA PROCESSING\\INTEGRALS\\ALTITUDEGRID");
						if (s.length()) {
							int n = atm::NFLOAT; wf::str2da(zgrid, &n, s.c_str());
							z0 = zgrid[0]; zmax=zgrid[1]; zstep=zgrid[2];
							zminstep=zgrid[3];
						}
						s = rdm::getvalue("DATA PROCESSING\\FLOATING LAYERS\\ALTITUDEGRID");
						if (s.length()) {
							int n = atm::NFIXED; wf::str2da(zgrid, &n, s.c_str());
							atm::setzcn2(atm::FLOATLAY,n,zgrid);
						}
						s = rdm::getvalue("DATA PROCESSING\\FIXED LAYERS\\ALTITUDEGRID");
						if (s.length()) {
							int n = atm::NFIXED; wf::str2da(zgrid, &n, s.c_str());
							atm::setzcn2(atm::FIXEDLAY,n,zgrid);
						}
						s = rdm::getvalue("OPERATIONS\\BACKGROUND MEASUREMENT\\BACKGROUND");
						if (s.length()) { int n = NCHAN; wf::str2da(bkgr, &n, s.c_str());}
						for (char c='A'; c!=(char)((int)'A'+NCHAN);c++) {
							int i = int(c) - int('A');
							string s = rdm::getvalue(string("SEGMENTATOR\\CHANNEL ")+c+"\\INNER");
							if (s.length()) {
								double d = strtod(s.c_str(),0);
								innerDiam[i] = d / 10.;
							}
							s = rdm::getvalue(string("SEGMENTATOR\\CHANNEL ")+c+"\\OUTER");
							if (s.length()) {
								double d = strtod(s.c_str(),0);
								outerDiam[i] = d / 10.;
							}
							s = rdm::getvalue(string("BICOUNTERS\\CHANNELS\\CHANNEL")+c);
							if (s.length()) {
								// new MASS: parse the line like "Counter 2A" 
								// into "2" and "A":
								char nbicnt = s[8], chanbicnt = s[9];
								string s2 = string("BICOUNTERS\\BICOUNTER ")+
									nbicnt+"\\NONLINEARITY"+chanbicnt;
								double d = strtod(rdm::getvalue(s2).c_str(),0);
								
								if (d>0) deadtime[i] = d * NS_TO_MS;
								s2 = string("BICOUNTERS\\BICOUNTER ")+
									nbicnt+"\\NONPOISSON"+chanbicnt;
								d = strtod(rdm::getvalue(s2).c_str(),0);
								if (d>0) nonpois[i] = d;
							} 
							// Universal (new and old MASS(-DIMM) devices) specifications
							// like Channels/Channel A/NonPoisson and NonLinearity
							// (may be inserted aposteriori for data reprocessing):
							s = rdm::getvalue(string("CHANNELS\\CHANNEL")+c+"\\NONLINEARITY");
							if (s.length()) {
								double d = strtod(s.c_str(),0);
								int i = int(c) - int('A');
								if (d>0) deadtime[i] = d * NS_TO_MS;
							}
							s = rdm::getvalue(string("CHANNELS\\CHANNEL")+c+"\\NONPOISSON");
							if (s.length()) {
								double d = strtod(s.c_str(),0);
								int i = int(c) - int('A');
								if (d>0) nonpois[i] = d;
							}
							s = rdm::getvalue(string("CHANNELS\\CHANNEL")+c+"\\SCATTER");
							if (s.length()) {
								double d = strtod(s.c_str(),0);
								int i = int(c) - int('A');
								if (d>0) scatter[i] = d;
							}
							// else: overwrite values manually above and rebuild:-(
						}
						double bt = strtod(rdm::getvalue("OPERATIONS\\NORMAL MODE\\BASETIME").c_str(),0);
						accumtime = strtod(rdm::getvalue("OPERATIONS\\NORMAL MODE\\ACCUMTIME").c_str(),0);
						int nmeas = 2*(int(accumtime/bt)+1);
						int basenum = int(1000*bt/microexp)+1;
						sc::init(basenum, nmeas, NCHAN);
						for (int i=0; i<NCHAN; i++) {
#ifdef ATMOSDEBUG
							CERR << "Channel "<< (char)('A'+i) 
							<<" ("<<innerDiam[i]*magnification<<" - "
							<<outerDiam[i]*magnification<<") cm P="
							<<nonpois[i]<<" n="<<deadtime[i]
							<<" ms bg="<<bkgr[i]<<" scat="<<scatter[i]<<'\n';
#endif
							sc::chan_init(i,bkgr[i],deadtime[i],nonpois[i],scatter[i],0);
						}
						double focfeed = aperdata["FEED"];
						double foclens = aperdata[withaper["G2"]] ;
						// here we assume that for positive focal lengths, the negative
						// shift is returned (version of weif at least 1.6 !)
						zshift2 = foclens!=0 ? wf::getzshift(focfeed,foclens) : 0;
				
						foclens = isgen ? aperdata[withaper["G1"]] :  aperdata[withaper["N"]];
						zshift1 = foclens!=0 ? wf::getzshift(focfeed,foclens) : 0;
#ifdef ATMOSDEBUG
						CERR << "Z1=" << zshift1 << " Z2=" << zshift2 << " ZMIN=" << z0 << " ZMAX=" <<zmax << " ZSTEP=" << zstep << " ZminSTEP=" << zminstep << '\n';
#endif
						pchanged = false;
					}
					// radian hour angle (from fractional args):
					ha = (s0 + mu*utd + longitude - ra) * PI2;
					// here we get secz as an approximate value of the more
					// finely determined airmass (stars are all near-zenith)
					secz = airmass(ha,dec,latitude);
#ifdef ATMOSDEBUG
					CERR << "UT= "<<ut<<" ha="<<ha<<" ra="<<ra<<" dec="<<dec<<" M="<<secz<<'\n';
#endif
					if (ochanged && !nocheck) {
						wf::checalc(sfile.c_str(), rfile, wfile.c_str(), NCHAN,
							magnification, innerDiam, outerDiam,
						z0, zmax, zstep, zminstep, true, wf::progress);					
						ochanged = false;
					}
					atm::update(wfile.c_str(),dfile.c_str(),zshift1,zshift2,z0,zmax,zstep,zminstep);
					isgen = (mode=='G');
					printf("%c %s %s %.3f\n",rdm::gettype(),
						rdm::getut().c_str(),rdm::getname(1).c_str(),secz);
				
				} else {
					mode = '?';
					printf("%c %s %s\n",rdm::gettype(),
					rdm::getut().c_str(),rdm::getname(1).c_str());
				}
				break;
				
			case 'i': case 'j': case 'm': case 's':
				if (mode=='N' || mode=='G') {
					if (measnorm+measgen==0) {UTbeg = ut; uthbeg = uth;}
					// flag of shifted indices (Generalized mode only)
					bool shifted = (t=='j' || t=='s');

					double* normidx, *shiftidx;// pointers to current indices

					if (t=='i' || t=='j') {
						// old mass-files: convert the record into inst.indices:
						int n = NIDX;
						/* We put the normal indices (SI,DSI,DESI) in first NIDX
						elements, shifted - after first NIDX skipped */
						wf::str2da(idx+(shifted)*NIDX,&n,rdm::getvalue().c_str());
						normidx = idx;
						shiftidx = &idx[NIDX];
					} else {
						// new mass-file: compute indices from moments:
						sc::str2mom(rdm::getvalue().c_str());
						sc::calcidx(microexp,shifted);
						if (shifted)
							measgen = sc::getcurmeas();
						else
							measnorm = sc::getcurmeas();
						normidx = sc::getidxptr(measnorm-1,sc::DSI);
						if (shifted) shiftidx = sc::getidxptr(measgen-1,sc::DSI);
					}
					// compute the atmospheric integrals:
					if (shifted) atm::calcint(normidx, shiftidx);
					else if (!isgen) atm::calcint(normidx, 0) ;
					// otherwise - wait for generalized indices
					UTend = ut;	// stored here until data interpretation
				}
				break;
			case '#' : {
			// echo the ##-messages of input-filtering scripts and observer:
				if (rdm::getvalue()[1]=='#')
					printf("%s\n",rdm::getvalue().c_str());
			}
			break;
		}

		if (nr::erget()) {
			fprintf(fout,"> %s %s\n",ut.c_str(),nr::ermessage().c_str()) ;
			if (nr::erget()!=nr::ERNOD && nr::erget()!=nr::EROFL && nr::erget()!=nr::recipes_ercode) {
				fprintf(stderr,"> %s %s. Exit\n",ut.c_str(),nr::ermessage().c_str()) ;
				exit(1);
			}
			nr::erreset();
		}
	} while (!feof(fmass));
	atm::done();
	sc::done();
	
	fclose(fout) ;	
	fclose(fmass) ;
	if (fstm) fclose(fstm);
	return 0 ;
}
#endif

/* private part of a namespace at */

namespace atm {

/*------------------------ Global parameters -----------------------------*/

/** Bad result placeholder and value */
double bad = 0 ;

/** Cn^2 profile restoration method:  */
char curcn2method = '?' ;

/** Dimension of (z)cn2 vectors */
const int MAXNLAYER = (NFIXED>NFLOAT) ? NFIXED+2 : NFLOAT+2 ;

/** Actual length of the computed Cn^2-vector from calcn2() */
int ncn2 = 0 ;

/** Cn^2-vector computed by Cn2 restoration calcn2() and Chi2 as last elem.*/
double cn2[MAXNLAYER] ;

/** Seeing computed by restored Cn2() profile. Whole atm. - for Generalized
mode restoration; free atmosphere - for Normal mode. */
double cn2seeing = 0;
double getcn2see() { return cn2seeing;}

/** Altitudes of Cn^2 layers from calcn2() */
double zcn2[MAXNLAYER] ;

/** Basic set of weight functions: */
wf_t atwf ;

/** Combined matrix of weights for Cn2 restoration: used in calcn2() and in
meritfunc() */
double ** wcomb = 0;

/** Shifts by altitude [km] by which normal (first) and generalized (both) 
modes operate (normally non-positive) */
double zshift[NMODE] = {-99,-99} ;

/** Two Matrices of coefficients for the moments of Cn2  */
double** coef[2] = {NULL, NULL} ;	/**< [0]: for non-shifted measurements, [1] 
for generalized mode measurements */

/** Number of photometric channels */
int nchan = 0 ;

/** Number of weights respective to nchan */
int nw = 0 ;

/* Combined average indices and their errors for restoration of
Cn2-profile */
double * sicomb = 0, * e2sicomb = 0 ;

/* Combined (observed-model)/observed indices produced by calcn2() */
double * o_csi = 0;

/** Storage of accumulated vectors of instantaneous turbulence moments */
double ** cn2int ;

/** Physical length of cn2int storage (allocated space) */
int cn2intlen = 0;

/** Current length of cn2int storage (measurement number) */
int curmeas = 0;

/** User's altitude grid for Cn2 moments decomposition. ZMAX serves also as
an upper limit for Cn2 restoration method. Filled in update(). */
double z0, zmax, dz, dzmin ;

/** User's parameter of altitude grid for Cn2 profiles restoration */
double userzcn2[2][MAXNLAYER] = {{Z0CN2_NORM, ZMAX, DZCN2_FIX, DZMINCN2_FIX},
{Z0CN2_NORM, ZMAX, DZCN2_FLO, DZMINCN2_FLO}};

/** Lengths of parameters set of userzcn2[] */
int userncn2[2] = {4,4};

/* Vectors of average integrals and their errors */
double  aint[NPOWER] ;
double e2int[NPOWER] ;

/* Vector of atmospheric parameters and their errors made from aint,e2int */
double atmpar[LAST] ;
double eatmpar[LAST] ;

/* Currently loaded weight file name: to check whether it's changed since last call */
std::string curwffile = "" ;

}

/*============================== Functions ===============================*/

/*------------------------------ getval ----------------------------*/

double atm::getval(ordint what) {
	switch (what) {
		case FSEE:	return (atmpar[FSEE]>0 && atmpar[FSEE]<MAXANGLE) ? atmpar[FSEE] : 
			(atmpar[FSEE]<=0 ? 0 : MAXANGLE);
		case WSEE:	return (atmpar[WSEE]>0 && atmpar[WSEE]<MAXANGLE) ? atmpar[WSEE] : 
			(atmpar[WSEE]<=0 ? 0 : MAXANGLE);
		case FM0:	return (aint[IFSE]>MINM0 && aint[IFSE]<MAXM0) ?	aint[IFSE]:
			(aint[IFSE]<=MINM0 ? MINM0 : MAXM0);
		case WM0:	return (aint[IWSE]>MINM0 && aint[IWSE]<MAXM0) ?	aint[IWSE]:
			(aint[IWSE]<=MINM0 ? MINM0 : MAXM0);
		case FHEFF:	return (atmpar[FHEFF]>MINHEFF && atmpar[FHEFF]<MAXHEFF) ? atmpar[FHEFF] : 
			(atmpar[FHEFF]<=MINHEFF ? MINHEFF : MAXHEFF);
		case WHEFF:	return (atmpar[WHEFF]>MINHEFF && atmpar[WHEFF]<MAXHEFF) ? atmpar[WHEFF] : 
			(atmpar[WHEFF]<=MINHEFF ? MINHEFF : MAXHEFF);
		case ISOPL:	return (atmpar[ISOPL]>0 && atmpar[ISOPL]<MAXANGLE) ? atmpar[ISOPL] : 
			(atmpar[ISOPL]<=0 ? 0 : MAXANGLE);
		case M2:	return (aint[IISK]>MINM2 && aint[IISK]<MAXM2) ?	aint[IISK] :
			(aint[IISK]<=MINM2 ? MINM2 : MAXM2);
		case TC:	return (atmpar[TC]>0 && atmpar[TC]<MAXTC) ? atmpar[TC] : 
			(atmpar[TC]<=0 ? 0 : MAXTC);
		default:
			nr::nrerror("ATMOS: getval(): wrong what-selection",nr::ERPAR);
			return bad ;
	}
}

/*------------------------------ geterr ----------------------------*/

double atm::geterr(ordint what) {
	double v, e ;
	switch (what) {
		case FSEE:	v = atmpar[FSEE]; e = eatmpar[FSEE]; break ;
		case WSEE:	v = atmpar[WSEE]; e = eatmpar[WSEE] ; break ;
		case FM0:	v = aint[IFSE]; 
					e = (e2int[IFSE]>0) ? sqrt(e2int[IFSE]) : bad; break ;
		case WM0:	v = aint[IWSE]; 
					e = (e2int[IWSE]>0) ? sqrt(e2int[IWSE]) : bad; break ;
		case ISOPL:	v = atmpar[ISOPL]; e = eatmpar[ISOPL] ; break ;
		case M2:	v = aint[IISK]; 
					e = (e2int[IISK]>0) ? sqrt(e2int[IISK]) : bad; break ;
		case FHEFF:	v = atmpar[FHEFF]; e = eatmpar[FHEFF] ; break ;
		case WHEFF:	v = atmpar[WHEFF]; e = eatmpar[WHEFF] ; break ;
		case TC:	v = atmpar[TC]; e = eatmpar[TC] ; break ;
		default:
			nr::nrerror("ATMOS: geterr(): wrong what-selection",nr::ERPAR);
			return bad ;
	}
	if (v>0 && v!=bad && e!=bad) 
		return e/v ;
	else
		return 0 ;
}

/*------------------------------ getncn2 -------------------------------*/

int atm::getncn2() {
	return ncn2 ;
}

/*------------------------------ getzcn2 -------------------------------*/

double atm::getzcn2(int i) {
	if (zcn2)
		return zcn2[i] ;
	else {
		nr::nrerror("ATMOS: getzcn2() while no Cn2 computed",nr::ERNUL) ;
		return bad ;
	}
}

/*------------------------------ getcn2 --------------------------------*/

double atm::getcn2(int i) {
	if (cn2)
		return cn2[i] ;
	else {
		nr::nrerror("ATMOS: getcn2() while no Cn2 computed",nr::ERNUL) ;
		return bad ;
	}
}

/*------------------------------ update --------------------------------*/

void atm::update(
	const char * wfile,
	const char * checkfile,
	double zshift1,
	double zshift2,
	double iz0, double izmax, double idz, 
	double idzmin) {

const double KM2M = 1e3 ;


	if (nr::erget()) return ;

	if (!wfile) {
		nr::nrerror("ATMOS: update() on null weight file name", nr::ERNUL) ;
		return ;
	}
#ifdef ATMOSDEBUG
	CERR << "atm::update("<<wfile<<",."<<zshift1<<",."<<iz0<<','<<izmax<<','<<idz<<','<<idzmin<<")\n";
#endif	
	/* Check the weight to coincide with the currently loaded by its file name */
	if (curwffile != string(wfile)) {
		/* weight is either not loaded or different */
		atwf.read(wfile) ;
	} else {
		/* Weight is the same. Check for the shifts and exit if they are so: */
		if ((zshift[0]==zshift1) & (zshift[1]==zshift2)) 
			goto CLEARLAB;
	}
	if (nr::erget()) return ;	/* returns if wf_t::read() got bad */
	
	/* Either weight functions or shifts have changed: redo decomposition */
	{	// this braket is to enclose declarations crossed by jump to CLEARLAB
	/* Cycle with 2 steps: 0-th for moments decomposition for the normal mode;
	   1-st for generalized (2 zshifts) mode */

	/* Check-file opening: */
	FILE * cf = 0 ;
	if (checkfile)
		if (!(cf=fopen(checkfile,"wt"))) {
			nr::nrerror((string("ATMOS: update(): problem to open the "
			"check-file ")+checkfile).c_str(), nr::ERFIO) ;
			return ;
		}

	/* Make the matrices of weights shifted by zshift1 and zshift2 */
	wf_t wfs[NMODE];
	/* Take care of atm::wcomb[][] if dimensions of atwf.ws[][] have changed: */
	if (atm::nw && atm::nw!=atwf.getnw()) {
		nr::free_dmatrix(wcomb,0,2*atm::nw-1,0,atm::nw);
		wcomb = 0;
	}
	atm::nw = atwf.getnw();

	int ibound = 0 ;	/* index of the boundary layer limit in WF z-grid */
	int nz = 0;
	
	// Save user's grid:
	atm::z0 = iz0;
	atm::zmax = izmax;
	atm::dz = idz;
	atm::dzmin = idzmin;
	for (int mode=0; mode<NMODE; mode++)  {
		zshift[mode] = (mode ? zshift2 : zshift1) ;

		/* COMPUTE THE TRANSPOSED WEIGHT MATRIX FOR ALTITUDE POWERS
		DECOMPOSITION */
		// we make the decomposition on the "optimal" grid with the upper limit
		// equal to that given in the weight file:
		for (int imode=0; imode<=mode; imode++) {
			if (mode==0) {
				// Normal mode: omit lower atmosphere in decomposition:
				wfs[imode].setzgrid(atm::HBOUND,atm::zmax,atm::dz, atm::dzmin) ; 
			} else {
				if (!imode) wfs[imode].setaper(0);
				wfs[imode].setzgrid(atm::z0,atm::zmax,atm::dz, atm::dzmin) ; 
			}
			wfs[imode].interpolate(&atwf, zshift[imode]) ;	
			if (nr::erget()) return ;
		}

		nz = wfs[mode].getnz();
		if (mode==0) {
			if (nchan != atwf.getnaper()) {
				nchan = atwf.getnaper();
				/* Allocate the coefficients matrix: */
		        for (int i=0; i<NMODE; i++) 
					if (atm::coef[i]==0) {
//						CERR << "Alloc "<<i<<"-th "<<NPOWER<<"x"<<atm::nw<<"\n";
						atm::coef[i] = nr::dmatrix(0,NPOWER-1,0,2*atm::nw-1) ;
					}
			}
		}
		/* SVD spaces */
		double ** wmatr=nr::dmatrix(1,nz,1,atm::nw*NMODE) ;
		double * wvect=nr::dvector(1,atm::nw*NMODE) ;
		double ** vmatr=nr::dmatrix(1,atm::nw*NMODE,1,atm::nw*NMODE) ;
		double * hvect=nr::dvector(0,nz-1) ;

		/* Find the zgrid-index of the boundary layer end: */

		for (int j=0; j<nz; j++) {
			if (wfs[mode].getalt(j)>=atm::HBOUND) {
				ibound = j;
				break ;
			}
		}

		if (ibound<0) {
			nr::nrerror("ATMOS: update(): failed to find the boundary layer"
			" index",nr::ERZGR); 
			goto CLEARLAB ;
		}
		if (cf) {
			fprintf(cf,"\nApproximated powers weight file %s, zshift=%5.2fkm", 
				wfile, zshift[0]);
			if (mode)
				fprintf(cf," and %5.2fkm",zshift[mode]);
			fprintf(cf,":");
		}
		/* Pointer to the string of previous use-weights flags */
		const char * prevuse = 0;
		int nwused=0;	// number of used weights for current approximation

		for (int ipow=int(atm::IFSE); ipow<=int(atm::IISK); ipow++) {
	
			/* Select the power: */
			double power ;
			switch (poweridx(ipow)) {
				case IFSE:
				case IWSE: power = POWSEE; break ;
				case IISP: power = POWISP; break ;
				case IEFF: power = POWEFF; break ;
				case IISK: power = POWISK; break ;
				default: nr::nrerror("ATMOS: update(): invalid power index",
					nr::ERPAR) ;
					return ;
			}
			if (cf) fprintf(cf, "\n# Ap%d: %6.3f ",(int)ipow,power) ;
			/* Generate the altitude powers: */
			for (int j=0; j<nz; j++) 
				hvect[j] = pow(wfs[mode].getalt(j)*KM2M, power) ;
			if (ipow==int(atm::IFSE) && ibound>0) {
				/* free atmosphere: force the hvect to resemble the smallest
					apertures pair weight behavior: */
				wfs[mode].copy(0,ibound-1,nchan,hvect) ;
				double d = wfs[mode].getval(ibound,nchan) ;
				for (int j=0; j<ibound; j++) hvect[j] /= d ;
				if (cf) fprintf(cf,"(W(AB)-dumped) ");
			}

			if (!prevuse || strcmp(prevuse,atm::USEWEIGHT[ipow])) {
				/* New SVD required: */
				/* Combined transposed matrix of weights for SVD */
				nwused = 0;
				for (int imode=0; imode<=mode; imode++) 
					for (int j=0; j<atm::nw; j++) 
						if (atm::USEWEIGHT[ipow][j+imode*atm::nw]!='0') {
							for (int i=0; i<nz; i++) 
								wmatr[i+1][nwused+1] = wfs[imode].getval(i,j) ;
							nwused++ ;
							if (cf) {
								if (imode)
									fprintf(cf,"%ss,",atwf.name(j));
								else
									fprintf(cf,"%s,",atwf.name(j));
							}
						}
				
				/* Run SVD of weight matrix: */
				nr::svdcmp(wmatr,nz,nwused,wvect,vmatr) ;
				if (nr::erget()) {mode=NMODE; goto ENDLAB ;}

				/* Maximal singular value: */
				double d = wvect[1] ;
				for (int i=2;i<=nwused;i++) 
					if (wvect[i]>d) d=wvect[i] ; /* maximal w */
				/* Minimal acceptable singular value: */
				d *= atm::MINRELW ;	

				int nbase = nwused ;
				/* Rejection of small singulars: */
				if (cf) fprintf(cf, " N(Base fun.): Init.=%d, ",nbase) ;
				for (int i=1;i<=nwused;i++) 
					if (wvect[i]<d) { 
						wvect[i] = 0 ; 
						nbase--; 
					}
				if (cf) fprintf(cf, " Smalls reject=> %d  ",nbase);
			}
			prevuse = atm::USEWEIGHT[ipow];

			/* Back-substitute to find the coefficients: */
			
			// normal mode coefficients for shifted weights and for M0 have to remain zero!
			for (int c=0; c<2*atm::nw; c++) atm::coef[mode][ipow][c] = 0;
			if (mode>0 || zshift[mode]!=0 || ipow!=atm::IWSE) {
				nr::svbksb(wmatr,wvect,vmatr,nz,nwused, hvect-1, 
					atm::coef[mode][ipow]-1);
				// now the coefficients are concentrated in first nwused
				// elements. Redistribute them starting from the tail:
				int k = nwused - 1 ;
				for (int c=atm::nw*(1+mode)-1; c>=0; c--)
					if (atm::USEWEIGHT[ipow][c]!='0') {
						atm::coef[mode][ipow][c] = atm::coef[mode][ipow][k] ;
						k--;
					} else
						atm::coef[mode][ipow][c] = 0 ;
			} // else - leave zero
		
		}
		if (cf) {
			/* Check the coefficients: */
			fprintf(cf,"\n# Coefficients:\n");
			for (int ipow=int(IFSE); ipow<=int(IISK); ipow++) {
				/* Search for the maximal coefficient and compute the noise-factor
				NAF=sum(coef^2)/sum^2(coef): */
				double d = 0 ;
				for (int j=0; j<atm::nw*2; j++) {
					double ac = fabs(coef[mode][ipow][j]) ;
					if (d<ac) d=ac ;
				}
				double naf = 0, sumcoef = 0;
				for (int j=0; j<atm::nw*2; j++) {
					naf += coef[mode][ipow][j] * coef[mode][ipow][j] ;
					sumcoef += coef[mode][ipow][j] ;
				}
				if (sumcoef>0)	
					naf = naf/(sumcoef*sumcoef); 
				else
					if (naf>0) naf = 99.9 ; 

				fprintf(cf,"# Ap%d: Max=%9.3e NAF=%5.1f",ipow,d,naf);
				for (int j=0; j<atm::nw*2; j++) 
					fprintf(cf," %7.4f",d>0?atm::coef[mode][ipow][j]/d:0);
				fprintf(cf,"\n");
			}
			/* Check the approximated relations: */	
			fprintf(cf,"# h[km]  ") ;
			for (int i=0; i<=atm::IISK ; i++) 
				fprintf(cf,"#Ap%d/h^ ",i) ;
			fprintf(cf,"\n") ;
			for (int i=0;i<nz;i++) {
				fprintf(cf,"%6.3f",wfs[mode].getalt(i)) ;
				for (int ipow=int(IFSE); ipow<=int(IISK); ipow++) {
			   		double power ;
					switch (poweridx(ipow)) {
						case IFSE:
						case IWSE: power = POWSEE; break ;
						case IISP: power = POWISP; break ;
						case IEFF: power = POWEFF; break ;
						case IISK: power = POWISK; break ;
						default: goto ENDLAB ;
					}
					/* power of altitude: */
					double a = pow(wfs[mode].getalt(i)*KM2M,power) ;
					double d = 0 ;
					for (int imode=0; imode<=mode; imode++)
						for (int j=0; j<atm::nw; j++)
							d+=wfs[imode].getval(i,j)*coef[mode][ipow][j+imode*atm::nw];
					fprintf(cf," %6.3f",a>0?d/a:0); 
				}
				fprintf(cf,"\n") ;
			}
		}
ENDLAB:
		/* Free the orthogonalized matrix, singular value vector and square 
		V-matrix: */
		nr::free_dmatrix(wmatr,1,nz,1,atm::nw*NMODE) ;
		nr::free_dvector(wvect,1,atm::nw*NMODE) ;
		nr::free_dmatrix(vmatr,1,atm::nw*NMODE,1,atm::nw*NMODE) ;
		nr::free_dvector(hvect,0,nz-1) ;
	}

	/* Close check-file */
	if (checkfile)
		fclose(cf) ;

	}	// closing braket to enclose declarations crossed by jump to CLEARLAB
CLEARLAB:
	/* reset counter of made measurements of integrals: */
	curmeas = 0 ;
	
	/* Reset measurement results: */
	atmpar[FSEE] = eatmpar[FSEE] = atmpar[WSEE] = eatmpar[WSEE] = atmpar[ISOPL] = eatmpar[ISOPL] = atmpar[WHEFF] = eatmpar[WHEFF] = 
	atmpar[FHEFF] = eatmpar[FHEFF] = atmpar[TC] = eatmpar[TC] = bad ;
	curcn2method = '?' ;
	ncn2 = 0;	// hense, no (z)Cn2 zeroing needed
	if (aint && e2int) 
		for (int i=0; i<NPOWER; i++) aint[i] = e2int[i] = 0 ;

	// save the weight file name as a signal of successful work:
	curwffile = wfile ;
}


/*------------------------------ done --------------------------------*/

void atm::done(void) {
	atm::nchan = 0 ;
	
	if (atm::coef[0]){ 
		nr::free_dmatrix(coef[0],0,NPOWER-1,0,atm::nw*2-1);
		nr::free_dmatrix(coef[1],0,NPOWER-1,0,atm::nw*2-1) ;
		coef[0] = coef[1] = 0 ;
	}

	/* Deallocate the storages for integrals: */
	if (cn2int) {
		nr::free_dmatrix(cn2int,0,cn2intlen-1,0,NPOWER-1) ;
		cn2int = 0 ;
		cn2intlen = curmeas = -1 ;
	}
	
	if (wcomb) {
		// last parameter is an unused foo: nz actually
		nr::free_dmatrix(wcomb,0,2*atm::nw-1,0,atm::nw);
		wcomb = 0;
	}
	if (o_csi) {
		nr::free_dvector(o_csi,0,2*atm::nw-1);
		o_csi = 0;
	}

	atwf.setaper(0);		
	atm::nw = 0;
}

/*------------------------------ calcint --------------------------------*/

void atm::calcint(double * scind, double * scinds) {

	if (!coef[0] || !scind) {
		nr::nrerror("ATMOS: calcint() on empty index or coef. array", nr::ERNUL) ;
		return ;
	}
	/* (Re)allocate the storages for integrals: */
	if (curmeas>=cn2intlen) {
		if (atm::cn2intlen>0) {
			double**tmp=nr::dmatrix(0,cn2intlen+NMEAS-1,0,NPOWER-1) ;
			// in a cycle, reassign the rows which are existent removing 
			// the ones in tmp which are not needed. Leave all the new rows intact
			for (int i=0; i<curmeas; i++) {
				delete[] tmp[i];	// simply reassign the allocated array
				tmp[i] = cn2int[i]; // ...
			}
			delete[] cn2int;	// no nr::free_dmatrix!!! It will kill the powers
			cn2intlen += NMEAS;
			cn2int = tmp ;
		} else {
			cn2intlen = NMEAS;
			cn2int = nr::dmatrix(0,cn2intlen-1,0,NPOWER-1) ;
		}
	}

	/* We detect the generalized mode by the appearance of non-NULL scinds */
	bool isgen = (scinds!=0) ;

	/* Compute the Cn2-integrals from indices: */
	for (int i=0; i<NPOWER; i++) {
		double cn2mom = 0 ;
		for (int j=0; j<atm::nw; j++) {
		 	cn2mom += scind[j] * coef[isgen][i][j] ;
		}
		if (isgen) 
			for (int j=0; j<atm::nw; j++) 
			 	cn2mom += scinds[j] * coef[isgen][i][j+atm::nw] ;
		cn2int[curmeas][i] = cn2mom ;
	}
	curmeas++ ;
}

/*------------------------------ getseeing -------------------------------*/

const double JPOW_SEE = 0.6 ;	/* Power of Cn-integral for seeing */

/**
	@brief Convert the given integral of turbulence into seeing
	@param	cn2 Integral of Cn2(h) over atmosphere
	@param	secz	secant of zenith distance
	@return	Seeing [arcsec]
*/
inline double getseeing(double cn2, double secz) {
	const double LPOW_SEE = -0.2 ;	/* Power of lambda for seeing */
	return atm::KS * pow( cn2 / secz, JPOW_SEE) * 
			pow(atm::atwf.getleff(),LPOW_SEE) ;
}

/*------------------------------ avgint ----------------------------------*/

/** Compute the formal relative error of the value of the power function of the
value \em mean given its squared error \em e2 */
inline double relerr(double mean, double e2, double power) {
	if ((e2<=0) || (mean<=0)) 
		return 0 ;
	else
		return fabs(power * sqrt(e2) / mean) ;
}

void atm::avgint(double desi, double edesi, double secz) {
const double JPOW_ISP = -0.6 ;	/* Power of Cn-5/3th moment for isoplan. angle*/
const double LPOW_ISP = 1.2 ;	/* Power of lambda for isoplan. angle */
const double SPOW_TC  = -0.6 ; /* Power of DESI for atm. time constant */
const double KSIGMA = 1; /* negative a is allowed if |a|<KSIGMA*rel.error */
const double KSIGMA2 = KSIGMA * KSIGMA;
	if (curmeas<1) {
		nr::nrerror("ATMOS: avgint() with no measurements done",nr::ERPAR) ;
	}

	/* average integrals: */
	sc::avgmatrix(cn2int,NPOWER,curmeas,0,0,aint,e2int,sc::MAXLAG) ;

	if (nr::erget()) return ;	/* for both two checks and avgmatrix() error */

	/* Convert integrals into atmospheric parameters */
	
	/* Seeing : free and whole atmosphere */	
	atmpar[FSEE] = atmpar[WSEE] = eatmpar[FSEE] = eatmpar[WSEE] = bad ;
	int isbadint = false ;
	if (aint[IFSE]>0) {
		atmpar[FSEE] = getseeing(aint[IFSE], secz) ;
		eatmpar[FSEE] = atmpar[FSEE] * 
			relerr(aint[IFSE],e2int[IFSE],JPOW_SEE);
		if (aint[IWSE]>0) {
			atmpar[WSEE] = getseeing(aint[IWSE], secz) ;
			eatmpar[WSEE] = atmpar[WSEE] * 
				relerr(aint[IWSE],e2int[IWSE],JPOW_SEE);
		} else {
			atmpar[WSEE] = eatmpar[WSEE] = 0;	// not generalized mode?
		}
	} else if (e2int[IFSE]*KSIGMA2 < aint[IFSE]*aint[IFSE]) {
		isbadint = true ;
	}
	
	/* Isoplanatic angle: */
	atmpar[ISOPL] = eatmpar[ISOPL] = bad ;
	if (aint[IISP]>0) {
		double zpow = pow(secz,POWISP+1) ;
		atmpar[ISOPL] = KP * pow(aint[IISP] / zpow, JPOW_ISP)
			* pow(atwf.getleff(), LPOW_ISP) ;
		eatmpar[ISOPL] = atmpar[ISOPL] *
			relerr(aint[IISP],e2int[IISP],JPOW_ISP); 
	} else if (e2int[IISP]*KSIGMA2 < aint[IISP]*aint[IISP]) {
		isbadint = true ;
	}

	/* Effective altitude of turbulence in [meters]: (free atm.) */
	atmpar[FHEFF] = atmpar[WHEFF] = eatmpar[FHEFF] = eatmpar[WHEFF] = bad ;
	if (aint[IEFF]>0 && aint[IISP]>0) {
		double power = 1.0/(POWISP-POWEFF) ;
		double zpow = pow(secz, POWISP-POWEFF) ;
		atmpar[FHEFF] = pow(aint[IISP]/aint[IEFF]/zpow,power) ;
		double d = relerr(aint[IISP],e2int[IISP],power) ;
		eatmpar[FHEFF] = d*d;
		d = relerr(aint[IEFF],e2int[IEFF],power) ;
		eatmpar[FHEFF] += d*d;
		eatmpar[FHEFF] = sqrt(eatmpar[FHEFF]) * atmpar[FHEFF] ;
	} else if (e2int[IEFF]*KSIGMA2 < aint[IEFF]*aint[IEFF]) {
		isbadint = true ;
	}
	
	/* Whole atm. effective altitude: */
	if (aint[IEFF]>0 && aint[IWSE]>0) {
		atmpar[WHEFF] = aint[IEFF]/aint[IWSE] / secz ;
		double d = relerr(aint[IEFF],e2int[IEFF],POWEFF) ;
		eatmpar[WHEFF] = d*d;
		d = relerr(aint[IWSE],e2int[IWSE],POWEFF) ;
		eatmpar[WHEFF] += d*d;
		eatmpar[WHEFF] = sqrt(eatmpar[WHEFF]) * atmpar[WHEFF] ;
	} 

	atmpar[TC] = eatmpar[TC] = bad ;
	if (desi>0) {
		atmpar[TC] = KT * pow(desi,SPOW_TC) ;
		eatmpar[TC] = atmpar[TC] * edesi * fabs(SPOW_TC) ;
	} else if (edesi*KSIGMA<1) {
		isbadint = true ;
	}
	if (isbadint) {
		nr::nrerror("ATMOS: at least one Cn2-integral <0",
			nr::ERNOD) ;
	}
}

/*---------------------------- meritfunc --------------------------------*/
/* Reopen atm once again: */
namespace atm {

int curnweight = 0 ;	// current number of weights (as 10 or 20) for meritfunc()

/**
	@brief	Merit function for the profile-restoration at fixed altitudes
	@param	y	vector[1..ncn2] of sqrt(Cn2) of layers (for powell()) 
					OR vector[\b 0..ncn2] of Cn2 
					if \a z[] is explicitly given (for floating layers)
	@param	z	vector of integer-type altitudes
	@return	merit Chi-square
	
*/
double meritfunc(double * y) ;
double atm::meritfunc(double * y, int * z) ;
}
// for powell():
double atm::meritfunc(double * y) {
	double chi2 = 0 ;	/* Khi-square */
	for (int j=0; j<curnweight; j++) {
		if (atm::FITSCIND[j]!='0') {
			double sc2 = 0 ;	/* scint. index */
			for (int i=0; i<ncn2; i++) 
				sc2 += y[i+1] * y[i+1] * wcomb[j][i] ;
			o_csi[j] = sicomb[j]-sc2 ;
			chi2 += o_csi[j]*o_csi[j] / e2sicomb[j] ;
			if (sicomb[j]>0) o_csi[j] /= sicomb[j] ;
		}
	}
	return chi2 ;
}
// for floating layers method:
double atm::meritfunc(double * y, int * z) {
	double chi2 = 0 ;	/* Khi-square */
	for (int j=0; j<curnweight; j++) {
		if (atm::FITSCIND[j]!='0') {
			double sc2 = 0 ;	/* scint. index */
			for (int i=0; i<ncn2; i++) 
				sc2 += y[i] * wcomb[j][z[i]] ;
			o_csi[j] = sicomb[j]-sc2 ;
			chi2 += o_csi[j]*o_csi[j] / e2sicomb[j] ;
			if (sicomb[j]>0) o_csi[j] /= sicomb[j] ;
		}
	}
	return chi2 ;
}

/*---------------------------- geto_csi ----------------------------------*/

const double * atm::geto_csi() { return atm::o_csi; }

/**
	@brief Matrix SVD invertor
	@param	ainv	Returned inverted matrix
	@param	a	Matrix to be inverted (content is \b modified on return!)
	@param	nrow	(row number in a) = (column number in ainv)
	@param	ncol	(column number in a) = (row number in ainv)
	@param	w	Decomposition work vector of singulars [1..ncol]
	@param	v	Decomposition work orthogonal matrix [1..ncol][1..ncol]
	@param	wcutoff	Minimal acceptable ratio of a w-element to maximal w-element
		for deletion of small singulars in w-vector
	
	The SVD decomposition of input matrix \em a is done and then the
	multiplication of resulted \em v, \em diag(1/w) and decomposed orthogonal
	matrix \em u being the transposed matrix on the place of \em a is done. This
	product is computed in \em ainv.
*/
void invwmatr(double ** ainv, double ** a, int nrow, int ncol, 
	double * w, double ** v, double wcutoff) {

	nr::svdcmp(a,nrow,ncol,w,v) ;
	/* Cancel small singulars */
	double d = 0 ;
	for (int i=1; i<=ncol; i++)	if (d<w[i]) d=w[i];
	d *= wcutoff ;	/* Minimal acceptable singular value */
	for (int i=1; i<=ncol; i++)
		if (w[i]>d) 
			w[i] = 1. / w[i] ;
		else 
			w[i] = 0 ;

	/* multiply three matrices: W^{-1}=(V*diag(1/w))*U^T */
	for (int i=1; i<=ncol; i++) 
		for (int j=1; j<=nrow; j++) {
			ainv[i][j] = 0 ;
			for (int k=1; k<=ncol; k++)
				ainv[i][j] += v[i][k]*w[k]*a[j][k] ;
		}
}

/*---------------------------- setzcn2 ---------------------------------*/

int atm::setzcn2(cn2method meth, int n, double * z) {
#ifdef ATMOSDEBUG
	CERR << "atm::setzcn2()\n";
#endif
	// recognize the type:
	int result = 1;
	if (n==4 && z[1]>z[0] && z[0]+z[2]<z[1]) {
		// likely parametric grid:
	} else {
		if (n<2 || n>NFIXED) result = 0;
		for (int i=1; i<n;i++) {
			if (z[i]<=z[i-1]) {
				// sorting violated
				result = 0;
				break;
			}
		}
	}
	if (result) {
		atm::userncn2[meth==FLOATLAY] = n;
		// likely the manual grid:
		for (int i=0; i<n; i++) userzcn2[meth==FLOATLAY][i] = z[i];
	} else {
		nr::nrerror("ATMOS: setzcn2() invalid parameters",nr::ERZGR);	
	}
	return result;
}

/*---------------------------- calcn2 ----------------------------------*/

double atm::calcn2(double * scind, double * scinds,
		double * e2scind, double * e2scinds, double secz, cn2method meth) {

const double AVGCN2	= 6.95e-13 ;	/* Cn2-integral giving 1" seeing */
const double FTOL = 1e-3 ;		/* Tolerance in powell() */

	if (nr::erget()) return bad;

	/* COMPUTE THE COMBINED WEIGHT MATRIX FOR CN2-profile RESTORATION.
	It is computed on a grid which is stretched with a factor of secant(z_dist)
	with respect to the "standard" grid with equal step DZCN2=1km. Meanwhile, the
	altitudes in zcn2[] are assigned as seen in zenith, with step DZCN2. */
	
	/* Secant of the zenith distance stretchs the zcn2[] grid and amplifies
	the strengths cn2[], so needs to be correctred twice! */

	int nz = 0;
	bool isgen = (scinds!=0) ;	// generalized mode flag
	double cn2zmax = atm::zmax*secz;
	// reject if altitude is twice the maximal or negative:
	if (cn2zmax>2*wf::ZMAX || cn2zmax<0) {
		return MAXANGLE;
	}
	if (cn2zmax>wf::ZMAX) cn2zmax = wf::ZMAX ;

	int IDX = (meth==FLOATLAY);
	
	// equidistant grid with a step corrected by secant of zenith distance:
	wf_t wfs;	// computed twice in this loop: for shifted and not weights
	// - we need this weight structure for altitude referencing further
	for (int mode=0; mode<=isgen; mode++) {
	
		// clear wfs (works before second turn in gen. mode):
		if (mode) wfs.setaper(0);

		// create the z-grid for restoration:
		if (userzcn2[IDX][2]<userzcn2[IDX][1] && userncn2[IDX]==4) {

			double cn2z0 = isgen ? Z0CN2_GEN*secz : userzcn2[IDX][0]*secz ;

			double cn2minzstep = userzcn2[IDX][3] * secz ;

			double cn2zstep = userzcn2[IDX][2] ;

			// step (fixed) must be stretched; modifier (prop. grid) - not
			if (cn2minzstep==0) cn2zstep *= secz;	

			wfs.setzgrid(cn2z0,cn2zmax,cn2zstep,cn2minzstep) ;
#ifdef ATMOSDEBUG
			CERR << " parametric grid (secz="<<secz<<"):"<<cn2z0<<' '<<cn2zmax<<' '<<cn2zstep
				<<' '<<cn2minzstep<<'\n';
#endif
		} else {
			double d[MAXNLAYER];
			for (int i=isgen; i<userncn2[IDX]+isgen; i++) {
				d[i] = userzcn2[IDX][i] * secz;
			}
			if (isgen) userzcn2[IDX][0] = Z0CN2_GEN*secz;
			wfs.setzgrid(userncn2[IDX], d);
#ifdef ATMOSDEBUG
			CERR << " manual grid:";
			for (int i=0; i<userncn2[IDX]+isgen; i++) CERR <<' '<<d[i];
			CERR << '\n';
#endif
		}
		wfs.interpolate(&atwf, zshift[mode]) ;
		nz = wfs.getnz();
		if (!wcomb && !mode) {
			// create weight matrix for Cn2 restoration:
			wcomb = nr::dmatrix(0,2*atm::nw-1,0,nz-1);				
		}
		if (!mode) {
			// create the model indices O-C vector:
			if (!o_csi) o_csi = nr::dvector(0,2*nw-1);
			for (int w=0; w<nw*2; w++) o_csi[w] = 0;
		}
		for (int w=0; w<nw; w++)
			wfs.copy(0,nz-1,w,wcomb[w+mode*nw]);
	}

#ifdef ATMOSDEBUG
// show the altitude grid for restoration:
CERR << "Resto-grid: ";
for (int i=0; i<nz; i++) {
	CERR << wfs.getalt(i)/secz << ' ';
}
CERR << '\n';
#endif

	/* Number of weights i.e. measured indices (not all may be used! see
		atm::FITSCIND): */
	atm::curnweight = atm::nw*(1+isgen);

	/* we need the combined indices and err-s global vectors for merit-function */
	atm::sicomb =   nr::dvector(0,curnweight-1) ;
	atm::e2sicomb = nr::dvector(0,curnweight-1) ;

	for (int i=0; i<nw; i++) {
			sicomb[i] = scind[i] ;
			e2sicomb[i] = e2scind[i] ;
	}

	if (isgen)
		for (int i=0; i<nw; i++) {
			/* works in generalized mode only */
			sicomb[i+nw] = scinds[i] ;
			e2sicomb[i+nw] = e2scinds[i] ;
		}

	/* arrays for selected matrix inversion: far larger than needed */
	double **w4=nr::dmatrix(1,curnweight,1,MAXNLAYER) ;
	double **w4inv=nr::dmatrix(1,MAXNLAYER,1,curnweight) ;
	double *wvect=nr::dvector(1,MAXNLAYER) ;
	double **vmatr=nr::dmatrix(1,MAXNLAYER,1,MAXNLAYER) ;
	double minchi2 = MAXCHI2 ;	/* arbitrary large number */
	double m0prof = 0 ;
	if (meth==FIXEDLAY) {
		/* Fixed layers method */
		/* Generate the grids: altitudes and 1st approximation of the profile */
		atm::ncn2 = NFIXED + isgen ;
		if (atm::ncn2>nz) atm::ncn2 = nz;
		const double cn2_0 = sqrt(AVGCN2 / atm::ncn2) ;
		// altitude grid must be cut to the demanded number of layers! say,@16km
		for (int i=0; i<ncn2; i++) {
			atm::zcn2[i] = wfs.getalt(i) / secz ;
			atm::cn2[i] = cn2_0;
		}

		/* Diagonal "identity" matrix: */
		double ** xi = nr::dmatrix(1,atm::ncn2,1,atm::ncn2) ;
		for (int i=1; i<=ncn2; i++) 
			for (int j=1; j<=ncn2; j++) xi[i][j] = (double)(i==j) ;
			
		/* Solve the task: find the minimum chi-square */
		int iter = 0 ;
		nr::powell(cn2-1,xi,ncn2,FTOL,&iter,&minchi2,atm::meritfunc);
		if (nr::erget()==nr::recipes_ercode) {
			ncn2 = 0;	// not converged
			minchi2 = MAXCHI2;
		} else {
			/* Back-square the profile, correct for sec(z) and compute the sum: */
			for (int i=0; i<ncn2; i++) {
				cn2[i] = (cn2[i]*cn2[i])/secz ;
				m0prof += cn2[i] ;
			}
		}
		nr::free_dmatrix(xi,1,ncn2,1,ncn2) ;
	} else {
		/* "Floating" layers method */
		atm::ncn2 = NFLOAT + isgen ;
		double * tr_cn2=nr::dvector(0,ncn2-1) ;// trial Cn2 vector
		int ih[NFLOAT+1] ;	// memory is either exact (gen) or larger (norm)
		// storage of best-chi2 residuals 
		double * o_csi0 = nr::dvector(0,atm::curnweight-1) ;
		// altitude index limits for layers:
		int maxi[NFLOAT+1] = {0};	// bottom layer is fixed if isgen==true
		maxi[NFLOAT+isgen-1] = nz - 1;
		for (int i=0; i<NFLOAT-1; i++) 
				for (int j=0; j<wfs.getnz(); j++) 
					if (wfs.getalt(j)>MAXHLAY[i]*secz) {
						maxi[i+isgen]=j; 
						break;
					}

		if (isgen) ih[0] = 0 ; /* Fixed ground layer */
		// Wander with (three) floating layers which altitude indices are ih[i]:
		for (int i=isgen; i<NFLOAT+isgen; i++) ih[i] = (i>0)?ih[i-1]+1:0;
		do {
			/* Compose the selected weights matrix w4 */
			for (int j=0;j<curnweight;j++)
				for (int i=0;i<ncn2; i++)
					w4[j+1][i+1] = wcomb[j][ih[i]] ;

			/* Invert w4-matrix */
			invwmatr(w4inv, w4, curnweight, ncn2, 
				wvect, vmatr, MINRELW) ;

			/* Compute the strengths: */
			for (int i=0; i<ncn2; i++) {
				tr_cn2[i] = 0 ;
				for (int j=0; j<curnweight; j++)
					tr_cn2[i] += w4inv[i+1][j+1] * sicomb[j] ;
				if (tr_cn2[i]<0) tr_cn2[i] = 0 ;	  /* >=0 constraint */
			}

			/* Compute indices and chi-square: */
			double chi2 = meritfunc(tr_cn2, ih);

			/* Save the altitudes and strengths if this is the best
				case up to now */
			if (chi2<minchi2) {
				minchi2 = chi2 ;
				m0prof = 0 ;
				for (int i=0; i<ncn2; i++) {
					zcn2[i] = wfs.getalt(ih[i]) / secz ;
					// correct for the sec(z) and save:
					cn2[i] = tr_cn2[i] / secz ;
					m0prof += cn2[i] ;
				}
				// save the residuals; they'll modify further for worse Chi2!
				for (int i=0; i<atm::curnweight; i++) o_csi0[i] = o_csi[i];
			}
			// modify the layer indices:
			bool atlim = false;
			for (int i = NFLOAT+isgen-1; i>=isgen; i--) {
				if (ih[i]<maxi[i]) {
					ih[i]++;
					if (atlim) {
						for (int j=i+1; j<=NFLOAT+isgen-1; j++) ih[j] =
							ih[j-1]+1;
						atlim = false;
					}
					break ;
				} else {
					atlim = true;
					if (i==isgen) 
						goto ENDFLOAT;	// end of iterations
				}
			}
		} while (1) ; // caught here the signal to stop*/
		ENDFLOAT:
		nr::free_dvector(tr_cn2,0,ncn2-1) ;
		// restore the best residuals from saved and remove latters:
		for (int i=0; i<atm::curnweight; i++) o_csi[i] = o_csi0[i];
		nr::free_dvector(o_csi0,0,atm::curnweight-1);
	}

	/* Save the method identity: */
	curcn2method = meth ;

#ifdef TESTIWW
	/* Compose the selected weights matrix w4 (having maybe 6 rows, not 4) */
	for (int j=0;j<curnweight;j++)
		for (int i=0;i<ncn2; i++) {
			int k = int(rint(zcn2[i])) ;
			w4[j+1][i+1] = wcomb[j][k] ;
		}
	/* Invert w4-matrix */
	invwmatr(w4inv, w4, curnweight, ncn2, wvect, vmatr, MINRELW) ;
	
	for (int h=0; h<=atm::zmax; h++) {
		printf("IWW %2d\t",h);
		double tot = 0;
		double prof = 0;
		for (int i=0; i<ncn2; i++) {
			double f = 0;
			for (int j=0; j<curnweight; j++) 
				f += w4inv[i+1][j+1] * wcomb[j][h] ;
			prof += f*cn2[i];
			tot += f;
			printf("%5.2f\t",f);
		}
		if (prof<0) prof=0;
//		printf("%8.2e\n",prof);
		printf("\n");
	}
#endif
	if (minchi2>MAXCHI2)	minchi2 = MAXCHI2 ;
	cn2[ncn2] = minchi2 ;	/* save chi-square in last element */
	nr::free_dmatrix(w4,1,curnweight,1,MAXNLAYER) ;
	nr::free_dmatrix(w4inv,1,MAXNLAYER,1,curnweight) ;
	nr::free_dvector(wvect,1,MAXNLAYER) ;
	nr::free_dmatrix(vmatr,1,MAXNLAYER,1,MAXNLAYER) ;
	nr::free_dvector(sicomb,0,curnweight-1) ;
	nr::free_dvector(e2sicomb,0,curnweight-1) ;

	nr::free_dmatrix(wcomb,0,2*nw-1,0,nw);// last parameter is an unused foo
	wcomb = 0;

	/* compute and return the seeing which corresponds to restored profile.
	Zenith distance is set to 1 since Cn2[] is already corrected for it */
	return (cn2seeing=getseeing(m0prof,1)) ;
}

/**
	@brief	Get the value of the instrument profile centered at \em z0
	@param	z	Altitude to give the result for
	@param	z0	Center altitude for the profile
	@param	resol 	relative resolution in altitude (FWHM/altitude)
	@param	resmin 	minimal altitude resolution (at lowest z) [km]
	@return	PSF value
	
	The instrument profile with the altitude resolution \em res is an arbitrary
	function symmetric in the domain of the altitude logarithm. At lowest
	altitudes, the gaussian is with half-width \em resmin in a linear domain of
	altitudes. The MASS resolution \em res is about 0.5. In current
	implementation, the PSF is a gaussian.
*/
inline double psf(double z, double z0, double resol, double resmin) {
	double ldx = (resmin<resol*z0) ? (log(z)-log(z0))/log(resol) :
				(z-z0)/resmin ;
	return exp(-ldx*ldx) ;
}

void atm::interpolate(double * z, double * intcn2, int nz, 
	double * zin, double * cn2in, int nzin) {
	if (!z || !intcn2 || nz<1) {
		nr::nrerror("ATMOS: intercn2() with null arguments",nr::ERNUL);
		return ;
	}

	const double resol = 0.4 ;	// altitude resolution of MASS
	const double resmin = 0.5; // minimal absolute resolution [km]

	/* In case the default parameters are left, override them with last-computed
	profile data: */
	double * zprof = (zin) ? zin : zcn2;
	double * cn2prof = (cn2in) ? cn2in : cn2;
	int nprof = (nzin) ? nzin : ncn2 ;

	int iupper = 0 ;	// index of zprof[] node which is above the current alt.
	int ilower = -1 ;	// index of zprof[] node which is below the current alt.
	for (int i=0; i<nz; i++) {
		if ((z[i]>zprof[iupper]) && (iupper<nprof)) {
			iupper++ ; ilower++ ;
		}
		// distance to lower layer in units of resolution:
		const double LARGE = 10 ;
		double dl = LARGE ;
		if (ilower>=0) {
			double curres = resol * zprof[ilower] ;
			if (curres<resmin) curres = resmin ;
			dl = fabs(z[i]-zprof[ilower])/curres ;
		}
		// distance to upper layer in units of resolution:
		double dh = LARGE ;
		if (iupper<nprof) {
			double curres = resol * zprof[iupper] ;
			if (curres<resmin) curres = resmin ;
			dh = fabs(z[i]-zprof[iupper])/curres ;
		}
//	printf("%d: %d %d\n",i,ilower,iupper);
		if (dl+dh<2) {
			// two close layers: linear interpolation
			double dz = (z[i]-zprof[ilower]) / (zprof[iupper]-zprof[ilower]) ;
			intcn2[i] =  cn2prof[ilower] * (1-dz) + cn2prof[iupper] * dz ;
		} else {
//	printf("%d: %3.1f %3.1f\n",i,dl,dh);
			double lcn2 = (ilower>=0) ? cn2prof[ilower] * psf(z[i], zprof[ilower], resol, resmin) : 0 ;
			double hcn2 = (iupper<nprof) ? cn2prof[iupper] * psf(z[i], zprof[iupper], resol, resmin) : 0 ;
			intcn2[i] = lcn2 + hcn2 ;
		}
	}
}


/*----------------------------- write -------------------------------*/

const char * atm::write(FILE * f, const char * stamp, atm::what what, 
	bool header) {

	static string os;
	char sbuf[100];
	os = "";
	if (!f) nr::nrerror("ATMOS: write() with null FILE", nr::ERPAR) ;
	if (nr::erget()) return os.c_str();
	
	/* write the stamp: */
	if (stamp) os += stamp;
		
	switch (what) {
		case ATMPAR:
		for (int i=0; i<int(atm::LAST); i++) {
			atm::ordint id = atm::ordint(i) ;
			const char * fl = getfmt(id) ;
			if (header) {
				const char * pn = getname(id) ;// parameter name
				// length of a number in current format: NO FLAGS ASSUMED
				char * endptr ;
				int fmtlen = strtol((char*)(strrchr(fl,'%')+1), &endptr, 10) ;
				string heading = string(ATMDLM)+pn;
				int spaceadded = fmtlen-strlen(pn) ;
				for (int c=0; c<spaceadded;c++) heading+=' ';
				// length of an accuracy number:
				fmtlen = strtol((char*)(strrchr(ERRFMT,'%')+1), &endptr, 10) ;
				heading += string(ATMDLM)+"e_"+pn;
				spaceadded = fmtlen-strlen(pn)-2 ;	// last "2" is len("e_")
				for (int c=0; c<spaceadded;c++) heading+=' ';
				os += heading;
			} else {
				/* write the average integrals and their errors: */
				double d = getval(id); 
				double e = geterr(id) ;
				sprintf(sbuf,fl,atm::ATMDLM,d) ;
				os += sbuf;
				sprintf(sbuf,atm::ERRFMT,atm::ATMDLM,e) ;
				os += sbuf;
			}
		}
		break;
		case CN2PROF:
			if (header) {
				sprintf(sbuf, "%sMet%sNz%sChi2  ",ATMDLM, ATMDLM, ATMDLM) ;
				os += string(sbuf) + ATMDLM + "SeeCn";
				for (int i=0; i<NFIXED; i++) {
					sprintf(sbuf,"%sz_%d %sCn2_%d   ",ATMDLM,i+1,ATMDLM,i+1) ;
					os += sbuf;
				}
			} else {
				/* save the dimention, altitudes and strengths of the restored 
					profile: */
				sprintf(sbuf,"%s  %c%s%2d",ATMDLM,curcn2method,ATMDLM,getncn2());
				os += sbuf;
				sprintf(sbuf,CHI2FMT, ATMDLM,getcn2(getncn2())) ;
				os += sbuf;
				// new addition: seeing by cn2 profile:
				sprintf(sbuf,ANGFMT, ATMDLM, cn2seeing);
				os += sbuf;
				for (int i=0; i<getncn2(); i++) {
					sprintf(sbuf,ALTFMT,ATMDLM,getzcn2(i)) ;
					os += sbuf;
					sprintf(sbuf,CN2FMT,ATMDLM,getcn2(i)) ;
					os += sbuf;
				}
			}
		break;
		case O_CSCIND:
			if (header) {
				os += string(ATMDLM) + "Met";
				// length of an accuracy number:
				char * endptr = 0;
				int fmtlen = strtol((char*)(strrchr(ERRFMT,'%')+1), &endptr, 10) ;
				string heading = string(ATMDLM)+"R_";
				string spaces;
				int spaceadded = fmtlen-2-2 ;	// 2+2 is strlen("R_")+strlen("00")
				for (int c=0; c<spaceadded;c++) spaces+=' ';
				for (int i=0;i<2*atm::nw;i++) {
					sprintf(sbuf,"%s%-2d%s",heading.c_str(),i,spaces.c_str()) ;
					os += sbuf;
				}
			} else {
				/* write the method identity: */
				sprintf(sbuf,"%s  %c",ATMDLM,curcn2method);
				os += sbuf;
				/* Write the normalized index residuals (observed-model/observed) : */
				for (int i=0;i<curnweight;i++) {
					sprintf(sbuf,atm::ERRFMT,atm::ATMDLM,o_csi[i]);
					os += sbuf;
				}
			}

	}

	if (!fprintf(f, "%s\n",os.c_str())) {
		nr::nrerror("ATMOS: write(): cannot write file", nr::ERFIO) ;
	}

	return os.c_str();
}
