/*--------------------------------------------------------------------------*/
/** @file atmos.hxx
                  MASS project: TURBINA module file                       
                       header file for atmos.cpp                             
                                                                          
 The module ATMOS specifies the top-level calculations which are aimed to
 compute the atmospheric parameters and the turbulence profile. The weight
 functions related to calculations are described themselves in the module WEIF.
 With the help of these functions, the various integrals over the altitude
 which approximate the certain moments of the turbulence (calibrated into
 seeing, isoplanatic angles etc) are calculated and the low-resolution
 turbulence profile is computed.

 The task of this module is split in following parts:
 - using the given altitude weight functions, produce the matrix of coefficients
 to convert the vector of scintillation indices into a vector of the turbulence
 moments
 - convert the vector of (base-time, instantaneous) scintillation indices into
 the vector of turbulence moments using the matrix multiplication
 - average the set of the turbulence moments vectors and convert the average
 vector into a set of atmospheric integral parameters (seeing, isoplanatic angle
 etc.)
 - fit the model with a few turbulent layers to reproduce an observed set of
 average scintillation indices (two methods are currently available)

 \b Usage:

 The module in current version does not need initialization, only shutdown
 (atm::done(), see below). Note, that  the initialization of the module SCIND
 has to be done separately while using this module in a usual pipeline of
 procedures within one program. Initialization and update of information on
 the system geometry is done only via the change of weight files provided to
 atm::update().

 Each time the new object is selected or the altitude shift values for the
 generalized mode measurements are changed, the indices-to-turbulence
 moments conversion matrices must be updated with atm::update() using the new
 weight file names and/or altitude shifts. Use wf::checalc() function of WEIF
 module to check the correspondence of the (file of) weight functions set to
 the object's spectral energy distribution (SED).

 Each time some (instantaneous) scintillation indices are obtained, the
 computation of integrals of Cn2*h^a over atmosphere (power a=0,1,5/3,2) has to
 be done with atm::calcint(). The results are stored internally.

 After the completion of the accumulation time, the integrals of Cn2 by powers 
 of altitude are averaged and converted into seeing, isoplanatic angle and
 effective altitude of turbulence by atm::avgint(). Also, the procedure 
 atm::calcn2() may be called which uses the \em average indices and their
 errors to  restore the low-resolution Cn2-profile with one of two methods.
 Results are accessible with \b atm::get*cn2*() functions.

 Once the profile and/or atmospheric characteristics are computed, they may be
 saved on disk by atm::write() with different switching keys atm::what. The
 converted average  integral values and their \b relative errors may be
 accessed by atm::getval() and atm::geterr() after atm::avgint() call.

 Finally, to end with the measurements, call a memory destructor atm::done().
 
 This module provides an option for a stand-alone program which accepts the
 weight functions file and the mass-file data (file or from standard input). On
 input of instantaneous scintillation indices, the turbulence moments are
 accumulated;  on averaged indices, the profile restoration using \em all
 available methods happens and integrals are averaged and converted into
 atmospheric parameters. The needed system parameters are read from
 preamble-type records but may also be specified with input parameters as
 starting values or if no preamble-records are present. Secant of zenith distance is not
 computed but taken from 'O'- and 'M'-records from the last parameter numbers
 (if present).  See main() which should be compiled with the macro ATMOSTEST
 set.

 @version 1.5: Translation into C++  
                                        
 		  2.0: New decomposition, independent executable option (not just a 
		  test) 
		  2.01: added method identity in write(...O_CSCIND...)  
		  2.02: corrected includes for compatibility with gcc v3.3 
		  2.03: write() function now returns the copy of the last written line 
		  2.04: getval() for atm::write(): if value is less than minimal
		  - outputs minimal; if >maximal - maximal (no unique "bad" anymore) 
		  2.1:  autonomous ATMOS: adapt to Turbina >1.3 data (stm-files
		  reading), thus full processing starting with calc. of scint.indices.
		  Also, atm::setzcn2() function is introduced (manual or
		  parametric grids now).
		  2.2: autonomous ATMOS: checks geometry and recomputes weights 
		  if needed; computes sec(z) from UT,lat,fi,ra,dec; some bug fixes.
                  BUG FIX in update(): aint[0..LAST]=0 ->aint[0..NPOWER].
		  2.14: scattered light account in ATMOS
*/

/* This header file is equipped with documentation made in comments in a
* format suitable for production of LaTeX or HTML documentation with the
* doxygen software.
* Editor tabular size: 4                                                    */
/****************************************************************************/

#ifndef _ATMOS_H
#define _ATMOS_H

#include <stdio.h>
#include "nrutil.h"
#include "weif.hxx"
/**
	@brief
	Namespace atm contains the set of functions of highest logical level for
	atmospheric calculations
*/
namespace atm {

/** Order of atmosphere integrals in their storage matrix and output file line.
In a file, each value is followed by its error. The element "LAST" must always 
stay at the end since it gives a number of integrals in enumeration */
enum ordint {
	FSEE,	/**< free seeing */
	WSEE,	/**< whole atm. seeing */
	FM0, 	/**< non-converted integral of Cn2 over free atmosphere */
	WM0, 	/**< non-converted integral of Cn2 over whole atmosphere */
	FHEFF,	/**< effective altitude of turbulence in free atm. */
	WHEFF,	/**< effective altitude of turbulence in whole atm. */
	ISOPL,	/**< isoplanatic angle */
	M2,		/**< second moment of turbulence (integral(Cn2(h)*h^2dh)) */
	TC,  	/**< atmospheric time constant */
	LAST
} ;

/** Writing selector in write(): */
enum what {
	ATMPAR,	/**< write atmospheric parameters */
	CN2PROF,	/**< write the restored Cn2 profile with altitudes and Chi-square */
	O_CSCIND	/**< write normalized observed-model scintillation indices */
} ;

/** Last (defaulted to "write data, not header", i.e. "false") parameter in 
write() */
const bool HEAD = true ;	

/****************************************************************************/
/*                     Declarations of local functions:                     */
/****************************************************************************/


/**
	@brief	Update the data structures for atmospheric calculations with a
		new weight function file
	@param	wfile	The file name of the weight matrix (see wf_t::write()) 
	@param	checkfile	file to print the moment approximation information for
		checking (may be null)
	@param	zshift1	Altitude shift (in [km], normally \b non-positive) for the
		normal mode or for \em one part of the generalized mode measurements
	@param	zshift2	Altitude shift (in [km], normally \b negative) for the 
		\em rest part of the generalized mode measurements
	@param  z0	Cn2 moments decomposition altitude grid: see wf_t::setzgrid()
	@param  zmax	see wf_t::setzgrid()
	@param  dz	see wf_t::setzgrid()
	@param  dzmin   see wf_t::setzgrid()

	This function should be called each time the scintillation weight functions
	are changed due to some reason (new star has another spectrum, new system
	magnification etc.) or the altitude shifts for Generalized mode are
	changed. 

	First, if the supplied name of the weight function file differs from that
	loaded before (the name is stored internally), this new weight functions
	set is read from the file. Also, from the weight file the number of
	indices (equal to that of weights) is derived.

	If neither the weight functions file name, nor the altitude z-shifts of 
	the pupil are changed, nothing is done but only the reset of the computed
	atmospheric parameters.
	
	If either the weights or z-shifts are changed, then the calculations of the
	Cn2 moments are made. They involve the shift of the weights set by \em
	zshift1 and \em zshift2. These shifts can be computed with wf::getzshift()
	and are returned negative by this function for positive focal lengths
	supplied. Thus, the negative zshift implements the generalized mode. 
	
	The decomposition also involves the approximation of the power-law
	dependences of altitude by a set of these shifted weight functions. The SVD
	method is used for this. Approximation enables us to represent the
	Cn2-moments by a set of scintillation indices. If \em zshift1 or \em
	zshift2 differ from each other, then both the coefficients for generalized
	and non-shifted modes are obtained. The coefficients are stored in global
	arrays to be used in calcint().

	The computations are done in a loop which is passed twice. 
	
	In a \b first loop, the job is done for the Normal mode measurements with
	one altitude shift equal to \em zshift1 (normally it is 0). It goes through
	steps:

	- the set of weight functions is shifted by \em zshift1 and
	interpolated into the \em user-defined altitude grid set by \a z0, \a zmax,
	\a dz, \a dzmin for the Cn2 moments decomposition. 

	- the altitude grid is scanned and the grid index IBOUND where the altitude
	reaches the boundary layer height (~1km) is found.

	- the shifted weight matrix is copied with \em transposition into the
	matrix \b W for SVD decomposition. It has dimensions \em ncol=Nw x \em
	nrow=Nz, where the number of weights \em Nw is less than the number of 
	altitudes \em Nz.
	
	- the obtained matrix \b W is SVD-decomposed by NR utility nr::svdcmp()
	into two \em orthogonal matrices \b U and transposed \b V and a diagonal
	matrix of \em singular  values \em diag(w):
	\f[
		({\bf W}) = ({\bf U}) \times diag(w_1 ... w_{Nw}) \times ( {\bf V^T} )
	\f]
	
	- the small singular values are rejected (reset to 0) to have the minimal
	ratio  \f$ w_i / w_{max} \f$ more than atm::MINRELW.
	
	- the altitude power vectors \em h=\em z^k are produced for k=0, 1, 5/3 and 2.
	The zero power (for seeing related Cn2-integral) is made two ways - normal
	exact unity for whole-atmosphere seeing and with "dumped" values below the
	altitude grid index IBOUND for "free atmosphere" seeing. The latter is
	equal to 1 above the boundary layer and mimics the AB-channels differential
	scintillation index	weight below 1km. For the normal mode, the decomposition
	is made starting at 1km altitude.
	
	- for each altitude power, the \em back-substitution utility nr::svbksb() is
	used to produce the coefficients of altitude powers decomposition by the
	weight functions presented in \b W:
	\f[
		{\bf coef} = ({\bf V}) \times diag(1/w_1 ... 1/w_{Nw}) \times ( {\bf
		U^T} \cdot {\bf h} )
	\f]
	These coefficients \b coef - vectors[ \em Nw ] - are stored as lines of a
	matrix with rows number equal to the number of decomposed powers of altitude
	(5 powers).

	- the quality of the approximation of the altitude moments with weight
	functions may be checked if one supplies the non-null file name \em
	checkfile. Then utility writes then the decomposition coefficients and the
	approximated relations in this file. The Noise Amplification Factor (NAF)
	for each coefficient set is also computed and written as a ratio of a sum
	of squared coefficients to the squared sum of coefficients. The file \em
	checkfile is \em rewritten each time (not appended), and consists of two
	similar parts - the first for approximations with \em zshift1 -shifted
	weights (normally - non-shifted) and the second for approximation with both
	\em zshift1 and \em zshift2 -shifted weights.

	In a \b second loop, the similar work is done using the second altitude
	shift \em zshift2. Thus, the \em shifted weights which are marked as usable
	in atm::USEWEIGHT are \b added in the weight matrix for the Cn2 profile
	restoration and in the matrix \b W (rightmost columns this time) for the
	decomposition of altitude powers. The dimension \em Nw is increased
	compared to the first loop by the number of used \em shifted weights.
	
	The derived decomposition coefficients are saved as another matrix of \b
	coef vectors. In the module, these two matrices are allocated in
	two loops and stored as a 2-element vector of pointers. First element
	(i.e. matrix) is accessed by the Normal mode results processing, the second
	is accessed by the Generalized mode processing.
	
	Before exit, the atmospheric integral converted parameters, Cn2 profile
	moments and their errors and number of layers (returned by getncn2()) are
	reset to zero.

	No job is done if error is set (nr::erget()).

*/
void update(const char * wfile, const char * checkfile, 
	double zshift1, double zshift2,
	double z0 = wf::Z0, double zmax = wf::ZMAX, double dz = wf::DZ, 
	double dzmin = wf::DZMIN) ;
				
/**
	@brief	Deallocate the memory reserved for module structures
	@return	void
	
	This function releases the memory allocated for weights, SED,
	coefficients of Cn2-moments decomposition etc by init().
	
	\note
	This function must be kept in strict sync with init(), update() and alloc().
*/
void done() ;

/****************************************************************************/
/*                    Global atmospheric parameters access                  */
/****************************************************************************/

/** 
	@brief Return calibrated atmospheric parameter
	@param	what	one of computed parameters - seeing, Heff, isoplanatic angle
	etc. presented in ordint enumeration
	@return	value or 0 on error
	
	@see avgint()
*/
double getval(ordint what) ;

/** 
	@brief Return the \b relative error of the calibrated atmospheric parameter
	@param	what	one of computed parameters - seeing, Heff, isoplanatic angle
	etc. presented in ordint enumeration
	@return	relative error value or "bad" value on error
	
	If the value of parameters is not "bad", the ratio of the error of the
	parameter to the parameter value itself is returned; otherwise, zero is
	returned
	
	@see avgint()
*/
double geterr(ordint what) ;

/**
	@brief	Return the seeing computed from the last restored profile
	@return	Seeing by Cn2[], arcsec
*/
double getcn2see();

/** 
	@brief Return number of the Turbulence profile altitudes 
	@return Length of the private computed Cn2-array 

	Note that this number is not necessary the length of allocated array
	\em cn2[].
	
	@see calcn2()
*/
int getncn2() ;

/** 
	@brief Get the altitude of the restored turbulence profile layer
	@param	i	layer index [0..getncn2()-1]
	@return altitude [km]
	
	@see calcn2()
*/
double getzcn2(int i) ;

enum cn2method {
	FIXEDLAY = 'X',	/**< "fiXed" layers method */
	FLOATLAY = 'L' /**< "fLoating" layers method */
};	

/**
	@brief Set the user-defined grid for Cn2 profile restoration
	@param meth	Restoration method ID (see atm::cn2method)
	@param n	length of z-array, up to atm::NFIXED
	@param z	grid parameters: either z0,zmax,dz,dzmin or z[0],..z[n]
	@return	1: grid is validated successfully; 0: grid is invalid.
	
	This function allows to set the user-defined altitude grid for the Cn2 
	profile	restoration instead of that used by default. The grid may be
	either of parametric or manual type. By default the grids specified by
	constants atm::Z0CN2_NORM, atm::ZMAX, atm::DZCN2_FIX, atm::DZMINCN2_FIX,
	atm::DZCN2_FLO and atm::DZMINCN2_FLO are used by atm::calcn2().
	
	Parametric grid (see wf_t::setzgrid()) is defined by four parameters 
	\em z0, \em zmax, \em dz and \em dzmin. \em n==4 MUST be supplied.
	It is validated by the checking: zmax>z0, z0+dz<zmax. It is normally
	used for setting if \em meth==atm::FLOATLAY.
	
	Manual grid (newly introduced in ATMOS v2.1) is defined directly as
	a sorted set of altitudes. It is validated by the checking the
	ascending sorting order and that z[0]>0. 
	
	For the Generalized mode restoration the zero altitude is inserted in
	the manual grid or z0 is replaced with 0 for parametric grid before 
	restoration begins (in atm::calcn2()).
	
	\note
	In case the validation described above fails, the return value is 0,
	the current (or default) grid for restoration is not modified and
	the error nr::ERZGR results.

	
*/
int setzcn2(cn2method meth, int n, double * z);

/** 
	@brief Get the strength of the restored turbulence profile layer or
	chi-square
	@param	i	layer index [0..getncn2()-1] or getncn2() for Chi-square 
	@return strength [m^-2/3]
	
	Last element of required array contains the Chi-square measure of the fitted
	profile accessible as getcn2(getncn2()).

	@see calcn2()
*/
double getcn2(int i) ;

/****************************************************************************/
/*                     Calculation of atmosphere parameters                 */
/****************************************************************************/

/**
	@brief Compute the atmosphere integral parameters and Cn2-profile
	@param	scind	array of non-shifted indices
	@param	scinds	array of shifted indices (may be NULL)
	
	Using the Cn2-moments decomposition coefficients and  scintillation
	indices, the calculation of integral atmospheric parameters is performed.
	Number of indices in either \em scind or \em scinds must be equal to number
	of weights determined in update() call. No check of the validity of
	supplied \em scind(s) arrays is done.

	The coefficients of multiplication of the indices in normal and generalized
	modes to obtain the turbulence moments are already precomputed in update();
	the results become accessible after the function call by getval() and
	geterr(). The atmospheric time constant is computed given the \em desi
	parameter.  If \em scind(s) or \em desi parameters are empty (NULL or 0,
	respectively), the calculations for supplied not-NULL parameters are only
	made.

	The turbulence moments of the power \em k (see power list in update()) are
	computed easily as a scalar product:
	\f[
		M_k = (\overline {\bf s} \cdot {\bf coef}_k)
	\f]
	Here the vector \b s of scintillation indices is composed of all non-shifted
	normal and differential indices and shifted indices marked by '1' in
	atm::USEWEIGHT.

	The computed quantities are saved in the current line of the global 
	storage for further averaging by avgint(). This storage is enlarged
	automatically once it is filled (initial size is atm::NMEAS).
	
	\b Performance
	
	One calculation of integrals takes \b 5 ms at PC P-III 667 MHz.

	\note
	If any computed integral appears non-positive, the error results and no
	calibrated parameters are produced.
*/
void calcint(double * scind, double * scinds) ;

/**
	@brief Restore the low-resolution Cn2 profile
	@param	scind	array of average non-shifted indices
	@param	scinds	array of average shifted indices (may be NULL)
	@param	e2scind	array of \em squared errors of \em scind
	@param	e2scinds	array of \em squared errors of \em scinds
	@param	meth	atm::FIXED: Use fixed-altitude layers restoration,
	atm::FLOAT: search for four strongest layers
	@param	secz	current approximate airmass
	@return Minimal chi-square reached, or "bad" value on error
	
	The profile restoration results in altitudes of the getncn2() layers 
	altitudes getzcn2() with strengths getcn2(). Two methods are implemented -
	the search for a few (atm::NFLOAT) strongest layers (\a meth=='L', \em 
	floating \em layers method)	and the search for the intensity of layers
	placed proportionally in the altitude range with relative resolution 0.5
	(\a meth='X', \em fixed \em layers method).

	Similarly to calcint(), the vector \b s of scintillation indices is
	composed of all non-shifted normal and differential indices and shifted
	indices marked by '1' in atm::FITSCIND. 

	Methods operate on the altitude grid created by parameters atm::Z0CN2 and
	others. The weight matrix loaded in atm::update() is interpolated to this
	grid stretched by a factor of secant(zenith distance) and give finally the 
	results	converted to zenith.
	
	\b Fixed layers method
	
	Fixed layers are placed proportionally in altitude like 1, 2, 4, 8, 16 km
	(see atm::DZCN2_FIX and other grid parameters). In the generalized mode
	when shifted indices \em scinds are available, the ground layer at 0km is
	added and the grid starts at atm::Z0CN2_GEN. Thus, \em Nlay=atm::NFLOAT or
	atm::NFLOAT+1 layers
	are fixed in the Cn2 profile altitude grid \em z[] (see getzcn2()). The 6
	or 7 free parameters give thus the strengths of these layers.

	The NR utility nr::powell() is used to minimize the merit function of the
	argument vector \em y[Nlay] used to produce the \em synthetic scintillation
	indices \f$\overline {s^2(syn)}\f$
	\f[
		\chi^2 = \sum\limits_{i=0}^{Nw} \frac{(s^2(syn)_i -
		s^2(obs)_i)^2}{\varepsilon^2_{s^2(obs)_i}}~~,~~\mbox{where}~~~~
		
		s^2(syn) = \sum\limits_{i=0}^{Nlay} y_i^2 \cdot W(z_i)
	\f]
	Here \b W is a weight matrix interpolated to the sec(z)-stretched
	proportional altitude grid (see above). The argument \em y of the
	merit function is taken quadratically to set the non-negativity constraint
	on the restored profile strengths. Output best fitted \em y-s are squared
	and divided by sec(z) to produce Cn2-s returned by getcn2().

	\b Floating layers method
	
	The three trial layers are placed sequentially in a grid defined by
	atm::DZCN2_FLO, atm::DZMINCN2_FLO between the lower limit atm::Z0CN2_NORM
	and the upper boundary supplied to update() as \a zmax. Minimal distance
	between layers is one step of the grid. Upper layer cannot go below lower
	one and lower layers have their maximal altitudes constraint put by
	atm::MAXHLAY.
	
	Number of layers is thus \em Nlay=atm::NFLOAT or one more in generalized
	mode, when one additional layer is \em fixed at the ground. Number of free
	parameters is 2*atm::NFLOAT or one more (6 or 7). For all trial
	combinations of layer altitudes, the \em direct method of the restoration
	of the layers' strengths is performed.

	For this direct method, the weight matrix \em W[Nlay x Nw] is composed from
	the weight matrix read in update() by selecting the needed
	columns corresponding to trial altitudes. Note that trial altitudes are
	expanded by a factor of secant(z) while interpolating the weight matrix as
	in \em fixed layers method.
	
	This matrix \b W is inverted with the help of SVD method nr::svdcmp() (see
	update())
	\f[
		({\bf W})^{-1} = ({\bf V}) \cdot (diag(1/w)) \cdot ({\bf U}^T)
	\f]
	
	Then the strengths corresponding to the trial layers are directly computed:
	\f[
		\overline {C_n^2(syn)} = \overline {\bf s_{obs}} \cdot ({\bf W})^{-1}
	\f]
	
	Using these strengths instead of \f$y^2\f$ in a merit function formulae
	(see above), the chi-square quality of the restored profile with current
	trial altitudes is computed. The combination of altitudes when this
	chi-square is minimal is accepted and available by getzcn2() and getcn2().

	The resulting strengths of layers are divided by secant(zenith distance),
	to get the zenith-referenced result. Thus, the non-zenith position is
	accounted for twice, in altitudes (see beginning of the description) and in
	strengths.
	
	Both methods result also in normalized residuals of scintillation indices as
	"(O-C)/O". They are written by write() with atm::O_CSCIND parameter and
	available also with atm::geto_csi();

	\b Performance
	
	One restoration takes typically \b 22 \em ms at PC P-III 667 Mhz for the
	method with 6 \em fixed layers, and \b 55 \em ms for placing 3 \em 
	floating altitude layers at the proportional grid with 24 steps (altitude
	modifier \f$2^{1/4}-1\f$). For two-times thinner grid at upper atmosphere
	(modifier \f$2^{1/8}-1\f$), it takes already about 320 \em ms.

	No job is done if error is set (nr::erget()).

*/
double calcn2(double * scind, double * scinds,
	double * e2scind, double * e2scinds, 
	double secz, cn2method meth) ;
	
/**
	@brief	Get the normalized observed-model indices restored by calcn2()
	@return	Pointer to residual scint.indices array where first Nidx elements are
		related to non-shifted indices and further go shifted indices (in generalized mode)
	
	This function returns the relative deviations of observed indices from
	restored indices (normalized by observed indices) which correspond to just
	restored Cn2  profile. The deviations are located in a place-holder, which is
	pointed by this function and which is deallocated by done(). Do not
	deallocate thus!

*/
const double * geto_csi();

/**
	@brief	Average Cn2 integrals and compute from them the atmospheric 
		parameters
	@param	desi	average DESI for shifted channel A
	@param	edesi	relative error of \em desi
	@param	secz	current approximate airmass
	
	Given the instantaneous integrals in their storage, their average values and
	standard deviations are computed and put in respective global vectors.
	The function sc::avgmatrix() is used for averaging. Errors thus take into
	account the correlation of values.

	Then the average integrals are converted into atmospheric parameters -
	seeing, isoplanatic angle and effective altitude of turbulence. They are
	available via getval() and - relative errors - via geterr(). Also, the
	atmospheric time constant is computed  from provided \em desi and \em
	edesi values.

	\b Seeing is computed as 
	\f[
		\beta = KS ~\lambda^{-1/5} M_0^{3/5}
	\f]
	where the zero-moment Mo of the turbulence is:
	\f[
		M_0 = \int C^2_n(h) dh
	\f]
	(integration goes from 0 for whole-atm. seeing and starting from \em 
	HBOUND for free-atm. seeing) and \em KS is the calibration constant.

	\b Isoplanatic angle is computed as 
	\f[
		\theta_0 = KP \lambda^{6/5} M_{5/3}^{-3/5} \sec^{8/5}(z)
	\f]
	where the 5/3-th moment of the turbulence profile is:
	\f[
		M_{3/5} = \int C^2_n(h) h^{5/3} dh
	\f]
	and \f$ KP=(2.91 \pi^2)^{-3/5} \f$ is the calibration constant.
	
	\b Second moment of the turbulence is related to the differential 
	astrometry precision and the operation of optical interferometers. Unlikely
	the isoplanatic angle, it is saved in files with no any calibration to any
	"angle".
	
	The \b free atmosphere \b effective \b altitude is computed as 
	\f[
		z_{eff}(free) = (M_{k_1} / M_{k_2})^{1/(k_1-k_2)}
	\f]
	where {\em k1} is 5/3 and {\em k2} is 1.
	
	The value of Heff in the \b whole atmosphere is computed by definition:
	\f[
		z_{eff}(whole) = M_1 / M_0
	\f]
	but only if the shifted indices are available.
	
	The related to adaptive optics \b atmospheric \b time \b constant is
	computed from Differential Exposure Scintillation Index (DESI) \f$
	\sigma^2_{DE}\f$ as

	\f[
		\tau_{de} = KT \left[ \sigma^2_{DE} \right]^{-3/5}
	\f]
	where \em KT is the empirical calibration constant.
	
	All the used integrals are corrected for non-zenith position of the star
	before conversion into atmospheric parameters. For this, they are divided
	by the factor \f$\sec(z)^{k+1}\f$, where \f$\sec(z)\f$ is the secant of the
	zenith distance.

	\note
	During the conversion of integrals into atmospheric parameters, the
	integrals are checked to be greater than \f$-\epsilon\kappa\f$ where
	\f$\epsilon\f$ is the relative error of integral and \f$\kappa\f$ is set to 1. 
	If at least one of integrals is thus
	invalid, the error nr::ERNOD is set which will block the writing of results
	by write() unless error is reset by nr::erreset() after avgint() and
	before write().

	No job is done if error is set (nr::erget()).

*/
void avgint(double desi, double edesi, double secz) ;

/**
	@brief	Write computed atmospheric parameters or restored Cn2 profile or
		residual relative indices or the header line for one of them, trailed by
		CR symbol
	@param	f	FILE pointer
	@param	stamp	leading word in the output line
	@param	what	atm::ATMPAR: write atm. parameters, atm::CN2PROF: write 
		the Cn2 profile, atm::O_CSCIND: write relative residuals of indices
	@param	header	Write the header line instead of data
	@return The line written in the file without CR symbol
	
	This utility writes the output results of atmospheric calculations on the
	disk. 
	
	For \em what==ATMPAR, the single line is written on disk, consisting
	of the average atmospheric parameters each followed by its error.
	Parameters are free seeing ["], seeing ["],	isoplanatic angle ["], second
	moment of turbulence [m^{4/3}],  and atmospheric time constant [ms].

	For \em what==CN2PROF, the Cn2-profile data line is written. It
	consists of 1) the number of adjusted Cn2 layers,  2) Chi-square 3) the
	getncn2() altitudes of layers and each followed by it strength in [m^-2/3].
	4) the seeing computed from the profile strengths
	
	For \em what==O_CSCIND, the relative O-C values for scintillation indices
	resulted from the last restored profile are written

	If \em header==true, the header line for respective \em what -selection is
	written instead of data. \em stamp must then contain the name of data which
	are normally written as a stamp when \em header==false.

	No job is done if error is set (nr::erget()).

*/
const char * write(FILE * f, const char * stamp, atm::what what, bool header = false) ;

/**
	@brief	Interpolate the \b last restored Cn2 profile into new grid
	@param	z	the grid of altitudes to interpolate to [km]
	@param	intcn2	output array[nz] of interpolated cn2 values
	@param	nz	length of \em z[]
	@param	zin	the grid of input altitudes (last computed profile by default)
	@param	cn2in	the grid of input Cn2 strength (last computed profile by default)
	@param	nzin	the length of \a zin[] and \a cn2in if not default
	
	The system response ("PSF") for the thin turbulence influence is assumed to
	be a  Gaussian in a logarithm altitude domain with a 0.5 resolution. In
	case of two "close" layers (e.g. fixed layers method with altitude
	modification factor equal to 2), the linear interpolation is used. In other
	cases, the influence of two adjacent PSFs are accounted for or of the one
	when all the layers are above or below  the current altitude \em z.

	It is suggested to use this function with a proportional altitude grid 
	\em	z[] (in a logarithmic altitude domain) for the graphic representation of
	the Cn2 profile. The output array \em intcn2[] must be allocated before the
	function call.
	
	The default parameters are aimed to replace the last-computed Cn2 profile
	data with other arrays \a zin, \a cn2in, both of the length \a nzin. This is
	needed when re-drawing all the profile data on the display.
*/
void interpolate(double * z, double * intcn2, int nz, 
	double * zin = 0, double * cn2in = 0, int nzin = 0);

/*==========================================================================*/
/*                             Definitions                                  */
/*==========================================================================*/

/** Number of modes: normal and generalized, hence = 2 */
const int NMODE	= 2 ;

/** Minimal ratio of the singular value to the maximal one */
const double MINRELW = 1e-3 ;

/** Maximal number of orthonormal functions to use in h-powers restoration: (to
apply after MINRELW based rejection) */
const int MAXNBASE = 10 ;

/** Starting capacity of the integrals storage, measured in base-times per
accumulation time. After filling the storage, it is reallocated with
NMEAS-more capacity */
const int NMEAS = 256 ;

const double POWSEE = 0 ;		/**< Power of Cn2 moment for seeing */
const double POWISP = 5/3. ;	/**< Power of Cn2 moment for isoplanatic angle  */
const double POWEFF = 1.0 ;		/**< Power of Cn2 moment for Effective altitude */
const double POWISK = 2.0 ;		/**< Cn2 second moment power */

/** Number of power laws to fit: 2 kinds of 0-th, 5/3-th, 1-st, 2-nd */
const int NPOWER = 5 ;

/** Binary mask for using the weights: "1"="use it". Each line consists of the
normal weights mask and shifted weights mask concatenated.
 Now we thus use all non-shifted and As, Bs and ABs weights for all powers. */
const char USEWEIGHT[NPOWER][2*10+1] = {
            "11001000000000000000",
            "10001000001000100000",
            "11111111111100100000",
            "11111111111100100000",
            "11111111111100100000" };
/* palette:  ABCDAAABBCABCDAAABBC
                 BCDCDD    BCDCDD
             nonshifted_shifted__
*/

/** Binary mask for fitting the particular indices with Cn2 model. As a rule,
all the non-shifted and two least aperture size shifted indices must be satisfied.*/
const char FITSCIND[] = 
            "11111111111100100000" ;
/* palette:  ABCDAAABBCABCDAAABBC
                 BCDCDD    BCDCDD
             nonshifted_shifted__
*/

/** Height of the boundary layer [kilometers] to fit the Cn2-integral for 
free seeing (h^0 in integral is forced to resemble the AB-pair weight function
below this height) */
const double HBOUND = 1.0 ;

/** Arrangement of the moment decomposition coefficients in the coef[] array */
enum poweridx {
IFSE,	/**< coef[Free seeing][]: */
IWSE,	/**< coef[Seeing][] */
IISP,	/**< coef[Isoplanatic angle][] */
IEFF,	/**< coef[Turbulence effective altitude][] */
IISK 	/**< coef[Isokinetic angle][] */
} ;

/** Calibration constant in seeing formula (AstL 45, p.399): 
\f$ seeing[''] = KS * \lambda_{eff}^{-1/5} M_0^{3/5} * 
\sec^{-1}(z)\f$, where \em lambda_eff is in [mkm]*/
const double KS = 1.73588e+07 ;

/** Calibration constant in isoplanatic angle formula (AstL 45, p.399)
\f$ \theta_0[''] = KP * (M_{5/3} /\lambda_{eff}^2)^{-3/5} * \sec^{8/5}(z) \f$, where
\f$ KP = (2.905*(2\pi)^2*(MKM2M)^{-2})^{-3/5}*206265 \f$*/
const double KP = 0.000756348 ;

/** Atmospheric time constant empirical calibration constant for DESI for 1,3ms
integrations in diameter=2cm apertures with -1km shift */
const double KT = 0.175 ;

/* Formats of the records of average integral and their errors in the file. 
The order of strings follows the order of integrals specified in \em ordint.
Each format must be preceded with "\%s" to insert a delimiter specified in 
INTDLM.
*/

/** The relative error format (same as in scind.hxx) with a leading "\%s" 
for a delimiter (length corresponds to the length of longest accuracy heading)*/
const char ERRFMT[] = "%s%7.3f" ;

/** Seeing and Isoplanatic angle [arcsec] format (with leading delimiter's \%s) */
const char ANGFMT[] = "%s%5.2f" ;

/** Effective altitude [m] format (with leading delimiter's \%s) */
const char HEFFMT[] = "%s%5.0f" ;

/** Atmospheric time constant format (with leading delimiter's \%s) */
const char TAUFMT[] = "%s%5.2f" ;

/** Delimiter of entries in the atmospheric data output file */
const char ATMDLM[] = " " ;

/** Chi-square writing format (with leading delimiter's \%s) */
const char CHI2FMT[] = "%s%6.2f" ;

/** Altitude of layers record format with a leading "\%s" for a delimiter */
const char ALTFMT[] =  "%s%4.1f" ;

/** Cn2-Profile values and second turbulence moment format  (with leading delimiter's \%s) */
const char CN2FMT[] = "%s%8.2e" ;

/** Number of fitted layers with fixed altitudes. Maximal possible also. Equals
to number of altitudes to interpolate the weight matrix to.
In generalized mode, an additional (NFIXED+1)-th ground layer is added at z=0 */
const int NFIXED = 6 ;	

/** Number of fitted layers with floating altitudes.
In generalized mode, an additional (NFLOAT+1)-th ground layer is added at z=0 */
const int NFLOAT = 3 ;	

/** Cn2-profile restoration altitude grid. Produced by wf::setzgrid(). */
const double Z0CN2_NORM = 0.5 ;	/**< starting altitude in normal mode, [km] */
const double Z0CN2_GEN = 0 ;	/**< starting altitude in generalized mode, [km] */

const double ZMAX = wf::ZMAX/2 ; /** <maximal altitude of a restored layer, [km] */
const double DZCN2_FIX = 1.0 ;	/**< Fixed method 
altitude modifier: =1.0 makes the grid like 1,2,4,8,16*/
const double DZCN2_FLO = pow(2.0,1./8.)-1; /**< Floating layers method 
altitude modifier: =2^0.25-1 makes a 24-step grid (see
atm::DZMINCN2_FLO), =2^0.125-1 - a finer grid but taking 8 times longer for
restoration. */

const double DZMINCN2_FIX = 0.5 ; 	/**< Minimal step for fixed layers method.
//Makes the grid strictly as (0) 0.5 1 2 4 8 16 (if NFIXED=6) */ 
//const double DZMINCN2_FIX = 1.0 ; 	
/*Alternative grid starting at 1km: Makes the grid strictly as (0) 1 2 4 8 16 
(if NFIXED=5) */ 

const double DZMINCN2_FLO = 0.09 ; 	/**< Minimal step for floating layers method.
The grid (0.5,max,2^0.25-1,0.09) (see wf_t::setzgrid()) has nodes at 1,2,4,6,16 km
and three intermediate points between each pair of adjacent nodes */ 

/** Maximal altitudes of the floating layers [km] except for the upper one which
can go up to the limit \a zmax set by atm::update(). */
const double MAXHLAY[NFLOAT-1] = {5., 12.} ;	
//const double MAXH1 = 5.; const double MAXH2 = 12.; for former atmos.cpp vers.

/* Limits for replacing the results with "bad" value if out of them
(format overflow prevent): */
const double MAXANGLE = 99;	// maximal seeing and isoplanatic angle
const double MINHEFF = 1;	// minimal effective altitude, meters
const double MAXHEFF = 99999.;	// maximal effective altitude, meters
const double MINM0 = 1e-17;	// minimal turbulence integral
const double MAXM0 = 1e-10;	// maximal turbulence integral
const double MINM2 = 1e-8;	// minimal turbulence second moment
const double MAXM2 = 1e-2;	// maximal turbulence second moment
const double MAXTC = 99;	// maximal time constant, ms

const double MAXCHI2 = 9999;	// maximal "interesting" chi-square value

/**
	@brief Get format respective to integral ID
	@param	i Integral ID
	@return C-format
*/
inline const char * getfmt(atm::ordint i) {
	switch (i) {
		case FSEE: case WSEE: case ISOPL: return ANGFMT;
		case FM0: case WM0: case M2: return CN2FMT;
		case FHEFF: case WHEFF: return HEFFMT;
		case TC: return TAUFMT;
		default: nr::nrerror("ATMOS: getfmt(): unexpected format switch",
						nr::ERPAR);
		return "";
	}
}

/**
	@brief	Give a name of a field in output file
	@param	i integral sequential number atm::ordint
	
	The names atm::SEENAME ... atm::TAUNAME are returned; for "whole atmosphere"
	parameters - starting from the second character (first is "f" (free)).
*/
inline const char * getname(atm::ordint i) {
static const char SEENAME[] = "fSee" ;	//< name of (free) Seeing in file
static const char ISOPLNAME[] = "Isopl" ;	//< name of Isoplanatic angle in file
static const char M0NAME[] = "fM0";	//< name of zero turbulence moment (Cn2 integral) in file
static const char HEFFNAME[] = "fHeff";	//< name of (free) effective turbulence altitude in file
static const char M2NAME[] = "M2" ;	//< name of second moment of turbulence in file
static const char TAUNAME[] = "Tau";	//< name of atm. time constant in file

	switch (i) {
		case FSEE: return SEENAME ;
		case WSEE: return SEENAME+1 ;
		case ISOPL: return ISOPLNAME ;
		case WM0: return M0NAME+1 ;
		case FM0: return M0NAME ; 
		case M2: return M2NAME ;
		case FHEFF: return HEFFNAME ;
		case WHEFF: return HEFFNAME+1 ;
		case TC: return TAUNAME ;
		default: nr::nrerror("ATMOS: getname(): unexpected name switch",
						nr::ERPAR);
		return "";
	}
}
} // of namespace atm
#endif
