/**
	\mainpage MASS Software. Part II: Data Processing

	\section intro Instroduction
	
	This document represents the second part of the MASS Software Reference 
	Manual and gives a detailed information on how the scientific output information
	is produced from the PMT count data. The preceding Part I dealt with
	the MASS device control issue where the place of scientific modules in
	the general hierarchy of TURBINA modules was specified. So, the
	question "where, when and how the [atmos/scind/scan] module is used in
	TURBINA" should be addressed there.
		
	\section modules Scientific modules for MASS
	
	There is a number of modules which serve for \em handling of scintillation
	indices (SI), atmospheric parameters and turbulence profiles etc.
	"Handling" means the creation/calculation of respective entities, their
	write/read operations with files and export in other modules via some so
	called "get"-functions.

	\image html scimod.jpg "Organization of scientific modules in MASS Software"
	\image latex scimod.eps "Organization of scientific modules in MASS	Software" height=10cm
	
	Scientific modules are organized in three hierarchical levels. On the top,
	the atmospheric calculations module stays (ATMOS), on the bottom - the
	memory and mathematics (NR and NRUTIL). In the middle - the count-processing
	utilities SCIND and SCAN and theoretic weight calculator WEIF.
	
	The module on a particular level can refer and use the modules from the
	same level and below which is reflected by the arrows on the figure. The 
	module names without boxes stand for the ones outside the "scientific" scope
	of modules for TURBINA (e.g. the night-report creator SUMMARY). Dashed
	arrows represent the \em weak dependence and the particular structures
	(functions or types) which are taken from the module for usage are
	indicated near such arrows. For example, it's only the type count_t of the
	detector pulse counts which is taken from the (non-scientific) module
	IOCOUNT.
	
	Each module has a header section names "Usage" which describes the
	sequence of calls to particular procedures which maintains the desired
	way of data processing.

	See the \b File \b Index for the further detailed description of these 
	modules.
*/

/**	@file nrutil.h
	@brief	memory and errors handling in NR and data processing utilities 
		of TURBINA.

	MASS project: header file for nrutil.cpp and nrutil.c 
	from Numerical Recipes, former rewritten in C++.

	The module implements the utilities for handling the dynamic memory storages
	(vectors and matrices) in the C++ style. The "new" and "delete"
	instructions are used instead of "malloc" as in original NR's nrutil.c.
	Thus, no nrerror() is called from the utilities themselves. 
	
	In addition to reimplementation of utilities, the error handling system is
	made with utilities nrerror() for setting and erget(), ermessage() for
	checking of error codes and their messages. The error codes are collected
	here for all the modules and made generalized in sense as much as possible.
	The sense of the error code may be verbalized by call to ercodemessage().

	If NR_CPP macro is defined, the C-linkage conventions for NR recipes are
	not needed. Their source files are assumed to be converted in C++
	compatible form and put in the namespace "nr", the same as one for
	utilities declared in this module. 

	
	@author	N.Shatsky after NR Inc.
	@version 1.5
*/

#ifndef _NRUTIL_H
#define _NRUTIL_H

/* Force C++ compilation of nr and nrutil modules: */
#ifndef NR_CPP
#define NR_CPP
#endif

#include <string>

/** All the mathematics with "float"-type is done actually in double-type in
nr.c, nrutil.cpp and MASS applications */
#define float double

#ifdef __cplusplus
#ifndef NR_CPP
extern "C" {
#else
namespace nr {
#endif
#endif	/* of __cplusplus */

/** The code set by nrerror() by default, thus - by any standard recipe. */
const int recipes_ercode = -1 ;	

/** 
	@brief Get the code of error (0 if no) set in the module
*/
int erget() ;

/** 
	@brief	Get the message string set with the code returned by erget().
*/
const std::string& ermessage() ;

/**
	@brief The error message by code
	@param	ercode code
	@return	static character string which corresponds to the code
	
	The codes are limited to the fixed set nr::ERLO..nr::ERHI; their meaning is
	obtained with this function. The particular cause of the error occured is 
	rather obtained by ermessage().
*/
const char * ercodemessage(int ercode) ;

/* Error codes which scientific modules functions may set : see ercodemessage()*/
const int ERLO = (0);
/* Fatal errors : programming error */
const int ERNUL =(-2);   /**< Error code: NULL pointer supplied */
const int ERMEM =(-3);   /**< Error code: MEMORY allocation error */
const int ERNNL =(-4);   /**< Error code: Allocation attempt on non-NULL array */
const int ERPAR =(-5);   /**< Error code: Non-sense parameter */
const int ERFATAL = ERPAR ;
 
/* Non-fatal errors, e.g. by bad input data */
const int ERFIO =(-6);   /**< Error code: Bad file info or I/O error */
const int EROFL =(-7);   /**< Error code: Not expected count series (index
                                                storages overfull) */
const int ERNOD =(-8);   /**< Error code: no data supplied for calculations */
const int ERZGR =(-9);   /**< WEIF: Altitude grids are incompatible */
const int ERSPE =(-10);   /**< WEIF: Spectral Energy Distributions are incompatible */
const int ERGEO =(-11);  /**< WEIF: Apertures geometry or number are different */
const int ERCRV =(-12);  /**< ATMOS: Spectrum and response curve do not match */
const int ERHI =(-12);

/**	
	@brief	Reset the error code 
	
	This function has to be called to allow the further processing of data
	outside the nr-module (i.e. in SCIND and ATMOS) where it is blocked on
	occurence of the error.
*/
void erreset() ;

/**
	@brief	Error code and message setting. 
	
	The body of the function is evaluated if no ercode is still set, to prevent
	the overwriting of the information on the error which occured first. Called
	from NRs with no second argument.  By convention, this function serves for
	declaring the error occured in any  other "scientific" module in MASS
	Software - SCIND, WEIF, SCAN and ATMOS.

*/
void nrerror(const char * ermsg, int ercode = recipes_ercode);

float *vector(int,int);
float **matrix(int,int,int,int);
double *dvector(int,int);
double **dmatrix(int,int,int,int);
int *ivector(int,int);
int **imatrix(int,int,int,int);
void free_vector(float*,int,int);
void free_dvector(double*,int,int);
void free_ivector(int*,int,int);
void free_matrix(float**,int,int,int,int);
void free_dmatrix(double **, int,int,int,int);
void free_imatrix(int**,int,int,int,int);
/* Not tested so far in C++ implementation: */
float **submatrix(float**,int,int,int,int,int,int);
void free_submatrix(float**,int,int,int,int);
float **convert_matrix(float*,int,int,int,int);
void free_convert_matrix(float**,int,int,int,int);

#ifdef __cplusplus
}
#endif

#undef float

#endif
