/*--------------------------------------------------------------------------*/
/** @file star.h
@brief  Reads the star list and interprets the star parameters when needed.

                  MASS project: TURBINA module file
                       header file for star.cpp

Can read the list of target stars, to search for the file of the weight 
functions corresponding to the given star SED,
keeps the current star parameters and position on sky.



 @author O. Voziakova, Sternberg Institute (ovoz@sai.msu.ru)
 
 @version 1.0
*/
#ifndef STAR_H
#define STAR_H

#include <qobject.h>
#include <qstring.h>
#include <qarray.h>
#include <qdatetime.h>
#include "trbconst.h"

double toValidTimeInterval(double);
double toDeg(double);
void toStrHour(double, QString&);		// 0-24h
void toStrHour12(double, QString&);		// +- 12h
void toStrDeg(double, QString&);

class StarInfo {
private:
	QString	_hr;
	QString	_name;

	double	_ra;		// ra in tsec
	double	_dec;		// declination in asec
	double	_decRad;	// declination in radians
	double	_ha;		// hour angle in tsec
	double	_zenit;		// deg
	double	_airMass;
	QString	_ra_str;	// qstring
	QString	_dec_str;	// qstring
	QString	_ha_str;	// qstring
	QString	_zenit_str;	// qstring
	QString	_airMass_str;	// qstring

	QString	_magV;
	double	_magVval;
	double	_magBVval;
	QString	_sed;
	QString _spectrum;
	QString	_duplicity;

public:

	StarInfo(QString& hr,
			QString& name,
			double ra,
			double dec,
			QString& magV,
			double	magBVval,
			QString& sed,
			QString& spectrum,
			QString& duplicity);
	~StarInfo();
	void	update(double);

	QString&	getHR(void) { return _hr; }
	QString&	getName(void) { return _name; }
	QString&	getMagV(void) { return _magV; }
	double		getMagVval(void) { return _magVval; }
	double		getMagBVval(void) { return _magBVval; }
	QString&	getSpectrum(void) { return _spectrum; }
	QString&	getSED(void) { return _sed; }
	QString&	getDuplicity(void) { return _duplicity; }
	QString&	getRA(void);
	QString&	getHA(void);
	QString&	getDec(void);
	double		getDecVal(void) { return _dec; }
	QString&	getZenit(void);
	double		getZenitVal(void) { return _zenit; }
	QString&	getAirMass(void);
	double		getAirMassVal(void) {return _airMass; }
};

class Star : public QObject {
	Q_OBJECT
private:
	QArray<StarInfo*>	starList;
	int			_nStars;
	int			_currentStarNumber;

// Files & Dirs:
	QString			_sedDirName;
	QString			_wfDirName;
	QString			_wfFileName;
	QString			_responseFileName;
	QString			_starFileName;

// Optic:
	double			_magnification;
	double			_innerDiam[N_CHAN];
	double			_outerDiam[N_CHAN];

// Weight Functions:
	double _minAltitude, _maxAltitude, _gridStep, _minGridStep;

public:
	Star(void);
	~Star();

	QString		composeWfFileName(QString&);
	QString		composeSedFileName(QString&);

	void		loadConst(bool isFirst=true);
	void		init(int dialog);

	QString		notFound(QString&);
	QString		invalidFormat(QString&, int);
	QString		invalidRA(QString&, int);
	QString		invalidDec(QString&, int);
	QString		emptyFile(QString&);

	int		loadStarData(int);

	int		searchMinAirMass(void);
	void		updateAll(void);
	void		updateCurrent(void);
	QString&	getHR(void);
	QString&	getName(void);
	QString&	getSpectrum(void);
	QString&	getSED(void);
	QString&	getWFName(void) { return _wfFileName; }
	QString&	getDuplicity(void);
	QString&	getMagV(void);
	double		getMagVval(void);
	double		getMagBVval(void);
	QString&	getRA(void);
	QString&	getHA(void);
	QString&	getDec(void);
	double		getDecVal(void);
	QString&	getZenit(void);
	double		getZenitVal(void);
	QString&	getAirMass(void);
	double		getAirMassVal(void);
	QArray<StarInfo*>&	getStarList(void) { return starList; }
	int			getNStars(void) { return _nStars; }
	int			setCurrentStar(int number);
	int			setCurrentStar(const QString& hr);
	int			getCurrentStar(void) {return _currentStarNumber;}
	bool		isStarSelected(void);

private slots:
	void change(void);
};

#endif
