/*--------------------------------------------------------------------------*/
/** @file widgets.h
@brief  Implementation of the different components the main window of TURBINA 
collected together by mywidget.h.

                  MASS project: TURBINA module file
                       header file for widgets.cpp



 @author O. Voziakova, Sternberg Institute (ovoz@sai.msu.ru)
 
 @version 1.0
*/
#ifndef WIDGETS_H
#define WIDGETS_H

#define QT_THREAD_SUPPORT
#include <qthread.h>
#include <qapplication.h>
#include <qpushbutton.h>
#include <qdialog.h>
#include <qlayout.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qmultilinedit.h>
#include <qcombobox.h>
#ifdef __QT3__
	#include "qttableview.h"
#else
	#include <qtableview.h>
	typedef QTableView QtTableView;
#endif
#include <qtextview.h>
#include <qlcdnumber.h>
#include "viewstar.h"
#include "trbexport.h"
#include "device.h"

#define DIALOG_POLLING_INT	500	// 0.5 sec
#define CENTER			0 // CorrectionDialog for Center
#define TRACK			1 // CorrectionDialog for Tracking


class QJKPushButton : public QPushButton {
public:
	QJKPushButton( const char*, QWidget*, const char* name=0 );
	~QJKPushButton();
protected:
	void focusInEvent( QFocusEvent* );
	void focusOutEvent( QFocusEvent* );
};

class QJKTable : public QtTableView {
public:
	QJKTable( QWidget* parent=0,
			const char* name=0,
			int numRows=MAX_ROW,
			int numCols=MAX_COL,
			bool grid=false );
	~QJKTable();

	void setCellContent( int row, int col, const QString& );

	void fillTable( void );
	void fitToFont( void );

protected:
	void paintCell( QPainter*, int row, int col );
	virtual void resizeEvent( QResizeEvent* );
	
private:
	int indexOf( int row, int col ) const;
	QString* contents;
	bool showGrid; 

	void setCellGeometry( void );
};

class HelpView : public QDialog {
private:
	QGridLayout* layout;
	QLabel* label;
	QPushButton* closebutton;

	QTextView* textView;
	void loadHelp( const QString& fileName );
public:
	HelpView( const QString& helpFileName,
			QWidget*, const char* name=0 );
	~HelpView();
};

/*
class QJKTextView : public QTextView {
private:
	QString text;

public:
	QJKTextView( QWidget*, const char* name=0 );
	~QJKTextView();

	void addLine( const QString& );
};
*/

class QJKMainView : public QWidget {
private:
	QJKTable *table;
	QLineEdit *textView;

	void setChildrenGeometry( void );
public:
	QJKMainView( QWidget* parent=0, const char* name=0 );
	~ QJKMainView();
	void fitToFont( void );
	void fillTable();
	void showTextLine();

protected:
	virtual void resizeEvent( QResizeEvent* );
};

class TalkDialog : public QDialog {
	Q_OBJECT
public:
	TalkDialog( void );
	~TalkDialog();
private:
	QVBoxLayout* layout;
	QLabel* label;
	QJKPushButton* closebutton;
	QComboBox* combo;
	QMultiLineEdit* edit;
private slots:
	void slotTalk( const QString& );
};

class AdjustDialog : public QDialog {
	Q_OBJECT
public:
	AdjustDialog( void );
	~AdjustDialog();
	void update( void );

private:
	int lastAperture;
	QGridLayout* layout;
	QLabel *label, *label_move, *label_apert, *label_correct;
	QLineEdit *result_window;
	QJKPushButton *acceptbutton, *cancelbutton;
	QJKPushButton *move_plus,*move_minus;
	QComboBox *combo_apert;

private slots:
	void movePlus( void );
	void moveMinus( void );
	void setCurrentItem( int );
	void setAperture( const QString& );
};

class CenteringDialog : public QDialog {
	Q_OBJECT
public:
	CenteringDialog( void );
	~CenteringDialog();
	void update( void );
	void fitToFont( void );

private:
	QGridLayout* layout;
	QLCDNumber *lcdDx,*lcdDy;
	QLabel *labelHeader,*labelDx,*labelDy;
	QJKPushButton *startButton;
	QJKPushButton *closeButton;

	int	timer;
	bool	_polling;

	PointDouble	delta;

private:
	void showResult( void );

private slots:
	void start( void );
	void finish( void );

protected:
	void timerEvent( QTimerEvent *e );
};

class CommentDialog : public QDialog {
	Q_OBJECT
public:
	CommentDialog( void );
	~CommentDialog();
	void update( QString* p_comment );
	void fitToFont( void );

private:
	QString* comment;
	QGridLayout* layout;
	QLineEdit* edit;
	QLabel *labelHeader;
	QPushButton *acceptButton,*cancelButton;

private slots:
	void acceptText( void );
};

class ShowStarDialog : public QDialog {
	Q_OBJECT
public:
	ShowStarDialog( int );
	~ShowStarDialog();
	void update( int );
	void showHA( void );

private:
	int _showHA;
	QGridLayout* layout;
	QLabel *labelName,*labelHA,*labelDec,*labelMagV,*labelSp,*labelDupl;
	QJKPushButton *closeButton;
};

class SelectStarDialog : public QDialog {
	Q_OBJECT
public:
	SelectStarDialog( void );
	~SelectStarDialog();
	void update( void );

private:
	QGridLayout* layout;
	QLabel* label;
	ViewStarList* viewStarList;
	QJKPushButton *selectbutton, *cancelbutton;

private slots:
	void select( void );
	void select( QListViewItem* );
};

class IllumDialog : public QDialog {
	Q_OBJECT
public:
	IllumDialog( void );
	~IllumDialog();
	void update( void );

private:
	QGridLayout* layout;
	QLabel* label;
	QLineEdit* bright_window;
	QJKPushButton *bright_plus,*bright_minus, *closebutton;
	void setBrightness( void );

private slots:
	void setBrightnessUp( void );
	void setBrightnessDown( void );
};

class LightDialog : public QDialog {
	Q_OBJECT
public:
	LightDialog( void );
	~LightDialog();
	void update( void );

private:
	QGridLayout* layout;
	QLabel* label;
	QLineEdit* level_window;
	QJKPushButton *acceptbutton, *cancelbutton;
	void setLevel( void );

private slots:
	void acceptLight( void );
};

class HVDialog : public QDialog {
	Q_OBJECT
public:
	HVDialog( void );
	~HVDialog();
	void update( void );

private:
	QGridLayout* layout;
	QLabel* label;
	QLineEdit* level_window;
	QJKPushButton *acceptbutton, *cancelbutton;
	void setLevel( void );

private slots:
	void acceptVoltage( void );
};

class BackCorrectionDialog : public QDialog {
	Q_OBJECT

private:
	QGridLayout* layout;
	QLabel *label,*label_cf,*label_l,*label_r,*label_av;
	QLabel *label_a,*label_b,*label_c,*label_d;
	QLineEdit* edit_l[N_CHAN];
	QLineEdit* edit_r[N_CHAN];
	QLineEdit* edit_av[N_CHAN];
	QJKPushButton *acceptbutton, *cancelbutton;
	QJKPushButton *stopbutton, *repeatbutton;

	int timer;
	matrix43 coef;

public:
	BackCorrectionDialog( void );
	~BackCorrectionDialog();
	void update( void );
	void finish( void );
	void setButtonsEnabled( bool );

protected:
	void timerEvent( QTimerEvent* );

private slots:
	void acceptCorrection( void );
	void correct( void );
	void showResult( void );
};

class CorrectionDialog : public QDialog {
	Q_OBJECT
private:
	int _what; // CenterCorrection - 0, TrackCalibration - 1
	QGridLayout* layout;
	QLabel *label,*label_x,*label_y;
	QLineEdit* edit_x;
	QLineEdit* edit_y;
	QJKPushButton *acceptbutton, *cancelbutton;
	QJKPushButton *stopbutton, *repeatbutton;

	bool		_polling;
	int		timer;
	PointDouble	delta;

public:
	CorrectionDialog( const QString name,
				const QString win1name,
				const QString win2name,
				int what );
	~CorrectionDialog();
	void fitToFont( void );
	void update( void );
	void finish( void );
	void setButtonsEnabled( bool );

protected:
	void timerEvent( QTimerEvent* );

private slots:
	void acceptCorrection( void );
	void correct( void );
	void showResult( void );
};

class DigitalClock : public QLCDNumber {
	Q_OBJECT
public:
	DigitalClock( QWidget* parent=0,const char *name=0 );
	~DigitalClock();
	void fitToFont( void );
public slots:
	void showTime();
};

#endif
